﻿using System.Windows.Forms;
using System.Collections.Generic;
using System;

namespace RetailDemoApp
{
    #region enum FormName
    /// <summary>
    /// Contains all usable form names.
    /// </summary>
    public enum FormName
    {
        Inventory,
        HighVolumeInventory,
        Monitor,
        Barcode,
        Read,
        Write,
        Battery,
        DetectorSettings,
        //COMPortSettings,
        NFCCard,
        WaitScreen
    }
    #endregion

    /// <summary>
    /// Class used to take care of form creation and switching. Creates all forms on init.
    /// </summary>
    public class FormsManager : IDisposable
    {
        #region Properties

        private Dictionary<FormName, Form> forms = new Dictionary<FormName, Form>();
        private delegate void ShowFormDelegate(FormName formName);

        private frmRetailDemo inventory;
        private frmRetailDemo highVolumeInventory;
        private frmMonitor monitor;
        private frmBarcode barcode;
        private frmReadSingle read;
        private frmWriteEPC write;
        private frmBatteryLife battery;
        private frmNFCCard nfcCard;
        private frmWaitScreen waitScreen;
        private frmDetectorSettings detectorSettings;
        //public frmCOMPortSettings comPortSettings;

        public FormName CurrentForm { get; private set; }
        public FormName PreviousForm { get; private set; }

        #endregion

        #region FormsManager() - constructor
        public FormsManager()
        {
            #region Create forms
            inventory = new frmRetailDemo(false);
            highVolumeInventory = new frmRetailDemo(true);
            monitor = new frmMonitor();
            barcode = new frmBarcode();
            read = new frmReadSingle();
            write = new frmWriteEPC();
            battery = new frmBatteryLife();
            nfcCard = new frmNFCCard();
            detectorSettings = new frmDetectorSettings();
            //comPortSettings = new frmCOMPortSettings();
            waitScreen = new frmWaitScreen();
            #endregion

            #region Set form names
            inventory.Text = Program.ApplicationName;
            highVolumeInventory.Text = Program.ApplicationName;
            monitor.Text = Program.ApplicationName;
            barcode.Text = Program.ApplicationName;
            read.Text = Program.ApplicationName;
            write.Text = Program.ApplicationName;
            battery.Text = Program.ApplicationName;
            nfcCard.Text = Program.ApplicationName;
            detectorSettings.Text = Program.ApplicationName;
            //comPortSettings.Text = Program.ApplicationName;
            waitScreen.Text = Program.ApplicationName;
            #endregion

            #region Add forms to Dictionary
            forms.Add(FormName.Inventory, inventory);
            forms.Add(FormName.HighVolumeInventory, highVolumeInventory);
            forms.Add(FormName.Monitor, monitor);
            forms.Add(FormName.Barcode, barcode);
            forms.Add(FormName.Read, read);
            forms.Add(FormName.Write, write);
            forms.Add(FormName.Battery, battery);
            forms.Add(FormName.NFCCard, nfcCard);
            forms.Add(FormName.DetectorSettings, detectorSettings);
            //forms.Add(FormName.COMPortSettings, comPortSettings);
            forms.Add(FormName.WaitScreen, waitScreen);
            #endregion

            CurrentForm = FormName.WaitScreen;
            PreviousForm = FormName.Inventory;
            Program.Controller.Session.OnSessionHandheldDisconnect += new EventHandler<EventArgs>(Session_OnSessionHandheldDisconnect);
        }
        #endregion

        #region Public methods

        //#region AttachToEvent()
        ///// <summary>
        ///// Attach to HandheldDisconnect event
        ///// </summary>
        //public void AttachToEvent()
        //{
        //    Program.Controller.Session.OnSessionHandheldDisconnect += new EventHandler<EventArgs>(Session_OnSessionHandheldDisconnect);
        //}
        //#endregion

        #region GetCurrentForm()
        /// <summary>
        /// Gets current form.
        /// </summary>
        /// <returns>Current form.</returns>
        public Form GetCurrentForm()
        {
            return forms[CurrentForm];
        }
        #endregion

        #region ShowForm(FormName formName)
        /// <summary>
        /// Shows desired form.
        /// </summary>
        /// <param name="formName">Form name to be shown.</param>
        public void ShowForm(FormName formName)
        {
            forms[formName].Show();
            forms[CurrentForm].Hide();
            PreviousForm = CurrentForm;
            CurrentForm = formName;
        }
        #endregion

        #region Dispose()
        /// <summary>
        /// Dispose forms.
        /// </summary>
        public void Dispose()
        {
            inventory.Dispose();
            highVolumeInventory.Dispose();
            monitor.Dispose();
            barcode.Dispose();
            read.Dispose();
            write.Dispose();
            battery.Dispose();
            nfcCard.Dispose();
            detectorSettings.Dispose();
            //comPortSettings.Dispose();
            waitScreen.Dispose();

            if (Program.Controller.Session != null)
                Program.Controller.Session.OnSessionHandheldDisconnect -= new EventHandler<EventArgs>(Session_OnSessionHandheldDisconnect);
        }
        #endregion

        #endregion

        #region Events

        #region Session_OnSessionHandheldDisconnect(object sender, EventArgs e)
        private void Session_OnSessionHandheldDisconnect(object sender, EventArgs e)
        {
            waitScreen.BeginInvoke(new ShowFormDelegate(ShowForm), FormName.WaitScreen);
        }
        #endregion

        #endregion
    }
}