﻿using System;
using System.Collections.Generic;
using System.Text;

namespace Nedap.EPC.TDT
{
    /// <summary>
    /// Class that represents an electronic product code (EPC) 
    /// </summary>
    public class EpcCode
    {
        /// <summary>
        /// Constructor for the EpcCode class.
        /// Note that this version of the constructor have an array of bytes as an input parameter.
        /// </summary>
        /// <param name="binaryEpc">EPC as an array of bytes in binary representation</param>
        public EpcCode(byte[] binaryEpc)
        {
            BinaryData = binaryEpc;
        }
        /// <summary>
        /// Constructor for the EpcCode class.
        /// First, check whether the translation of the input string in binary format succeeded.
        /// If it is succeeded, first convert it to a hexadecimal string and then, convert that
        /// hexadecimal string to an array of bytes. If it is failed, convert a input string to
        /// an array of bytes.
        /// Note that this version of the constructor have have a string as an input parameter.
        /// </summary>
        /// <param name="stringEpc">EPC string in binary or hexadecimal representation</param>
        public EpcCode(string stringEpc)
        {
            try
            {
                // try to parse with TDT
                string binary = TDTEngine.Translate(stringEpc, LevelTypeList.BINARY);
                HexString = EpcConvert.ConvertToHexStringFromBinaryString(binary);
                BinaryData = EpcConvert.ConvertToDataFromHexString(hexString);
            }
            catch
            {
                try
                {
                    // TODO: check if it's an HEX string
                    BinaryData = EpcConvert.ConvertToDataFromHexString(stringEpc);
                }
                catch
                {
                    throw;
                }
            }
        }
        /// <summary>
        /// Constructor for the EpcCode class used for the 14-digit GTIN.
        /// </summary>
        /// <param name="GTIN14">14-digit GTIN</param>
        /// <param name="serial">Serial number for GTIN</param>
        /// <param name="companyPrefixLength">Length of GS1 Company Prefix</param>
        public EpcCode(string GTIN14, ulong serial, uint companyPrefixLength)
        {
            // create legacy encoding
            //legacy = [NSString stringWithFormat:@"gtin=%@;serial=%llu", GTIN14, [serial unsignedLongLongValue]];
            Legacy = "gtin=" + GTIN14 + "serial=" + serial.ToString();
            //Log("legacy: {0}", Legacy);

            // create parameter list
            string parameterList = "filter=1;taglength=96;gs1companyprefixlength=" + companyPrefixLength.ToString();
            //Log("parlist: {0}", parameterList);		

            string binary = TDTEngine.Translate(Legacy, parameterList, "BINARY");
            HexString = EpcConvert.ConvertToHexStringFromBinaryString(binary);
            BinaryData = EpcConvert.ConvertToDataFromHexString(hexString);
            //Log("hexString: {0}", HexString);
        }

        /// <summary>
        /// Public property that contains the EPC code as how it's stored in an RFID tag - binary data.
        /// </summary>
        public byte[] BinaryData { get; private set; }
        /// <summary>
        ///  Public property that returns length in bits of the EPC code which is in binary format.
        /// </summary>
        public uint LengthInBits
        {
            get
            {
                return (uint)BinaryData.Length * 8;
            }
        }

        private string binaryString;
        /// <summary>
        /// Public property that contains the EPC code as how it's stored in an RFID tag - in a string with ones and zeros.
        /// </summary>
        public string BinaryString
        {
            get
            {
                if (binaryString == null)
                {
                    binaryString = EpcConvert.ConvertToBinaryStringFromData(BinaryData);
                }
                return binaryString;
            }
            private set
            {
                binaryString = value;
            }
        }

        private string hexString;
        /// <summary>
        /// Public property that contains the EPC code converted to hexadecimal string from a binary string.
        /// </summary>
        public string HexString
        {
            get
            {
                if (hexString == null)
                {
                    hexString = EpcConvert.ConvertToHexStringFromBinaryString(BinaryString);
                }
                return hexString;
            }
            private set
            {
                hexString = value;
            }
        }

        private string tagEncoding;
        /// <summary>
        /// Public property that contains the EPC code translated from binary string to TAG_ENCODING level
        /// of representation which is in string format.
        /// </summary>
        public string TagEncoding
        {
            get
            {
                if (tagEncoding == null)
                {
                    tagEncoding = TDTEngine.Translate(BinaryString, LevelTypeList.TAG_ENCODING);
                }
                if (tagEncoding == null)
                {
                    tagEncoding = HexString;
                }
                return tagEncoding;
            }
            private set
            {
                tagEncoding = value;
            }
        }

        private string pureIdentityURI;
        /// <summary>
        /// Public property that contains the EPC code translated from binary string to PURE_IDENTITY level
        /// of representation which is in string format.
        /// </summary>
        public string PureIdentityURI
        {
            get
            {
                if (pureIdentityURI == null)
                {
                    pureIdentityURI = TDTEngine.Translate(BinaryString, LevelTypeList.PURE_IDENTITY);
                }

                if (pureIdentityURI == null)
                {
                    pureIdentityURI = hexString;
                }
                return pureIdentityURI;
            }
            private set
            {
                pureIdentityURI = value;
            }
        }

        private string legacy;
        /// <summary>
        /// Public property that contains the EPC code translated from binary string to LEGACY level
        /// of representation which is in string format.
        /// </summary>
        public string Legacy
        {
            get
            {
                if (legacy == null)
                {
                    legacy = TDTEngine.Translate(BinaryString, LevelTypeList.LEGACY);
                }

                if (legacy == null)
                {
                    legacy = hexString;
                }
                return legacy;
            }
            private set
            {
                legacy = value;
            }
        }

        private string onsHostname;
        /// <summary>
        /// Public property that contains the EPC code translated from binary string to ONS_HOSTNAME level
        /// of representation which is in string format.
        /// </summary>
        public string OnsHostname
        {
            get
            {
                if (onsHostname == null)
                {
                    onsHostname = TDTEngine.Translate(BinaryString, LevelTypeList.ONS_HOSTNAME);
                }

                if (onsHostname == null)
                {
                    onsHostname = hexString;
                }
                return onsHostname;
            }
            private set
            {
                onsHostname = value;
            }
        }
        /// <summary>
        /// Equals
        /// </summary>
        /// <param name="obj">Input object</param>
        /// <returns>False if input parameter is null or input parameter type is not
        /// equal to matching EPC type or input parameter value is not equal to matching EPC,
        /// true if input parameter value is equal to matching EPC</returns>
        public override bool Equals(object obj)
        {
            if (obj == null || this.GetType() != obj.GetType())
            {
                return false;
            }
            EpcCode epc = (EpcCode)obj;
            return ByteArraysEqual(this.BinaryData, epc.BinaryData);
        }
        /// <summary>
        /// Method that checks whether the two bytes are equal
        /// </summary>
        /// <param name="b1">First byte</param>
        /// <param name="b2">Second byte</param>
        /// <returns>True if equal, false if not equal</returns>
        public bool ByteArraysEqual(byte[] b1, byte[] b2)
        {
            if (b1 == b2) return true;
            if (b1 == null || b2 == null) return false;
            if (b1.Length != b2.Length) return false;
            for (int i = 0; i < b1.Length; i++)
            {
                if (b1[i] != b2[i]) return false;
            }
            return true;
        }

        /// <summary>
        /// Get hash code.
        /// </summary>
        /// <returns>Returns hash code from base implementation.</returns>
        public override int GetHashCode()
        {
            return base.GetHashCode();
        }

        /// <summary>
        /// Returns EPC as an array of bytes where input parameter is an array of bytes.
        /// </summary>
        /// <param name="binaryEpc"></param>
        /// <returns></returns>
        public static EpcCode CodeWithBinaryData(byte[] binaryEpc)
        {
            return new EpcCode(binaryEpc);
        }

        /// <summary>
        /// Returns EPC as an array of bytes where input parameter is in string format.
        /// </summary>
        /// <param name="stringEpc"></param>
        /// <returns></returns>
        public static EpcCode CodeWithString(string stringEpc)
        {
            return new EpcCode(stringEpc);
        }

        /// <summary>
        /// Returns the 14-digit GTIN as an array of bytes.
        /// </summary>
        /// <param name="GTIN14">14-digit GTIN</param>
        /// <param name="serial">Serial number for GTIN</param>
        /// <param name="companyPrefixLength">Length of GS1 Company Prefix</param>
        /// <returns>14-digit GTIN</returns>
        public static EpcCode CodeWithGTIN14(string GTIN14, ulong serial, uint companyPrefixLength)
        {
            return new EpcCode(GTIN14, serial, companyPrefixLength);
        }
        /// <summary>
        /// Property that returns hash code
        /// </summary>
        public uint Hash
        {
            get
            {
                return (uint)((short)BinaryData[BinaryData.Length - 2] * 256 + (short)BinaryData[BinaryData.Length - 1]);
            }
        }
        /// <summary>
        /// Returns PURE_IDENTITY URI string
        /// </summary>
        /// <returns></returns>
        public string Description()
        {
            return PureIdentityURI;
        }
    }
}