﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Text;
using System.Text.RegularExpressions;
using System.Xml;
using System.Reflection;

namespace Nedap.EPC.TDT
{
    #region enum XmlDefinitions
    /// <summary>
    /// Enumeration of  TDT library XML definitions
    /// </summary>
    public enum XmlDefinitions
    {
        /// <summary>
        /// Represents GDTI-113 TDT XML definition
        /// </summary>
        GDTI_113,
        /// <summary>
        /// Represents GDTI-96 TDT XML definition
        /// </summary>
        GDTI_96,
        /// <summary>
        /// Represents GIAI-202 TDT XML definition
        /// </summary>
        GIAI_202,
        /// <summary>
        /// Represents GIAI-64 TDT XML definition
        /// </summary>
        GIAI_64,
        /// <summary>
        /// Represents GIAI-96 TDT XML definition
        /// </summary>
        GIAI_96,
        /// <summary>
        /// Represents GID-96 TDT XML definition
        /// </summary>
        GID_96,
        /// <summary>
        /// Represents GRAI-170 TDT XML definition
        /// </summary>
        GRAI_170,
        /// <summary>
        /// Represents GRAI-64 TDT XML definition
        /// </summary>
        GRAI_64,
        /// <summary>
        /// Represents GRAI-96 TDT XML definition
        /// </summary>
        GRAI_96,
        /// <summary>
        /// Represents GSRN-96 TDT XML definition
        /// </summary>
        GSRN_96,
        /// <summary>
        /// Represents SGLN-195 TDT XML definition
        /// </summary>
        SGLN_195,
        /// <summary>
        /// Represents SGLN-64 TDT XML definition
        /// </summary>
        SGLN_64,
        /// <summary>
        /// Represents SGLN-96 TDT XML definition
        /// </summary>
        SGLN_96,
        /// <summary>
        /// Represents SGTIN-198 TDT XML definition
        /// </summary>
        SGTIN_198,
        /// <summary>
        /// Represents SGTIN-64 TDT XML definition
        /// </summary>
        SGTIN_64,
        /// <summary>
        /// Represents SGTIN-96 TDT XML definition
        /// </summary>
        SGTIN_96,
        /// <summary>
        /// Represents SSCC-64 TDT XML definition
        /// </summary>
        SSCC_64,
        /// <summary>
        /// Represents SSCC-96 TDT XML definition
        /// </summary>
        SSCC_96,
        /// <summary>
        /// Represents USDOD-64 TDT XML definition
        /// </summary>
        USDOD_64,
        /// <summary>
        /// Represents USDOD-96 TDT XML definition
        /// </summary>
        USDOD_96
    }
    #endregion

    // Idea is to convert all XML files to binary format, which is smaller and easier to load when application starts.
    // Deployed app would have only these binary files, which we would create beforehand. 
    // This way we could load all files in acceptable time frames. Downside of this is that we need to implement and unit test our own saving and loading to file, and that would take roughly 2-3 days.

    ///<summary>
    /// This class provides methods for translating an electronic product code (EPC)
    /// between various levels of representation including BINARY, TAG_ENCODING,
    /// PURE_IDENTITY and LEGACY formats. An additional output level ONS_HOSTNAME may
    /// be defined for some coding schemes.
    ///</summary>

    public class TDTEngine
    {
        static Dictionary<XmlDefinitions, string> data;
        // 'prefixes' is a list of levels to prefix trees. Each prefix tree is a
        // Trie structure that is useful for quickly finding a
        // matching prefix.
        static List<PrefixMatch> prefixes = new List<PrefixMatch>();

        // The GEPC64Table is a xml that provides a lookup of GS1 Company Prefix based on
        // the corresponding integer-based Company Prefix Index and vice versa.
        static GEPC64Table gEPC64Table;
        // Dictionary 'gs1cpi' is a dictionary providing a lookup between
        // a GS1 Company Prefix and the corresponding integer-based Company Prefix
        // Index, where one of these has been registered for use with 64-bit EPCs.
        static Dictionary<string, string> gs1cpi;
        // Dictionary 'gs1cpirev' is a dictionary providing a lookup between 
        // a integer-based Company Prefix Index and the GS1 Company Prefix
        // , where one of these has been registered for use with 64-bit EPCs.
        static Dictionary<string, string> gs1cpirev;
        // Instance of TDTEngine class
        static readonly TDTEngine Instance = new TDTEngine();
        // Legacy constructor for a new Tag Data Translation engine.
        // The ManagerTranslation.xml file is loaded from a from a directory called OtherXML.
        static TDTEngine()
        {
        }
        /// <summary>
        ///  Returns an instance of TDTEngine class.
        /// </summary>
        public static TDTEngine UniqueInstance
        {
            get { return Instance; }
        }
        /// <summary>
        ///  Method that loads all schemes included on the resources.
        /// </summary>
        public void Initialize()
        {
            List<string> definitionFileNames = new List<string>();
            //create string array of definition names form resources
            foreach (string resourceName in Assembly.GetExecutingAssembly().GetManifestResourceNames())
            {
                // Load only selection of definitions.
                if (resourceName.StartsWith("Nedap.EPC.TDT.Definitions."))
                {
                    definitionFileNames.Add(resourceName);
                }
            }

            // Load TDT definitions form resource files
            foreach (string definitionFileName in definitionFileNames)
            {
                LoadEPCTagDataTranslation(Assembly.GetExecutingAssembly().GetManifestResourceStream(definitionFileName));
            }
            // Load data from ManagerTranslation.xml
            LoadGEPC64Table(Assembly.GetExecutingAssembly().GetManifestResourceStream("Nedap.EPC.TDT.OtherXML.ManagerTranslation.xml"));
            LoadGEPC64TableInReverseOrder(Assembly.GetExecutingAssembly().GetManifestResourceStream("Nedap.EPC.TDT.OtherXML.ManagerTranslation.xml"));
        }
        /// <summary>
        ///  Method that loads only selection of TDT library XML definitions included on the resources.
        /// </summary>
        /// <param name="definitions">Array of selected TDT library XML definitions</param>
        public void Initialize(XmlDefinitions[] definitions)
        {
            LoadDataToDictionary();
            List<XmlDefinitions> definitionsList = new List<XmlDefinitions>(definitions);

            foreach (XmlDefinitions definition in definitions)
            {
                if (definitionsList.Contains(definition))
                {
                    LoadEPCTagDataTranslation(Assembly.GetExecutingAssembly().GetManifestResourceStream(data[definition].ToString()));
                    while (definitionsList.Contains(definition))
                    {
                        definitionsList.Remove(definition);
                    }
                }
            }
            // Load data from ManagerTranslation.xml
            LoadGEPC64Table(Assembly.GetExecutingAssembly().GetManifestResourceStream("Nedap.EPC.TDT.OtherXML.ManagerTranslation.xml"));
            LoadGEPC64TableInReverseOrder(Assembly.GetExecutingAssembly().GetManifestResourceStream("Nedap.EPC.TDT.OtherXML.ManagerTranslation.xml"));
        }
        private void LoadDataToDictionary()
        {
            data = new Dictionary<XmlDefinitions, string>();
            data.Add(XmlDefinitions.GDTI_113, "Nedap.EPC.TDT.Definitions.GDTI-113.xml");
            data.Add(XmlDefinitions.GDTI_96, "Nedap.EPC.TDT.Definitions.GDTI-96.xml");
            data.Add(XmlDefinitions.GIAI_202, "Nedap.EPC.TDT.Definitions.GIAI-202.xml");
            data.Add(XmlDefinitions.GIAI_64, "Nedap.EPC.TDT.Definitions.GIAI-64.xml");
            data.Add(XmlDefinitions.GIAI_96, "Nedap.EPC.TDT.Definitions.GIAI-96.xml");
            data.Add(XmlDefinitions.GID_96, "Nedap.EPC.TDT.Definitions.GID-96.xml");
            data.Add(XmlDefinitions.GRAI_170, "Nedap.EPC.TDT.Definitions.GRAI-170.xml");
            data.Add(XmlDefinitions.GRAI_64, "Nedap.EPC.TDT.Definitions.GRAI-64.xml");
            data.Add(XmlDefinitions.GRAI_96, "Nedap.EPC.TDT.Definitions.GRAI-96.xml");
            data.Add(XmlDefinitions.GSRN_96, "Nedap.EPC.TDT.Definitions.GSRN-96.xml");
            data.Add(XmlDefinitions.SGLN_195, "Nedap.EPC.TDT.Definitions.SGLN-195.xml");
            data.Add(XmlDefinitions.SGLN_64, "Nedap.EPC.TDT.Definitions.SGLN-64.xml");
            data.Add(XmlDefinitions.SGLN_96, "Nedap.EPC.TDT.Definitions.SGLN-96.xml");
            data.Add(XmlDefinitions.SGTIN_198, "Nedap.EPC.TDT.Definitions.SGTIN-198.xml");
            data.Add(XmlDefinitions.SGTIN_64, "Nedap.EPC.TDT.Definitions.SGTIN-64.xml");
            data.Add(XmlDefinitions.SGTIN_96, "Nedap.EPC.TDT.Definitions.SGTIN-96.xml");
            data.Add(XmlDefinitions.SSCC_64, "Nedap.EPC.TDT.Definitions.SSCC-64.xml");
            data.Add(XmlDefinitions.SSCC_96, "Nedap.EPC.TDT.Definitions.SSCC-96.xml");
            data.Add(XmlDefinitions.USDOD_64, "Nedap.EPC.TDT.Definitions.USDOD-64.xml");
            data.Add(XmlDefinitions.USDOD_96, "Nedap.EPC.TDT.Definitions.USDOD-96.xml");

        }
        // Load an xml file from the given Stream objects
        static void LoadEPCTagDataTranslation(Stream xmlFileStream)
        {
            // deserialize in an object
            XmlReader xmlReader = new XmlTextReader(xmlFileStream);
            System.Xml.Serialization.XmlSerializer ser = new System.Xml.Serialization.XmlSerializer(typeof(EpcTagDataTranslation));
            EpcTagDataTranslation epcTagDataTranslation = (EpcTagDataTranslation)ser.Deserialize(xmlReader);

            // build prefix tree
            InitFromTDT(epcTagDataTranslation);
        }
        // Adds prefixes from XML Standard definitions into 'prefixes' list
        static void InitFromTDT(EpcTagDataTranslation epcTagDataTranslation)
        {
            foreach (Scheme scheme in epcTagDataTranslation.scheme)
            {
                foreach (Level level in scheme.level)
                {
                    prefixes.Add(new PrefixMatch(level.prefixMatch, scheme, level));
                }
            }
        }
        // Adds values from ManagerTranslation.xml into 'gs1cpi' dictionary
        // where GS1 Company Prefix is a key and Company Prefix Index is a value
        // of a dictionary object.
        static void LoadGEPC64Table(Stream xmlFileStream)
        {
            gs1cpi = new Dictionary<string, string>();

            // deserialize in an object
            XmlReader xmlReader = new XmlTextReader(xmlFileStream);
            System.Xml.Serialization.XmlSerializer ser = new System.Xml.Serialization.XmlSerializer(typeof(GEPC64Table));
            gEPC64Table = (GEPC64Table)ser.Deserialize(xmlReader);

            foreach (GEPC64TableEntry g in gEPC64Table.entry)
            {
                gs1cpi.Add(g.companyPrefix, g.index);
            }
        }
        //  Adds values from ManagerTranslation.xml into 'gs1cpirev' dictionary
        //  where Company Prefix Index is a key and GS1 Company Prefix  is a value
        //  of a dictionary object.
        static void LoadGEPC64TableInReverseOrder(Stream xmlFileStream)
        {
            gs1cpirev = new Dictionary<string, string>();

            // deserialize in an object
            XmlReader xmlReader = new XmlTextReader(xmlFileStream);
            System.Xml.Serialization.XmlSerializer ser = new System.Xml.Serialization.XmlSerializer(typeof(GEPC64Table));
            gEPC64Table = (GEPC64Table)ser.Deserialize(xmlReader);

            foreach (GEPC64TableEntry g in gEPC64Table.entry)
            {
                gs1cpirev.Add(g.index, g.companyPrefix);
            }
        }

        // For the given input string, and optionally a tag length, find a scheme/level
        // with a matching prefix and tag length.
        static PrefixMatch GetPrefixMatch(string epc, int tagLength)
        {
            StringBuilder prefixBuilder = new StringBuilder();

            List<PrefixMatch> prefixesReduced = new List<PrefixMatch>();
            List<PrefixMatch> prefixesReducedNew;
            prefixesReduced.AddRange(prefixes.ToArray());

            int i = 0;

            // each time check if prefix still matches
            while (prefixesReduced.Count > 1)
            {
                prefixBuilder.Append(epc[i++]);
                prefixesReducedNew = new List<PrefixMatch>();

                // reduce number of prefixMatches
                foreach (PrefixMatch p in prefixesReduced)
                {
                    if (p.Prefix != null)
                    {
                        if (p.Prefix.StartsWith(prefixBuilder.ToString()))
                        {
                            if (tagLength > 0)
                            {
                                if (int.Parse(p.SchemeData.tagLength) == tagLength)
                                {
                                    prefixesReducedNew.Add(p);
                                }
                            }
                            else
                            {
                                prefixesReducedNew.Add(p);
                            }
                        }
                    }
                }
                // prefix(es) found
                prefixesReduced = prefixesReducedNew;
            }

            if (prefixesReduced.Count == 1)
            {
                return prefixesReduced.ToArray()[0];
            }
            else
            {
                throw new TDTTranslationException("TDTSchemeNotFound.");
            }
        }
        // For the given Prefix match, input string, and optionally a tag length,
        // find a corresponding Option key with a matching pattern and grammar.
        static Option GetOption(PrefixMatch prefixMatch, string epcIdentifier, int companyPrefixLength)
        {

            int start = 0;
            int end = 0;
            int total = 0;
            int iterations = 0;

            Option option = new Option();
            int matchCount = 0;
            // if input level have more than one Option key
            if (prefixMatch.LevelData.option.Length > 1)
            {

                start = Environment.TickCount;
                iterations += 1;
                // reduce number of Option keys
                foreach (Option o in prefixMatch.LevelData.option)
                {
                    if (Regex.IsMatch(epcIdentifier, o.pattern))
                    {
                        if (companyPrefixLength > 0)
                        {
                            if (companyPrefixLength == int.Parse(o.optionKey))
                            {
                                option = o;
                                matchCount++;
                            }
                        }
                        else
                        {
                            option = o;
                            matchCount++;
                        }
                    }
                }

                end = Environment.TickCount;
                total = (end - start);

            }
            else if (prefixMatch.LevelData.option.Length == 1)
            {
                matchCount = 1;
                option = prefixMatch.LevelData.option[0];
            }

            if (matchCount == 0)
            {
                throw new TDTTranslationException("TDTOptionNotFound (no option key found).");
            }
            else if (matchCount > 1)
            {
                throw new TDTTranslationException("TDTOptionNotFound (multiple option keys found).");
            }
            return option;
        }
        // Returns an integer value of a given input string or returns an integer value of
        //  a specified fieldname from the specified dictionary if a dictionary contains an input string.
        static int GetIntvalue(Dictionary<string, string> data, string input)
        {
            int output;
            if (Regex.IsMatch(input, "^\\d+$"))
            {
                output = int.Parse(input);
            }
            else
            {
                if (data.ContainsKey(input))
                {
                    output = int.Parse(data[input]);
                }
                else
                {
                    output = -1;
                }
            }
            return output;
        }
        // Calculates the check digit for a supplied input string (assuming that the
        // check digit will be the digit immediately following the supplied input
        // string). GS1 check digit calculation methods are used.
        static string CalculateGS1Checksum(string input)
        {
            int checksum;
            int weight;
            int total = 0;
            int len = input.Length;
            int d;
            for (int i = 0; i < len; i++)
            {
                if (i % 2 == 0)
                {
                    weight = -3;
                }
                else
                {
                    weight = -1;
                }
                d = int.Parse(input.Substring(len - 1 - i, 1));
                total += weight * d;
            }
            checksum = (10 + total % 10) % 10;
            return checksum.ToString();
        }
        // Adds additional entries to the 'data' dictionary by processing various
        // rules defined in the TDT definition files. Typically used for string
        // processing functions, lookup in tables, calculation of check digits etc.
        static void ProcessRules(Dictionary<string, string> data, Rule rule)
        {
            string function = rule.function;
            int openBracket = function.IndexOf("(");
            int closeBracket = function.IndexOf(")");

            string[] functionParameters = function.Substring(openBracket + 1, closeBracket - openBracket - 1).Split(',');
            string functionName = function.Substring(0, openBracket);

            string newFieldName = rule.newFieldName;

            string givenValue;

            switch (functionName)
            {
                // case "TABLELOOKUP:"
                // Stores in the 'data' dictionary the value obtained from a lookup in a specified XML table.
                case "TABLELOOKUP":
                    if (functionParameters.Length != 4)
                    {
                        throw new Exception("Number of parameters is incorrect for rule " + functionName + ".");
                    }
                    // The first parameter is the given value already known
                    givenValue = functionParameters[0];
                    // The second parameter is the string filename of the table
                    string table = functionParameters[1];
                    // The third parameter is the column in which the supplied input value should be sought
                    string inputColumnSupplied = functionParameters[2];
                    // The fourth parameter is the column whose value should be read for the corresponding row,
                    // in order to obtain the result of the lookup.
                    string outputColumnRequired = functionParameters[3];

                    if (table == "tdt64bitcpi")
                    {   // lookup of Company Prefix Index based on the corresponding  GS1 Company Prefix
                        if (inputColumnSupplied == "gs1companyprefix")
                        {
                            data[outputColumnRequired] = gs1cpi[data[inputColumnSupplied]];
                        }
                        // lookup of GS1 Company Prefix based on the corresponding  Company Prefix Index
                        if (inputColumnSupplied == "gs1companyprefixindex")
                        {
                            data[outputColumnRequired] = gs1cpirev[data[inputColumnSupplied]];
                        }
                    }
                    else
                    {
                        throw new NotImplementedException("Rule: " + functionName + " (table: " + table + ").");
                    }

                    break;
                // case "LENGTH:"
                // Stores the length of the specified string under the new fieldname 
                // specified by the corresponding rule of the definition file.
                case "LENGTH":
                    if (functionParameters.Length != 1)
                    {
                        throw new Exception("Number of parameters is incorrect for rule " + functionName + ".");
                    }

                    givenValue = functionParameters[0];
                    //string s = data[givenValue];

                    data[newFieldName] = data[givenValue].Length.ToString();
                    break;
                // case "GS1CHECKSUM:"
                // Stores a GS1 check digit in the 'data' dictionary, keyed under the
                // new fieldname specified by the corresponding rule of the definition file.
                case "GS1CHECKSUM":
                    data[newFieldName] = CalculateGS1Checksum(data[functionParameters[0]]);
                    break;
                // case "SUBSTR:"
                // Obtains a substring of the string provided as the first parameter. If
                // only a single second parameter is specified, then this is considered
                // as the start index and all characters from the start index onwards
                // are stored in the 'data' dictionary under the key named
                // 'newfieldname' in the corresponding rule of the definition file. If a
                // second and third parameter are specified, then the second parameter
                // is the start index and the third is the length of characters
                // required. A substring consisting characters from the start index up
                // to the required length of characters is stored in the 'data' dictionary
                // , keyed under the new fieldname specified by the corresponding
                // rule of the defintion file.
                case "SUBSTR":
                    if (functionParameters.Length == 2)
                    {
                        string currentFieldName = functionParameters[0];
                        int startPoint = GetIntvalue(data, functionParameters[1]);

                        if (startPoint >= 0)
                        {
                            data[newFieldName] = data[currentFieldName].Substring(startPoint);
                        }
                    }
                    else if (functionParameters.Length == 3)
                    {
                        string currentFieldName = functionParameters[0];
                        int startPoint = GetIntvalue(data, functionParameters[1]);
                        int endPoint = GetIntvalue(data, functionParameters[2]);

                        if (startPoint >= 0 && endPoint >= 0)
                        {
                            data[newFieldName] = data[currentFieldName].Substring(startPoint, endPoint);
                        }
                    }
                    break;
                // case "CONCAT:"
                // Concatenates specified string parameters together. Literal values
                // must be enclosed within single or double quotes or consist of
                // unquoted digits. Other unquoted strings are considered as fieldnames
                // and the corresponding value from the 'data' dictionary are
                // inserted. The result of the concatenation (and substitution) of the
                // strings is stored as a new entry in the 'data' dictionary, keyed
                // under the new fieldname specified by the rule.
                case "CONCAT":
                    StringBuilder output = new StringBuilder();
                    foreach (string functionParameter in functionParameters)
                    {
                        if (data.ContainsKey(functionParameter))
                        {
                            output.Append(data[functionParameter]);
                        }
                        else
                        {
                            output.Append(functionParameter);
                        }
                    }
                    data[newFieldName] = output.ToString();
                    break;
                default:
                    throw new NotImplementedException("Rule: " + functionName + ".");
            }
        }
        // Pad a field value according the field definition
        static void PadField(Dictionary<string, string> data, Field field)
        {
            if (field.padChar == null)
                return;

            //RE char padChar = char.Parse(field.padChar);
            char padChar = Convert.ToChar(field.padChar);

            string fieldName = field.name;
            string fieldValue = data[fieldName];

            int requiredLength = 0;
            if (field.length != null)
            {
                requiredLength = int.Parse(field.length);
            }

            if (field.padDirSpecified)
            {
                if (field.padDir == PadDirectionList.RIGHT)
                {
                    fieldValue = fieldValue.PadRight(requiredLength, padChar);
                }
                else
                {
                    fieldValue = fieldValue.PadLeft(requiredLength, padChar);
                }
            }
            else
            {
                fieldValue = fieldValue.PadLeft(requiredLength, padChar);
            }

            data[fieldName] = fieldValue;
        }
        // If the outbound level is BINARY, convert the string field to binary, then
        // pad to the left with the appropriate number of zero bits to reach a
        // number of bits specified by the bitLength attribute of the TDT definition
        // file.
        static void PadFieldBinary(Dictionary<string, string> data, Field field)
        {
            string fieldName = field.name;
            string fieldValue = data[fieldName];

            if (field.bitLength == null || fieldValue == "")
                return;

            int requiredLength = int.Parse(field.bitLength);

            fieldValue = System.Convert.ToString(Int64.Parse(fieldValue), 2);

            if (requiredLength < fieldValue.Length)
            {
                throw new Exception("Required length (" + requiredLength + ") is less than current length (" + fieldValue.Length + ").");
            }

            fieldValue = fieldValue.PadLeft(requiredLength, '0');


            data[fieldName] = fieldValue;
        }
        // Returns a string built using a particular grammar. Single-quotes strings
        // are counted as literal strings, whereas all other strings appearing in
        // the grammar require substitution with the corresponding value from the
        // 'data' dictionary.
        static string BuildGrammar(string grammar, Dictionary<string, string> data)
        {
            StringBuilder output = new StringBuilder();

            string[] grammarSplit = Regex.Split(grammar, "\\s+");
            foreach (string s in grammarSplit)
            {
                if (s.Substring(0, 1) == "'")
                {
                    output.Append(s.Substring(1, s.Length - 2));
                }
                else
                {
                    output.Append(data[s]);
                }
            }

            return output.ToString();
        }
        /// <summary>
        /// The 'Translate' method translates a String input to a specified outbound
        /// level of the same coding scheme. For example, the input string value may
        /// be a tag-encoding URI and the outbound level specified by string
        /// outboundlevel may be BINARY, in which case the return value is a binary
        /// representation expressed as a string. 
        /// Note that this version of the method have input string parameter that contains 
        /// additional parameters which are needed to be provided because they
        /// cannot always be determined from the input value alone and also output level type is in string format.
        /// </summary>
        /// <param name="epcIdentifier">The EPC identifier to be converted</param>
        /// <param name="parameterList">String that contains additional parameters which are needed to be provided because they
        /// cannot always be determined from the input value alone.
        /// Examples include the taglength, companyprefixlength and filter
        ///  values.</param>
        /// <param name="outputFormat">The outbound level required for the ouput. Permitted values
        /// include BINARY, TAG_ENCODING, PURE_IDENTITY, LEGACY and
        /// ONS_HOSTNAME</param>
        /// <returns>The identifier converted to the output level</returns>
        static public string Translate(string epcIdentifier, string parameterList, string outputFormat)
        {
            Dictionary<string, string> data = new Dictionary<string, string>();

            foreach (string parameters in parameterList.Split(';'))
            {
                string[] keyValue = parameters.Split('=');

                if (keyValue.Length == 2)
                {
                    data[keyValue[0]] = keyValue[1];
                }
            }

            //RE LevelTypeList levelOutputType = (LevelTypeList)Enum.Parse(typeof(LevelTypeList), outputFormat);
            LevelTypeList levelOutputType = (LevelTypeList)Enum.Parse(typeof(LevelTypeList), outputFormat, true);

            return Translate(epcIdentifier, data, levelOutputType);
        }
        /// <summary>
        /// The 'Translate' method translates a String input to a specified outbound
        /// level of the same coding scheme. For example, the input string value may
        /// be a tag-encoding URI and the outbound level specified by string
        /// outboundlevel may be BINARY, in which case the return value is a binary
        /// representation expressed as a string. 
        /// Note that this version of the method have no additional parameters. 
        /// </summary>
        /// <param name="epcIdentifier">The EPC identifier to be converted</param>
        /// <param name="levelOutputType">The outbound level required for the ouput. Permitted values
        /// include BINARY, TAG_ENCODING, PURE_IDENTITY, LEGACY and
        /// ONS_HOSTNAME</param>
        /// <returns>The identifier converted to the output level</returns>
        static public string Translate(string epcIdentifier, LevelTypeList levelOutputType)
        {
            return Translate(epcIdentifier, new Dictionary<string, string>(), levelOutputType);
        }
        /// <summary>
        /// The 'Translate' method translates a String input to a specified outbound
        /// level of the same coding scheme. For example, the input string value may
        /// be a tag-encoding URI and the outbound level specified by string
        /// outboundlevel may be BINARY, in which case the return value is a binary
        /// representation expressed as a string.
        /// </summary>
        /// <param name="epcIdentifier">The EPC identifier to be converted</param>
        /// <param name="data"> Dictionary that contains additional parameters which are needed to be provided because they
        /// cannot always be determined from the input value alone.
        /// Examples include the taglength, companyprefixlength and filter
        ///  values.</param>
        /// <param name="levelOutputType">The outbound level required for the ouput. Permitted values
        /// include BINARY, TAG_ENCODING, PURE_IDENTITY, LEGACY and
        /// ONS_HOSTNAME</param>
        /// <returns>The identifier converted to the output level</returns>
        static public string Translate(string epcIdentifier, Dictionary<string, string> data, LevelTypeList levelOutputType)
        {
            // translation according to process defined in the EPCGlobal Tag Data Translation (TDT) 1.4

            // read (and parse) the supplied extra parameters to extract the value and encoding for each key

            int tagLength = 0;
            if (data.ContainsKey("taglength"))
            {
                tagLength = int.Parse(data["taglength"]);
            }

            //=== vanaf hier 78ms

            int companyPrefixLength = 0;
            if (data.ContainsKey("gs1companyprefixlength"))
            {
                companyPrefixLength = int.Parse(data["gs1companyprefixlength"]);
            }

            //=== vanaf hier 75ms

            // detect the input scheme and level within that scheme from the input value
            PrefixMatch prefixMatch = GetPrefixMatch(epcIdentifier, tagLength);

            //=== vanaf hier 69ms

            //=====================================
            int start = 0;
            int end = 0;
            int total = 0;
            start = Environment.TickCount;
            //=====================================

            // determine the appropriate option value matching the input value string
            Option option = GetOption(prefixMatch, epcIdentifier, companyPrefixLength);

            //=====================================
            end = Environment.TickCount;
            total = (end - start);
            //=====================================

            string optionValue = option.optionKey;

            //=== vanaf hier 28ms

            // seperate all the input values
            Match match = Regex.Match(epcIdentifier, option.pattern);
            List<string> values = new List<string>();
            foreach (Group g in match.Groups)
            {
                values.Add(g.Value);
            }

            //=== vanaf hier 17ms

            foreach (Field f in option.field)
            {
                int seq = int.Parse(f.seq);
                string fieldName = f.name;
                string fieldValue = values[seq];

                //Int64 fieldValueCheck = 0;

                if (!f.compactionSpecified)
                {
                    if (f.characterSet == "")
                    {
                        // check if charset matches
                        string pattern = "^" + f.characterSet + "$";
                        if (!Regex.IsMatch(fieldValue, pattern))
                        {
                            throw new TDTTranslationException("TDTUnmatchingCharacterSet (name: " + fieldName + ", value: " + fieldValue + ").");
                        }
                    }

                    // if input type is binary, apply padding, according to schematic on page 35
                    if (prefixMatch.LevelData.type == LevelTypeList.BINARY)
                    {
                        if (f.bitPadDirSpecified)
                        {
                            if (f.bitPadDir == PadDirectionList.LEFT)
                            {
                                fieldValue = fieldValue.TrimStart('0');
                            }
                            else
                            {
                                fieldValue = fieldValue.TrimEnd('0');
                            }
                            if (fieldValue == "")
                            {
                                fieldValue = "0";
                            }
                        }
                        Int64 d = System.Convert.ToInt64(fieldValue, 2);
                        data[fieldName] = d.ToString();

                        Field fNonBinary = new Field();
                        foreach (Level lo in prefixMatch.SchemeData.level)
                        {
                            if (lo.type == LevelTypeList.TAG_ENCODING)
                            {
                                foreach (Option oo in lo.option)
                                {
                                    if (optionValue == oo.optionKey)
                                    {
                                        foreach (Field fo in oo.field)
                                        {
                                            if (fo.name == fieldName)
                                            {
                                                fNonBinary = fo;
                                                break;
                                            }
                                        }
                                    }
                                }
                            }
                        }

                        // determine relevant non-binary output field
                        if (f.padChar != null)
                        {
                            if (fNonBinary.padChar != null)
                            {
                                throw new Exception("Invalid TDT definition file.");
                            }
                            else
                            {
                                if (f.padDirSpecified)
                                {
                                    if (f.padDir == PadDirectionList.LEFT)
                                    {
                                        data[fieldName] = data[fieldName].TrimStart(f.padChar[0]);
                                    }
                                    else
                                    {
                                        data[fieldName] = data[fieldName].TrimEnd(f.padChar[0]);
                                    }
                                }
                            }
                        }
                        else
                        {
                            if (fNonBinary.padChar != null)
                            {
                                if (fNonBinary.padDirSpecified)
                                {
                                    if (fNonBinary.padDir == PadDirectionList.LEFT)
                                    {
                                        data[fieldName] = data[fieldName].PadLeft(int.Parse(fNonBinary.length), fNonBinary.padChar[0]);
                                    }
                                    else
                                    {
                                        data[fieldName] = data[fieldName].PadRight(int.Parse(fNonBinary.length), fNonBinary.padChar[0]);
                                    }
                                }
                            }
                        }
                    }
                    else
                    {
                        data[fieldName] = fieldValue;
                    }

                    // no 128-bit types in C#, may raise unnecessary exception
                    try
                    {
                        if (f.decimalMinimum != "")
                        {
                            Int64 minimum = System.Convert.ToInt64(f.decimalMinimum);
                            if (Int64.Parse(data[fieldName]) < minimum)
                            {
                                throw new Exception("TDTNumericOverflow. Value is lower than minimum (name: " + fieldName + ", value: " + fieldValue + ", minimum: " + minimum.ToString() + ").");
                            }
                        }

                        if (f.decimalMaximum != "")
                        {
                            Int64 maximum = System.Convert.ToInt64(f.decimalMaximum);
                            if (Int64.Parse(data[fieldName]) > maximum)
                            {
                                throw new Exception("TDTNumericOverflow. Value is higher than maximum (name: " + fieldName + ", value: " + fieldValue + ", minimum: " + maximum.ToString() + ").");
                            }
                        }
                    }
                    catch
                    {
                    }
                }
                else
                {
                    //int fraction = 1;
                    int bits = 1;

                    switch (f.compaction)
                    {
                        case CompactionMethodList.Item5bit:
                            //fraction = 32;
                            bits = 5;
                            break;
                        case CompactionMethodList.Item6bit:
                            //fraction = 64;
                            bits = 6;
                            break;
                        case CompactionMethodList.Item7bit:
                            //fraction = 128;
                            bits = 7;
                            break;
                        case CompactionMethodList.Item8bit:
                            //fraction = 256;
                            bits = 8;
                            break;
                        default:
                            throw new Exception("Unsupported compaction method: " + f.compaction + ".");
                    }

                    if (f.bitPadDirSpecified)
                    {
                        while (fieldValue.Length % bits != 0)
                        {
                            if (f.bitPadDir == PadDirectionList.LEFT)
                            {
                                fieldValue = fieldValue.Substring(1);
                            }
                            else
                            {
                                fieldValue = fieldValue.Substring(0, fieldValue.Length - 1);
                            }
                        }
                    }

                    StringBuilder sb = new StringBuilder();
                    while (fieldValue.Length > 0)
                    {
                        int number = System.Convert.ToInt32(fieldValue.Substring(0, bits), 2);
                        fieldValue = fieldValue.Substring(bits);

                        if (number < 32)
                            number += 64;

                        if (f.compaction == CompactionMethodList.Item5bit)
                            number += 64;

                        //RE sb.Append(ASCIIEncoding.ASCII.GetString(new byte[] { (byte)number }));
                        sb.Append(ASCIIEncoding.ASCII.GetString(new byte[] { (byte)number }, 1, 1));
                    }

                    data[fieldName] = sb.ToString();
                }
            }

            //=== vanaf hier 9ms

            // EXTRACT rules
            int seqRules = 0;
            if (prefixMatch.LevelData.rule != null)
            {
                foreach (Rule r in prefixMatch.LevelData.rule)
                {
                    if (r.type == ModeList.EXTRACT)
                    {
                        if (seqRules < int.Parse(r.seq))
                        {
                            ProcessRules(data, r);
                        }
                        else
                        {
                            throw new Exception("TDTRuleOutOfOrder (rule: " + r.ToString() + ").");
                        }
                    }
                }
            }

            // retrieve output level
            Level levelOutput = null;
            foreach (Level l in prefixMatch.SchemeData.level)
            {
                if (l.type == levelOutputType)
                {
                    levelOutput = l;
                }
            }

            // retrieve output option
            Option optionOutput = null;
            foreach (Option o in levelOutput.option)
            {
                if (o.optionKey == optionValue)
                {
                    optionOutput = o;
                }
            }

            // FORMAT rules
            seqRules = 0;
            if (levelOutput.rule != null)
            {
                foreach (Rule r in levelOutput.rule)
                {
                    if (r.type == ModeList.FORMAT)
                    {
                        if (seqRules < int.Parse(r.seq))
                        {
                            //seqRules = int.Parse(r.seq);
                            ProcessRules(data, r);
                        }
                        else
                        {
                            throw new Exception("TDTRuleOutOfOrder (rule: " + r.ToString() + ").");
                        }
                    }
                }
            }

            if (levelOutput.type == LevelTypeList.BINARY)
            {
                foreach (Field f in optionOutput.field)
                {
                    string fieldName = f.name;

                    Field fNonBinary = new Field();
                    foreach (Level lo in prefixMatch.SchemeData.level)
                    {
                        if (lo.type == LevelTypeList.TAG_ENCODING)
                        {
                            foreach (Option oo in lo.option)
                            {
                                if (optionValue == oo.optionKey)
                                {
                                    foreach (Field fo in oo.field)
                                    {
                                        if (fo.name == f.name)
                                        {
                                            fNonBinary = fo;
                                            break;
                                        }
                                    }
                                }
                            }
                        }
                    }

                    if (fNonBinary.padChar != null)
                    {
                        if (f.padChar != null)
                        {
                            throw new Exception("Invalid TDT definition file.");
                        }
                        else
                        {
                            if (fNonBinary.padDirSpecified)
                            {
                                if (fNonBinary.padDir == PadDirectionList.LEFT)
                                {
                                    if (Regex.IsMatch(data[fieldName], "[^1-9A-Za-z]") == false)
                                    {
                                        data[fieldName] = data[fieldName].TrimStart(fNonBinary.padChar[0]);
                                    }

                                }
                                else
                                {
                                    data[fieldName] = data[fieldName].TrimEnd(fNonBinary.padChar[0]);
                                }
                            }
                        }
                    }
                    else
                    {
                        if (f.padChar != null)
                        {
                            if (f.padDirSpecified)
                            {
                                if (f.padDir == PadDirectionList.LEFT)
                                {
                                    data[fieldName] = data[fieldName].PadLeft(int.Parse(f.length), f.padChar[0]);
                                }
                                else
                                {
                                    data[fieldName] = data[fieldName].PadRight(int.Parse(f.length), f.padChar[0]);
                                }
                            }
                        }
                    }

                    if (f.compactionSpecified)
                    {
                        int fraction = 1;
                        int bits = 1;

                        switch (f.compaction)
                        {
                            case CompactionMethodList.Item5bit:
                                fraction = 32;
                                bits = 5;
                                break;
                            case CompactionMethodList.Item6bit:
                                fraction = 64;
                                bits = 6;
                                break;
                            case CompactionMethodList.Item7bit:
                                fraction = 128;
                                bits = 7;
                                break;
                            case CompactionMethodList.Item8bit:
                                fraction = 256;
                                bits = 8;
                                break;
                            default:
                                throw new Exception("Unsupported compaction method: " + f.compaction + ".");
                        }

                        StringBuilder bitsString = new StringBuilder();
                        foreach (byte b in ASCIIEncoding.ASCII.GetBytes(data[fieldName]))
                        {
                            string app = System.Convert.ToString(b % fraction, 2).PadLeft(8, '0').Substring(8 - bits, bits);
                            bitsString.Append(app);

                        }

                        data[fieldName] = bitsString.ToString();
                    }
                    else
                    {
                        data[fieldName] = System.Convert.ToString(Int64.Parse(data[fieldName]), 2);
                    }

                    if (f.bitPadDirSpecified)
                    {
                        if (f.bitPadDir == PadDirectionList.LEFT)
                        {
                            data[fieldName] = data[fieldName].PadLeft(int.Parse(f.bitLength), '0');
                        }
                        else
                        {
                            data[fieldName] = data[fieldName].PadRight(int.Parse(f.bitLength), '0');
                        }
                    }
                }
            }
            //=====================================
            end = Environment.TickCount;
            total = (end - start);
            //=====================================

            return BuildGrammar(optionOutput.grammar, data);
        }
        /// <summary>
        /// Raises exceptions for cases which are not implemented
        /// </summary>
        static public void RefreshTranslations()
        {
            throw new NotImplementedException();
        }
    }
}
