﻿using System;

using System.Collections.Generic;
using System.Text;
using Nedap.HandheldApi.CommunicationQueue;
using InTheHand.Net.Bluetooth;
using InTheHand.Net.Sockets;
using InTheHand.Net;
using System.Net.Sockets;
using System.Threading;
using System.Diagnostics;

namespace Nedap.HandheldApi.CommunicationPort
{
    /// <summary>
    /// This class provides Bluetooth Serial Port functionality implemented with 32feet.net library.
    /// It can be used to automatically discover available !D Hand devices, pair an connect to one
    /// of these, without user interaction.
    /// </summary>
    public class BluetoothCommunicationPort : ICommunicationPort
    {
        private List<BluetoothDeviceInfo> handheldDevices = new List<BluetoothDeviceInfo>();
        private int selectedHandheldIndex;
        private bool HandheldReaderFound { get; set; }
        private BluetoothClient btClient;
        private BluetoothEndPoint ep;
        private NetworkStream deviceStream;
        private Thread dataRecievedNotifier;

        /// <summary>
        /// This event is trigered when during device discovery more than one !D Hand is found.
        /// </summary>
        public event EventHandler<MultipleHandheldsFoundEventArgs> OnMultipleHandheldsFound;

        /// <summary>
        /// Constructor
        /// </summary>
        public BluetoothCommunicationPort()
        {
            StartDeviceDiscovery();
        }

        private void StartDeviceDiscovery()
        {
            BluetoothComponent bco = new BluetoothComponent();
            bco.DiscoverDevicesComplete += HandleDiscoDevicesComplete;
            HandheldReaderFound = false;
            bco.DiscoverDevicesAsync(10, true, true, true, false, null);
        }

        private void HandleDiscoDevicesComplete(object sender, DiscoverDevicesEventArgs e)
        {
            foreach (BluetoothDeviceInfo device in e.Devices)
            {
                if (device.DeviceName.IndexOf("!D Hand",StringComparison.OrdinalIgnoreCase) >= 0)
                {
                    handheldDevices.Add(device);
                }
            }
            if (handheldDevices.Count >= 1)
            {
                SelectHandheldDevice();
                HandheldReaderFound = true;
            }
        }

        private void SelectHandheldDevice()
        {
            //default value (first handheld selected)
            selectedHandheldIndex = 0;

            if (handheldDevices.Count > 1 && OnMultipleHandheldsFound != null)
            {
                //Application should provide to the user a dialog with a list of available handhels,
                //and the user should select apropriate one
                MultipleHandheldsFoundEventArgs rearedsEvenArgs = new MultipleHandheldsFoundEventArgs();
                rearedsEvenArgs.Readers.RearedNames = new List<string>(handheldDevices.Count);
                foreach (BluetoothDeviceInfo deviceInfo in handheldDevices)
                {
                    rearedsEvenArgs.Readers.RearedNames.Add(deviceInfo.DeviceName);
                }
                OnMultipleHandheldsFound(this, rearedsEvenArgs);
                selectedHandheldIndex = rearedsEvenArgs.Readers.selectedReaderIndex;
            }
        }

        #region ICommunicationPort Members

        /// <summary>
        /// Connect to the discovered reader.
        /// </summary>
        /// <returns></returns>
        public bool Connect()
        {
            if (!HandheldReaderFound)
            {
                return false;
            }
            try
            {
                //pair reader
                BluetoothSecurity.PairRequest(handheldDevices[selectedHandheldIndex].DeviceAddress, "0000");
                ep = new BluetoothEndPoint(handheldDevices[selectedHandheldIndex].DeviceAddress, BluetoothService.SerialPort);
                btClient = new BluetoothClient();
                btClient.Connect(ep);
                deviceStream = btClient.GetStream();
                return btClient.Connected;
            }
            catch (Exception ex)
            {
                Debug.WriteLine(ex.Message + "/n" + ex.StackTrace);
                return false;
            }
        }
        /// <summary>
        /// Disconnect from reader.
        /// </summary>
        /// <returns></returns>
        public bool Disconnect()
        {
            btClient.Close();
            btClient.Dispose();
            if (handheldDevices.Count > 0)
            {
                BluetoothSecurity.RemoveDevice(handheldDevices[selectedHandheldIndex].DeviceAddress);
            }
            return true;
        }

        /// <summary>
        /// Retrieves if connection to reader is established
        /// </summary>
        public bool Connected
        {
            get
            {
                return (btClient != null) ? btClient.Connected : false;
            }
        }

        /// <summary>
        /// Writes data to reader.
        /// </summary>
        /// <param name="buffer"></param>
        /// <param name="offset"></param>
        /// <param name="length"></param>
        /// <returns></returns>
        public int WriteData(byte[] buffer, int offset, int length)
        {
            try
            {
                deviceStream.Write(buffer, offset, length);
            }
            catch (Exception ex)
            {
                Debug.WriteLine(ex.Message + "/n" + ex.StackTrace);
                return 0;
            }
            return length;
        }

        /// <summary>
        /// Reads data from reader.
        /// </summary>
        /// <param name="buffer"></param>
        /// <param name="offset"></param>
        /// <param name="length"></param>
        /// <returns></returns>
        public int ReadData(byte[] buffer, int offset, int length)
        {
            try
            {
                int readCount = deviceStream.Read(buffer, offset, length);
                dataRecievedNotifier = new Thread(new ThreadStart(DataRecieved));
                dataRecievedNotifier.Start();
                return readCount;
            }
            catch (Exception ex)
            {
                Debug.WriteLine(ex.Message + "/n" + ex.StackTrace);
                return 0;
            }
        }

        private void DataRecieved()
        {
            if (OnDataReceived != null)
            {
                OnDataReceived(null, EventArgs.Empty);
            }
        }

        /// <summary>
        /// This event is trigered when data are received form reader.
        /// </summary>
        public event EventHandler OnDataReceived;

        #endregion

        #region IDisposable Members

        /// <summary>
        /// Dispose.
        /// </summary>
        public void Dispose()
        {
            if (btClient != null)
            {
                btClient.Close();
                btClient.Dispose();
            }
            if (dataRecievedNotifier != null)
            {
                dataRecievedNotifier.Join(1000);
            }
            if (handheldDevices.Count > 0)
            {
                try
                {
                    BluetoothSecurity.RemoveDevice(handheldDevices[selectedHandheldIndex].DeviceAddress);
                }
                catch
                {
                }
            }
        }

        #endregion
    }
}
