﻿using System;

using System.Collections.Generic;
using System.Text;
using Nedap.HandheldApi.CommunicationQueue;
using System.IO.Ports;

namespace Nedap.HandheldApi.CommunicationPort
{
    /// <summary>
    /// This clas wraps SerialPort class and imelements IComunication interface. It is used as an apstraction for 
    /// a serial port.
    /// </summary>
    public class SerialCommunicationPort : ICommunicationPort
    {
        private SerialPort serialPort;
        private bool disposed = false;
        private bool subscribed = false;
#if DEBUG
        // Create a logger for use in this class
        private static readonly log4net.ILog log = log4net.LogManager.GetLogger(typeof(SerialCommunicationPort));
#endif

        /// <summary>
        /// This is a constructor for serial port. It creates new instance of SerialPort class with required parameters.
        /// </summary>
        /// <param name="portName">String that holds name of the serial port (e.g. COM7)</param>
        /// <param name="baudRate">Integer number that represents port baud rate</param>
        /// <param name="parity">Port parity checking protocol</param>
        /// <param name="dataBits">Standard length of data bits per byte.</param>
        /// <param name="stopBits">Standard number of stopbits per byte.</param>
        public SerialCommunicationPort(string portName, int baudRate, Parity parity, int dataBits, StopBits stopBits)
        {
            serialPort = new SerialPort();
            serialPort.PortName = portName;
            serialPort.BaudRate = baudRate;
            serialPort.Parity = parity;
            serialPort.DataBits = dataBits;
            serialPort.StopBits = stopBits;
        }

        /// <summary>
        /// Destructor.
        /// </summary>
        ~SerialCommunicationPort()
        {
            Dispose();
        }

        #region IDisposable Members

        /// <summary>
        /// Disposes private members.
        /// </summary>
        public void Dispose()
        {
            if (!disposed)
            {
                if (subscribed)
                {
                    serialPort.DataReceived -= new SerialDataReceivedEventHandler(OnDataReceived);
                    subscribed = false;
                }
                serialPort.Dispose();
                disposed = true;
            }
        }
        #endregion


        #region ICommunicationPort Members

        /// <summary>
        /// Connect to port
        /// </summary>
        /// <returns>True if success</returns>
        public bool Connect()
        {
            try
            {
                serialPort.Open();
                if (serialPort.IsOpen)
                {
                    if (!subscribed)
                    {
                        serialPort.DataReceived += new SerialDataReceivedEventHandler(OnDataReceived);
                        subscribed = true;
                    }
                    Connected = true;
                    return true;
                }
            }
            catch (Exception e)
            {
#if DEBUG
                log.Error("Connection error", e); 
#endif
            }
            return false;
        }

        /// <summary>
        /// Disconnect from port
        /// </summary>
        /// <returns>True if success</returns>
        public bool Disconnect()
        {
            try
            {
                if (subscribed)
                {
                    serialPort.DataReceived -= new SerialDataReceivedEventHandler(OnDataReceived);
                    subscribed = false;
                }
                serialPort.Close();
                if (!serialPort.IsOpen)
                {
                    Connected = false;
                    return true;
                }
            }
            catch (Exception e)
            {
#if DEBUG
                log.Error("Connection error", e); 
#endif
            }
            return false;
        }

        /// <summary>
        /// Defines if port is connected
        /// </summary>
        public bool Connected { get; private set; }

        /// <summary>
        /// Writes data to port
        /// </summary>
        /// <param name="buffer">Data buffer</param>
        /// <param name="offset">Position where to start writing</param>
        /// <param name="length">Number of bytes to write</param>
        /// <returns></returns>
        public int WriteData(byte[] buffer, int offset, int length)
        {
            try
            {
                if (serialPort.BytesToWrite == 0)
                {
                    //all bytes are sent so we can continue
                    serialPort.Write(buffer, offset, length);
                    return length;
                }
            }
            catch (Exception e)
            {
#if DEBUG
                log.Error("Write error", e); 
#endif
            }
            return 0;
        }

        /// <summary>
        /// Reads data from port, if no data is available it should return with 0 result.
        /// </summary>
        /// <param name="buffer">Buffer to read into</param>
        /// <param name="offset">Position where to start writing</param>
        /// <param name="length">Maximum number of bytes to read</param>
        /// <returns>Actual read bytes, which is possible to be 0 if no data is available</returns>
        public int ReadData(byte[] buffer, int offset, int length)
        {
            try
            {
                int butesToRead = serialPort.BytesToRead;
                if (butesToRead != 0)
                {
                    return serialPort.Read(buffer, offset, length);
                }
            }
            catch (Exception e)
            {
#if DEBUG
                log.Error("Read error", e); 
#endif
            } 
            return 0;
        }
        /// <summary>
        /// This event is trigered when data are received on the serial port.
        /// </summary>
        public event EventHandler OnDataReceived;
        #endregion
    }
}
