﻿using System;

using System.Collections.Generic;
using System.Text;
using System.Threading;
using System.Diagnostics;

namespace Nedap.HandheldApi.CommunicationQueue
{
    /// <summary>
    /// Data representing a response.
    /// </summary>
    internal class ResponseReceivedEventArgs : EventArgs
    {
        /// <summary>
        /// Property that holds one of available responces.
        /// </summary>
        public Session.Response Response { get; set; }

        /// <summary>
        /// Byte array property that hold received data.
        /// </summary>
        public byte[] Data { get; set; }

        /// <summary>
        /// Constructor.
        /// </summary>
        /// <param name="response"></param>
        /// <param name="data"></param>
        public ResponseReceivedEventArgs(Session.Response response, byte[] data)
        {
            this.Response = response;
            this.Data = data;
        }
    }

    /// <summary>
    /// Defines 2-way communication queue which interacts with communication port
    /// </summary>
    internal class CommunicationQueue : ICommunicationQueue, IDisposable
    {
        private ICommunicationPort port;
        private Thread receiveThread;
        private Thread sendThread;
        private AutoResetEvent receiveEvent;
        private AutoResetEvent sendEvent;
        private byte[] sendBuffer;
        private byte[] receiveBuffer;
        private object sendBufferSync = new object();
        private object receiveBufferSync = new object();
        private bool disposed = false;
#if DEBUG
        // Create a logger for use in this class
        private static readonly log4net.ILog log = log4net.LogManager.GetLogger(typeof(CommunicationQueue));
#endif

        public bool Disposed
        {
            get
            {
                lock (disposedSync)
                {
                    return disposed;
                }
            }
            set
            {
                lock (disposedSync)
                {
                    disposed = value;
                }
            }
        }

        private object disposedSync = new object();
        private const byte StartByte = 0x06;
        private const byte StartByteLong = 0x07;

        /// <summary>
        /// Constructor.
        /// </summary>
        /// <param name="port"></param>
        public CommunicationQueue(ICommunicationPort port)
        {
            this.port = port;
            port.OnDataReceived += new EventHandler(port_OnDataReceived);

            receiveEvent = new AutoResetEvent(false);
            sendEvent = new AutoResetEvent(false);

            sendBuffer = new byte[0];
            receiveBuffer = new byte[0];

            sendThread = new Thread(new ThreadStart(SendThreadExecute));
            receiveThread = new Thread(new ThreadStart(ReceiveThreadExecute));

            sendThread.Start();
            receiveThread.Start();
        }

        void port_OnDataReceived(object sender, EventArgs e)
        {
#if DEBUG
            log.Debug("Serial port data recieved)");
#endif
            receiveEvent.Set();
        }

        /// <summary>
        /// Method for sendind a command.
        /// </summary>
        /// <param name="command">Command to send</param>
        /// <param name="data">Data to send</param>
        public void SendCommand(Session.Command command, byte[] data)
        {
            if (!Disposed)
            {
                lock (sendBufferSync)
                {
                    byte[] buffer = EncodeCommand(command, data);
                    byte[] oldBuffer = sendBuffer;
                    sendBuffer = new byte[sendBuffer.Length + buffer.Length];
                    oldBuffer.CopyTo(sendBuffer, 0);
                    buffer.CopyTo(sendBuffer, oldBuffer.Length);
                    sendEvent.Set();
                }
#if DEBUG
                log.Debug("Sent :" + command.ToString() + ":" + BitConverter.ToString(data));
#endif
            }
        }

        private byte[] EncodeCommand(Session.Command command, byte[] data)
        {
            Debug.Assert(data.Length + 4 < 256, "Data too long.");

            byte[] buffer = new byte[data.Length + 4];
            // start byte	
            buffer[0] = StartByte;
            // length	
            buffer[1] = (byte)(3 + data.Length);
            // commandID
            buffer[2] = (byte)command;
            // data	
            for (int i = 0; i < data.Length; i++)
            {
                buffer[3 + i] = data[i];
            }
            // crc	
            int crc = 0x100;
            for (int j = 1; j < data.Length + 3; j++)
            {
                crc -= buffer[j];
            }
            buffer[3 + data.Length] = (byte)(crc & 0xFF);

            return buffer;
        }

        private void SendThreadExecute()
        {
            try
            {
                while (!Disposed)
                {
                    if (port.Connected)
                    {
                        sendEvent.WaitOne(1000, false);

                        bool done = false;

                        while (!done && !Disposed)
                        {
                            lock (sendBufferSync)
                            {
                                if (sendBuffer.Length > 0)
                                {
                                    int bytesWritten = port.WriteData(sendBuffer, 0, sendBuffer.Length);
                                    if (bytesWritten > 0)
                                    {

                                        byte[] oldBuffer = sendBuffer;
                                        sendBuffer = new byte[oldBuffer.Length - bytesWritten];

                                        for (int i = bytesWritten; i < oldBuffer.Length; i++)
                                            sendBuffer[i - bytesWritten] = oldBuffer[i];

                                        done = sendBuffer.Length == 0;
                                    }
                                    else
                                    {
                                        done = true;
                                    }
                                }
                                else
                                {
                                    done = true;
                                }
                            }
                        }
                    }
                    else
                    {
                        while (!port.Connected && !Disposed)
                        {
                            //Reconnect loop
                            Thread.Sleep(1000);
                            //Due to Bluetooth serial port implemntation we have problems with loosing end of messages from handheld.
                            //Here we are resseting receive buffer in order to remove this beggining of the message to be able to continue.
                            receiveBuffer = new byte[0];
                            port.Connect();
                        }
                    }
                }
            }
            catch (Exception e)
            {
#if DEBUG
                log.Error("Error in send thread", e);
#endif
                throw;
            }
        }

        private void ReceiveThreadExecute()
        {
            try
            {
                int bufferLength = 1024;
                byte[] buffer = new byte[bufferLength];

                while (!Disposed)
                {
                    if (port.Connected)
                    {
                        receiveEvent.WaitOne(1000, false);

                        bool done = false;

                        while (!done && !Disposed)
                        {

                            int bytesRead = port.ReadData(buffer, 0, bufferLength);

                            if (bytesRead > 0)
                            {
#if DEBUG
                                log.Debug("Received (" + bytesRead + ") bytes:" + BitConverter.ToString(buffer, 0, bytesRead));
#endif
                                List<ResponseReceivedEventArgs> responses = new List<ResponseReceivedEventArgs>();

                                lock (receiveBufferSync)
                                {
                                    byte[] oldBuffer = receiveBuffer;
                                    receiveBuffer = new byte[oldBuffer.Length + bytesRead];
                                    oldBuffer.CopyTo(receiveBuffer, 0);

                                    for (int i = 0; i < bytesRead; i++)
                                    {
                                        receiveBuffer[oldBuffer.Length + i] = buffer[i];
                                    }
                                    while (true && receiveBuffer.Length >= 4)
                                    {
                                        ResponseReceivedEventArgs responce = null;
                                        // Try to get a response from buffer
                                        responce = TryParseResponse();
                                        if (responce == null)
                                        {
                                            break;
                                        }
                                        responses.Add(responce);
                                    }
                                }

                                if (responses.Count > 0)
                                {
                                    foreach (ResponseReceivedEventArgs response in responses)
                                    {
#if DEBUG
                                        log.Debug("Received " + response.Response.ToString() + "(" + response.Data.Length + "):" + BitConverter.ToString(response.Data));
#endif
                                        DoResponseReceived(response);
                                    }
                                }
                            }
                            else
                            {
                                done = true;
                            }
                        }
                    }
                    else
                    {
                        Thread.Sleep(1000);
                    }
                }
            }
            catch (Exception e)
            {
#if DEBUG
                log.Error("Error in receive thread", e);
#endif
                throw;
            }
        }

        private ResponseReceivedEventArgs TryParseResponse()
        {
            ResponseReceivedEventArgs result = null;

            #region Ignore bytes from previous message
            int startIndex = 0;
            while ((startIndex < receiveBuffer.Length) && (receiveBuffer[startIndex] != StartByte) && (receiveBuffer[startIndex] != StartByteLong))
            {
                startIndex++;
            }

            if (startIndex > 0)
            {
#if DEBUG
                log.Warn(startIndex + " bytes skipped from the in-buffer");
#endif
                byte[] oldBuffer = receiveBuffer;
                receiveBuffer = new byte[receiveBuffer.Length - startIndex];
                for (int i = startIndex; i < oldBuffer.Length; i++)
                    receiveBuffer[i - startIndex] = oldBuffer[i];
            }
            #endregion

            if (receiveBuffer.Length >= 4)
            {
                // verify start byte	
                Debug.Assert(receiveBuffer[0] == StartByte || receiveBuffer[0] == StartByteLong);
                
                bool parsingLongPacket = receiveBuffer[0] == StartByteLong;
                int length = (parsingLongPacket) ? ((receiveBuffer[1] << 8) | receiveBuffer[2]) + 1 : receiveBuffer[1] + 1;
                int indexOfResponce = (parsingLongPacket) ? 3 : 2;

                if (receiveBuffer.Length >= length)
                {
                    // verify CRC	
                    int crc = 0x00;
                    for (int i = 1; i < length - 1; i++) crc -= receiveBuffer[i];

                    if (receiveBuffer[length - 1] != (crc & 0xFF))
                    {
#if DEBUG
                        log.Error("Incorrect CRC in message!");
#endif
                    }
                    else
                    {
                        // get command ID	
                        Session.Response response = (Session.Response)receiveBuffer[indexOfResponce];
                        // get data	
                        byte[] data = new byte[length - 1 - indexOfResponce - 1];
                        for (int i = 0; i < data.Length; i++)
                        {
                            data[i] = receiveBuffer[i + indexOfResponce + 1];
                        }

                        //create event args
                        result = new ResponseReceivedEventArgs(response, data);
                    }

                    #region Remove parsed bytes
                    int endIndex = length;

                    byte[] oldBuffer2 = receiveBuffer;
                    receiveBuffer = new byte[receiveBuffer.Length - endIndex];
                    for (int i = endIndex; i < oldBuffer2.Length; i++)
                        receiveBuffer[i - endIndex] = oldBuffer2[i];

                    #endregion
                }
            }

            return result;
        }

        private void DoResponseReceived(ResponseReceivedEventArgs args)
        {
            if (OnResponseReceived != null)
            {
                OnResponseReceived(this, args);
            }
        }

        /// <summary>
        /// Event that willl be trigered when responce is received.
        /// </summary>
        public event EventHandler<ResponseReceivedEventArgs> OnResponseReceived;

        #region IDisposable Members

        /// <summary>
        /// Dispose.
        /// </summary>
        public void Dispose()
        {
            Monitor.Enter(disposedSync);
            if (!disposed)
            {
                disposed = true;
                Monitor.Exit(disposedSync);

                port.OnDataReceived -= new EventHandler(port_OnDataReceived);

                sendThread.Join(10000);
                receiveThread.Join(10000);

                sendEvent.Close();
                receiveEvent.Close();
            }
            else
            {
                Monitor.Exit(disposedSync);
            }
        }

        /// <summary>
        /// Destructor.
        /// </summary>
        ~CommunicationQueue()
        {
            Dispose();
        }

        #endregion
    }
}
