﻿using System;
using System.Threading;
using System.ComponentModel;
using Nedap.HandheldApi.CommunicationQueue;

namespace Nedap.HandheldApi
{
    /// <summary>
    /// Defines connection state with timed state update and device wakeup with timeout.
    /// </summary>
    internal class ConnectionState : IConnectionState, IDisposable, INotifyPropertyChanged
    {
        #region Properties

        #region Private properties

        private ICommunicationQueue communicationQueue;
        private static Timer timer;
        private bool commandSent = false;
        private bool isDisposed = false;
        private object lockMe = new object();
        private int startDelay = 2000;                     // 2 second (in miliseconds)

        private DateTime lastAnswer = DateTime.MinValue;   // time of the last good device response

        #endregion

        #region Public event handlers

        /// <summary>
        /// Triggers when device connects.
        /// </summary>
        public event EventHandler<EventArgs> DeviceConnected;

        /// <summary>
        /// Triggers when device disconnects.
        /// </summary>
        public event EventHandler<EventArgs> DeviceDisconnected;

        /// <summary>
        /// Triggers when device connection state changes
        /// </summary>
        public event PropertyChangedEventHandler PropertyChanged;

        #endregion

        #region IsConnected
        private bool isConnected;
        /// <summary>
        /// Is device connected ?
        /// </summary>
        internal bool IsConnected
        {
            get { return isConnected; }
            private set
            {
                if (value != isConnected)
                {
                    isConnected = value;
                    OnConnectionStateChanged("IsConnected");

                    if (value) DeviceConnected(this, new EventArgs());
                    else DeviceDisconnected(this, new EventArgs());
                }
            }
        }
        #endregion

        #region MaxWithNoAnswer
        private int maxWithNoAnswer = 5000;
        /// <summary>
        /// Maximum number of seconds to pass without answer from device (in miliseconds).
        /// Default value is 5 seconds (5000 miliseconds).
        /// </summary>
        public int MaxWithNoAnswer
        {
            get { return maxWithNoAnswer; }
            set { maxWithNoAnswer = value; }
        }
        #endregion

        #region CheckInterval
        private int checkInterval = 5000;
        /// <summary>
        /// Device connection status checking interval (in miliseconds).
        /// Default value is 5 seconds (5000 miliseconds).
        /// </summary>
        public int CheckInterval
        {
            get { return checkInterval; }
            set { checkInterval = value; }
        }
        #endregion

        #endregion

        #region ConnectionState() - constructor/destructor

        public ConnectionState(ICommunicationQueue CommunicationQueue)
        {
            communicationQueue = CommunicationQueue;
            communicationQueue.OnResponseReceived += new EventHandler<ResponseReceivedEventArgs>(communicationQueue_OnResponseReceived);

            timer = new Timer(WorkerMethod, null, startDelay, checkInterval);
        }

        ~ConnectionState()
        {
            Dispose();
        }

        #endregion

        #region Methods

        #region communicationQueue_OnResponseReceived(object sender, ResponseReceivedEventArgs e)
        private void communicationQueue_OnResponseReceived(object sender, ResponseReceivedEventArgs e)
        {
            lock (lockMe)
            {
                if (!isDisposed)
                {
                    lastAnswer = DateTime.Now;
                    IsConnected = true;
                    commandSent = false;
                }
            }
        }
        #endregion

        #region WorkerMethod(object state)
        private void WorkerMethod(object state)
        {
            lock (lockMe)
            {
                if (!isDisposed)
                {
                    if (commandSent)
                        IsConnected = false;

                    if (DateTime.Now > lastAnswer.AddMilliseconds(maxWithNoAnswer))
                    {
                        SendCommandToDevice();
                    }
                }
            }
        }
        #endregion

        #region SendCommandToDevice()
        private void SendCommandToDevice()
        {
            commandSent = true;
            communicationQueue.SendCommand(Session.Command.GeneralGetBatteryPercentage, new byte[0]);
        }
        #endregion

        #endregion


        #region IDisposable Members

        public void Dispose()
        {
            isDisposed = true;
            timer.Dispose();
            communicationQueue.OnResponseReceived -= new EventHandler<ResponseReceivedEventArgs>(communicationQueue_OnResponseReceived);
        }

        #endregion

        #region IConnectionState Members
        /// <summary>
        /// Check device connection state.
        /// </summary>
        /// <returns>Returns boolean that represents device connection state (connected/disconnected).</returns>
        public bool CheckConnection()
        {
            return IsConnected;
        }

        #endregion

        #region INotifyPropertyChanged Members

        private void OnConnectionStateChanged(string info)
        {
            if (PropertyChanged != null)
                PropertyChanged(this, new PropertyChangedEventArgs(info));
        }

        #endregion
    }
}