﻿using System;
using System.Collections.Generic;
using System.Text;
using Nedap.HandheldApi;

namespace Nedap.HandheldApi
{
    /// <summary>
    /// Interface to presentation actions.
    /// This methods are called from presentation context when state transition occurs.
    /// </summary>
    internal interface IPresentationActions
    {
        /// <summary>
        /// Executes read of NFC card.
        /// </summary>
        void ActionNfcTryToReadCard();

        /// <summary>
        /// Executes sending of EPC m parameter value to the handheld.
        /// </summary>
        /// <param name="theMValue"></param>
        void ActionRfidSetMValue(RfidEpcMValue theMValue);
        /// <summary>
        /// Executes sending of EPC q parameter value to the handheld.
        /// </summary>
        /// <param name="theQValue"></param>
        /// <param name="dynamic"></param>
        void ActionRfidSetQValue(int theQValue, bool dynamic);
        /// <summary>
        /// Executes sending of EPC session parameter value to the handheld.
        /// </summary>
        /// <param name="theSession"></param>
        void ActionRfidSetSession(RfidEpcSession theSession);
        /// <summary>
        /// Executes sending of EPC target parameter value to the handheld.
        /// </summary>
        /// <param name="theTarget"></param>
        /// <param name="toggle"></param>
        void ActionRfidSetTarget(RfidEpcTarget theTarget, bool toggle);
        /// <summary>
        /// Sets RFID on time parameter.
        /// </summary>
        /// <param name="timeout"></param>
        void ActionRfidSetTimeout(int timeout);
        /// <summary>
        /// Executes read of EPC values.
        /// </summary>
        void ActionRfidReadEpc();
        /// <summary>
        /// Executes write of EPC values.
        /// </summary>
        void ActionRfidWriteEpc();
        /// <summary>
        /// Executes read of RFID data.
        /// </summary>
        /// <param name="power"></param>
        void ActionRfidReadDetailedTagInformation(int power);

        /// <summary>
        /// Sets last read rfid value as current value. 
        /// </summary>
        void ActionRFIDSetSelectLastRead();
        /// <summary>
        /// Sets last written rfid value as current value.
        /// </summary>
        void ActionRFIDSetSelectLastWritten();
        /// <summary>
        /// Resets rfid current select value to null.
        /// </summary>
        void ActionRFIDResetSelect();

        /// <summary>
        /// Executes read of barcode value.
        /// </summary>
        void ActionBarcodeRead();

        /// <summary>
        /// Executes start or stop of UI led blinking on handheld.
        /// Should UI led blinking start or stop depends on given blinking parameter.
        /// </summary>
        /// <param name="blinking"></param>
        void ActionUILedSetBlinking(bool blinking);
        /// <summary>
        /// Executes play of given tune on handheld.
        /// </summary>
        /// <param name="beepTune"></param>
        void ActionUITune(UITune beepTune);

        /// <summary>
        /// This method is executed when device connects to the handheld.
        /// Triggers OnConnect event and start battery percentage timer.
        /// </summary>
        void ActionHandheldConnected();
        /// <summary>
        /// This method is executed when device disconnects from the handheld.
        /// Triggers OnDisconnect event and stops battery percentage timer.
        /// </summary>
        void ActionHandheldDisconnected();
        /// <summary>
        /// Executes configure of the handheld. Sets handheld region to Europe.
        /// </summary>
        void ActionHandheldConfigure();
        /// <summary>
        /// Acquires battery state (voltage) from the handheld.
        /// </summary>
        void ActionHandheldGetBatteryPercentage();

        /// <summary>
        /// Logs given error using debug logging.
        /// </summary>
        /// <param name="errorName"></param>
        void ActionLogError(String errorName);
        /// <summary>
        /// Logs given event using debug logging.
        /// </summary>
        /// <param name="eventName"></param>
        void ActionLogEvent(String eventName);
        /// <summary>
        /// Logs given timed event using debug logging.
        /// </summary>
        /// <param name="eventName"></param>
        void ActionLogTimedEvent(String eventName);
        /// <summary>
        /// Logs end of the given timed event using debug logging.
        /// </summary>
        /// <param name="eventName"></param>
        void ActionLogEndTimedEvent(String eventName);
        /// <summary>
        /// Logs end of the given current timed event using debug logging.
        /// </summary>
        void ActionLogEndCurrentTimedEvent();

        /// <summary>
        /// Starts timer for given presentation timer method. Created timer is added to the timers list.
        /// </summary>
        /// <param name="timer"></param>
        /// <param name="duration"></param>
        void StartTimer(Presentation.PresentationTimerMethod timer, int duration);
        /// <summary>
        /// Stops timer for given presentation timer method. Stoped timer is removed from the timers list.
        /// </summary>
        /// <param name="timer"></param>
        void StopTimer(Presentation.PresentationTimerMethod timer);
        /// <summary>
        /// Changes time when event for given presentation timer method will be triggered. 
        /// </summary>
        /// <param name="timer"></param>
        /// <param name="duration"></param>
        void ResetTimer(Presentation.PresentationTimerMethod timer, int duration);
        /// <summary>
        /// This method simulates button push on the handheld reader.
        /// </summary>
        void SimulateButtonPush();
    }
}
