﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Threading;
using Nedap.EPC.TDT;
using Nedap.HandheldApi.CommunicationQueue;
using System.Collections.ObjectModel;

namespace Nedap.HandheldApi
{
    internal sealed class Presentation : IPresentationActions, IDisposable
    {
        #region Fields
        private const String INVENTORY_PREFIX = "MainMap.RFIDInventory";
        private const String READ_DETAILED_TAG_INFORMATION_PREFIX = "MainMap.RFIDReadDetailedTagInformation";

        private PresentationContext presentationContext;

        private Select rfidCurrentSelect;
        private Password rfidCurrentPassword;
        private EpcCode rfidEPCToWrite;

        private Collection<BarcodeType> barcodeCurrentTypes;
        private int barcodeCurrentDuration;
        private bool barcodeTypesInitialized;

        private RfidEpcMemoryBank rfidReadDataMemoryBank;
        private int rfidReadDataAddressInWords;
        private int rfidReadDataLengthInWords;
        private int readPower;

        private int rfidOffTime;
        private int rfidOnTime;

        private Collection<EpcCode> inventoryTags;
        private Dictionary<String, Collection<EpcCode>> inventoryTagsDictionary;
        private Collection<EpcObservation> singleTags;
        private Collection<EpcObservation> singleTagsFiltered;

        private SyncTimer batteryPercentageTimer;

        private ISession session;

        private Object presentationSync;
        private bool disposed;

        private Dictionary<PresentationTimerMethod, SyncTimer> timers;

        private String currentTimedEvent;
        private bool changeSession;
        private int currentSession = 0;

#if DEBUG
        // Create a logger for use in this class
        private static readonly log4net.ILog log = log4net.LogManager.GetLogger(typeof(Presentation)); 
#endif

        #endregion

        #region Events

        /// <summary>
        /// This event is triggered when inventory list is cleared (All items from the inventory list are removed).
        /// </summary>
        public event EventHandler<EventArgs> OnClearInventory;
        /// <summary>
        /// This event is triggered when battery percentage value is received from the handheld.
        /// </summary>
        public event EventHandler<ReceiveBatteryPercentageEventArgs> OnReceiveBatteryPercentage;
        /// <summary>
        /// This event is triggered when low battery warning is received from the handheld.
        /// </summary>
        public event EventHandler<EventArgs> OnLowBatteryWarning;
        /// <summary>
        /// This event is triggered when empty battery status is received from the handheld.
        /// </summary>
        public event EventHandler<EventArgs> OnEmptyBatteryStatus;
        /// <summary>
        /// This event is triggered when charging status is received from the handheld.
        /// </summary>
        public event EventHandler<EventArgs> OnChargingStatus;
        /// <summary>
        /// This event is triggered when one RFID tag is red using the handheld.
        /// </summary>
        public event EventHandler<ReadRfidTagEventArgs> OnReadRFIDTag;
        /// <summary>
        /// This event is triggered when more than one RFID tags are red using the handheld.
        /// </summary>
        public event EventHandler<ReadRfidTagsEventArgs> OnReadRFIDTags;
        /// <summary>
        /// This event is triggered when more than one new RFID tags are red using the handheld.
        /// </summary>
        public event EventHandler<ReadRfidTagsFilteredEventArgs> OnReadRFIDTagsFiltered;
        /// <summary>
        /// This event is triggered when detailed tag information is red using the handheld.
        /// </summary>
        public event EventHandler<ReadDetailedTagInformationEventArgs> OnReadDetailedTagInformation;
        /// <summary>
        /// This event is triggered when barcode is red using the handheld.
        /// </summary>
        public event EventHandler<ReadBarcodeEventArgs> OnReadBarcode;
        /// <summary>
        /// This event is triggered when the handheld was not able to read barcode.
        /// </summary>
        public event EventHandler<EventArgs> OnNotReadBarcode;
        /// <summary>
        /// This event is triggered when button on the handheld is pushed.
        /// </summary>
        public event EventHandler<EventArgs> OnPushButton;
        /// <summary>
        /// This event is triggered when handheld is shaken.
        /// </summary>
        public event EventHandler<EventArgs> OnShake;
        /// <summary>
        /// This event is triggered when NFC card is red using the handheld.
        /// </summary>
        public event EventHandler<ReadNfcCardEventArgs> OnReadNFCCard;
        /// <summary>
        /// This event is triggered when the handheld is disconnected and action is executed on the device.
        /// </summary>
        public event EventHandler<EventArgs> OnIsDisconnected;
        /// <summary>
        /// This event is triggered when single read of RFID is finished.
        /// </summary>
        public event EventHandler<EventArgs> OnFinishSingleRead;
        /// <summary>
        /// This event is triggered when when inventory process or monitor process are continued.
        /// </summary>
        public event EventHandler<EventArgs> OnContinue;
        /// <summary>
        /// This event is triggered when when inventory process or monitor process are paused.
        /// </summary>
        public event EventHandler<EventArgs> OnPause;
        /// <summary>
        /// This event is triggered when writing of EPC has failed with multiple read.
        /// </summary>
        public event EventHandler<EventArgs> OnFailWriteEpcMultipleRead;
        /// <summary>
        /// This event is triggered when writing of EPC has failed with zero read.
        /// </summary>
        public event EventHandler<EventArgs> OnFailWriteEpcZeroRead;
        /// <summary>
        /// This event is triggered when writing of EPC has failed. 
        /// Failure happened during the write of data.
        /// </summary>
        public event EventHandler<EventArgs> OnFailWriteEpcWriteFailed;
        /// <summary>
        /// This event is triggered when writing of EPC has failed. 
        /// Failure happened during the verification of written of data.
        /// </summary>
        public event EventHandler<EventArgs> OnFailWriteEpcVerificationFailed;
        /// <summary>
        /// This event is triggered when writing of EPC has finished successfully.
        /// </summary>
        public event EventHandler<EventArgs> OnSucceedWriteEPC;
        /// <summary>
        /// This event is triggered when device is connected to the handheld.
        /// </summary>
        public event EventHandler<EventArgs> OnConnect;
        /// <summary>
        /// This event is triggered when device is disconnected from the handheld.
        /// </summary>
        public event EventHandler<EventArgs> OnDisconnect;
        /// <summary>
        /// This event is triggered when supported regions are received from the handheld.
        /// </summary>
        public event EventHandler<ReceiveSupportedRegionsEventArgs> OnReceiveSupportedRegions;

        #endregion

        #region Event triggers

        /// <summary>
        /// Triggers event which indicates that button on the handheld is pushed.
        /// </summary>
        public void DoPushButton()
        {
            if (OnPushButton != null)
            {
                OnPushButton(this, EventArgs.Empty);
            }
        }

        /// <summary>
        /// Triggers event which indicates that the handheld is disconnected and action is executed on the device.
        /// </summary>
        public void DoIsDisconnected()
        {
            if (OnIsDisconnected != null)
            {
                OnIsDisconnected(this, EventArgs.Empty);
            }
        }

        /// <summary>
        /// Triggers event which indicates that read of RFID is finished.
        /// </summary>
        public void DoFinishSingleRead()
        {
            if (OnFinishSingleRead != null)
            {
                OnFinishSingleRead(this, EventArgs.Empty);
            }
        }

        /// <summary>
        /// Triggers event which indicates thate inventory process or monitor process are continued.
        /// </summary>
        public void DoContinue()
        {
            if (OnContinue != null)
            {
                OnContinue(this, EventArgs.Empty);
            }
        }

        /// <summary>
        /// Triggers event which indicates thate inventory process or monitor process are paused.
        /// </summary>
        public void DoPause()
        {
            if (OnPause != null)
            {
                OnPause(this, EventArgs.Empty);
            }
        }

        /// <summary>
        /// Triggers event which indicates that writing of EPC has failed with multiple read.
        /// </summary>
        public void DoFailWriteEpcMultipleRead()
        {
            if (OnFailWriteEpcMultipleRead != null)
            {
                OnFailWriteEpcMultipleRead(this, new EventArgs());
            }
        }

        /// <summary>
        /// Triggers event which indicates that writing of EPC has failed with zero read.
        /// </summary>
        public void DoFailWriteEpcZeroRead()
        {
            if (OnFailWriteEpcZeroRead != null)
            {
                OnFailWriteEpcZeroRead(this, new EventArgs());
            }
        }

        /// <summary>
        /// Triggers event which indicates that writing of EPC has failed.
        /// Failure happened during the write of data.
        /// </summary>
        public void DoFailWriteEpcWriteFailed()
        {
            if (OnFailWriteEpcWriteFailed != null)
            {
                OnFailWriteEpcWriteFailed(this, new EventArgs());
            }
        }

        /// <summary>
        /// Triggers event which indicates that writing of EPC has failed.
        /// Failure happened during the verification of written of data.
        /// </summary>
        public void DoFailWriteEpcVerificationFailed()
        {
            if (OnFailWriteEpcVerificationFailed != null)
            {
                OnFailWriteEpcVerificationFailed(this, new EventArgs());
            }
        }

        /// <summary>
        /// Triggers event which indicates that writing of EPC has finished successfully.
        /// </summary>
        public void DoSucceedWriteEPC()
        {
            if (OnSucceedWriteEPC != null)
            {
                OnSucceedWriteEPC(this, EventArgs.Empty);
            }
        }

        /// <summary>
        /// Triggers event which indicates that device is connected to the handheld.
        /// </summary>
        public void DoConnect()
        {
            if (OnConnect != null)
            {
                OnConnect(this, EventArgs.Empty);
            }
        }

        /// <summary>
        /// Triggers event which indicates that device is disconnected from the handheld.
        /// </summary>
        public void DoDisconnect()
        {
            if (OnDisconnect != null)
            {
                OnDisconnect(this, EventArgs.Empty);
            }
        }

        /// <summary>
        /// Triggers event which indicates that supported regions are received from the handheld.
        /// </summary>
        /// <param name="regionFlags"></param>
        public void DoReceiveSupportedRegions(int regionFlags)
        {
            if (OnReceiveSupportedRegions != null)
            {
                OnReceiveSupportedRegions(this, new ReceiveSupportedRegionsEventArgs(regionFlags));
            }
        }

        /// <summary>
        /// Triggers event which indicates that inventory list is cleared (All items from the inventory list are removed).
        /// </summary>
        public void DoClearInventoryEvent()
        {
            if (OnClearInventory != null)
            {
                OnClearInventory(this, EventArgs.Empty);
            }
        }

        #endregion

        public enum PresentationTimerMethod
        {
            NfcReadErrorTimer,
            NfcReadRetry,
            RfidSingleReadReadErrorTimer,
            RfidInventoryReadErrorTimer,
            RfidInventoryIdleTimer,
            RfidMonitorReadErrorTimer,
            RfidMonitorIdleTimer
        }

        #region Constructor/destructor

        /// <summary>
        /// Constructor
        /// </summary>
        /// <param name="session"></param>
        public Presentation(ISession session)
        {
            this.session = session;
            presentationSync = new Object();
            SubscribeToSessionEvents();
            presentationContext = new PresentationContext(this);

            inventoryTagsDictionary = new Dictionary<string, Collection<EpcCode>>();
            inventoryTags = new Collection<EpcCode>();
            singleTags = new Collection<EpcObservation>();
            singleTagsFiltered = new Collection<EpcObservation>();

            timers = new Dictionary<PresentationTimerMethod, SyncTimer>();

            rfidOffTime = 0;
            rfidOnTime = 500;
            BeeperEnabled = true;
            VibrationEnabled = true;

            barcodeTypesInitialized = false;
        }

        /// <summary>
        /// Destructor
        /// </summary>
        ~Presentation()
        {
            Dispose();
        }
        #endregion

        #region Session events subscribe/unsubscribe
        /// <summary>
        /// Subscribing to events form session.
        /// </summary>
        private void SubscribeToSessionEvents()
        {
            session.OnSessionHandheldConnect += new EventHandler<EventArgs>(SessionHandheldConnect);
            session.OnSessionHandheldDisconnect += new EventHandler<EventArgs>(SessionHandheldDisconnect);
            session.OnSessionHandheldReceiveBatteryPercentage += new EventHandler<HandheldReceiveBatteryPercentageEventArgs>(SessionHandheldReceiveBatteryPercentage);
            session.OnSessionHandheldReceiveLowBatteryWarning += new EventHandler<EventArgs>(SessionHandheldReceiveLowBatteryWarning);
            session.OnSessionHandheldReceiveEmptyBatteryStatus += new EventHandler<EventArgs>(SessionHandheldReceiveEmptyBatteryStatus);
            session.OnSessionHandheldReceiveChargingStatus += new EventHandler<EventArgs>(SessionHandheldReceiveChargingStatus);
            session.OnSessionHandheldReceiveError += new EventHandler<HandheldReceiveErrorEventArgs>(SessionHandheldReceiveError);
            session.OnSessionRfidReadTag += new EventHandler<RfidReadTagEventArgs>(SessionRfidReadTag);
            session.OnSessionRfidReadData += new EventHandler<RfidReadDataEventArgs>(SessionRfidReadData);
            session.OnSessionRfidFinishReadCycle += new EventHandler<EventArgs>(SessionRfidFinishDownloadTags);
            session.OnSessionRfidFailButTagsAvailable += new EventHandler<RfidFailButTagsAvailableEventArgs>(SessionRfidFailButTagsAvailable);
            session.OnSessionRfidWriteDataMoreTagsAvailable += new EventHandler<RfidWriteDataMoreTagsAvailableEventArgs>(SessionRfidWriteDataMoreTagsAvailable);
            session.OnSessionRfidWriteNoTagsInField += new EventHandler<EventArgs>(SessionRfidWriteNoTagsInField);
            session.OnSessionRfidWriteTooMuchTagsInField += new EventHandler<EventArgs>(SessionRfidWriteTooMuchTagsInField);
            session.OnSessionRfidWriteToTagFailed += new EventHandler<EventArgs>(SessionRfidWriteToTagFailed);
            session.OnSessionRfidWriteSuccess += new EventHandler<EventArgs>(SessionRfidWriteSuccessful);
            session.OnSessionRfidReceiveSupportedRegions += new EventHandler<RfidReceiveSupportedRegionsEventArgs>(SessionRfidReceiveSupportedRegions);
            session.OnSessionBarcodeRead += new EventHandler<RfidReadABarcodeEventArgs>(SessionBarcodeReadBarcode);
            session.OnSessionBarcodeNotRead += new EventHandler<EventArgs>(SessionBarcodeFailReadingBarcode);
            session.OnSessionUIPushButton += new EventHandler<EventArgs>(SessionUIPushButton);
            session.OnSessionUIShake += new EventHandler<EventArgs>(SessionUIShake);
            session.OnSessionNfcCardNotFound += new EventHandler<EventArgs>(SessionNfcNotFoundACard);
            session.OnSessionNfcCardRead += new EventHandler<RfidReadCardEventArgs>(SessionNfcReadCard);
        }

        /// <summary>
        /// Unsubscribing from session events.
        /// </summary>
        private void UnsubscribefromSessionEvents()
        {
            session.OnSessionHandheldConnect -= new EventHandler<EventArgs>(SessionHandheldConnect);
            session.OnSessionHandheldDisconnect -= new EventHandler<EventArgs>(SessionHandheldDisconnect);
            session.OnSessionHandheldReceiveBatteryPercentage -= new EventHandler<HandheldReceiveBatteryPercentageEventArgs>(SessionHandheldReceiveBatteryPercentage);
            session.OnSessionHandheldReceiveLowBatteryWarning -= new EventHandler<EventArgs>(SessionHandheldReceiveLowBatteryWarning);
            session.OnSessionHandheldReceiveEmptyBatteryStatus -= new EventHandler<EventArgs>(SessionHandheldReceiveEmptyBatteryStatus);
            session.OnSessionHandheldReceiveChargingStatus -= new EventHandler<EventArgs>(SessionHandheldReceiveChargingStatus);
            session.OnSessionHandheldReceiveError -= new EventHandler<HandheldReceiveErrorEventArgs>(SessionHandheldReceiveError);
            session.OnSessionRfidReadTag -= new EventHandler<RfidReadTagEventArgs>(SessionRfidReadTag);
            session.OnSessionRfidReadData -= new EventHandler<RfidReadDataEventArgs>(SessionRfidReadData);
            session.OnSessionRfidFinishReadCycle -= new EventHandler<EventArgs>(SessionRfidFinishDownloadTags);
            session.OnSessionRfidFailButTagsAvailable -= new EventHandler<RfidFailButTagsAvailableEventArgs>(SessionRfidFailButTagsAvailable);
            session.OnSessionRfidWriteDataMoreTagsAvailable -= new EventHandler<RfidWriteDataMoreTagsAvailableEventArgs>(SessionRfidWriteDataMoreTagsAvailable);
            session.OnSessionRfidReceiveSupportedRegions -= new EventHandler<RfidReceiveSupportedRegionsEventArgs>(SessionRfidReceiveSupportedRegions);
            session.OnSessionBarcodeRead -= new EventHandler<RfidReadABarcodeEventArgs>(SessionBarcodeReadBarcode);
            session.OnSessionBarcodeNotRead -= new EventHandler<EventArgs>(SessionBarcodeFailReadingBarcode);
            session.OnSessionUIPushButton -= new EventHandler<EventArgs>(SessionUIPushButton);
            session.OnSessionUIShake -= new EventHandler<EventArgs>(SessionUIShake);
            session.OnSessionNfcCardNotFound -= new EventHandler<EventArgs>(SessionNfcNotFoundACard);
            session.OnSessionNfcCardRead -= new EventHandler<RfidReadCardEventArgs>(SessionNfcReadCard);
        }
        #endregion

        #region INHPresentationActions Members

        /// <summary>
        /// Executes read of NFC card.
        /// </summary>
        public void ActionNfcTryToReadCard()
        {
            lock (presentationSync)
            {
                if (!disposed)
                {
                    session.NfcTryToReadCard();
                }
            }
        }

        /// <summary>
        /// Executes sending of EPC m parameter value to the handheld.
        /// </summary>
        /// <param name="theMValue"></param>
        public void ActionRfidSetMValue(RfidEpcMValue theMValue)
        {
            lock (presentationSync)
            {
                if (!disposed)
                {
                    session.RfidSetMValue(theMValue);
                }
            }
        }

        /// <summary>
        /// Executes sending of EPC q parameter value to the handheld.
        /// </summary>
        /// <param name="theQValue"></param>
        /// <param name="dynamic"></param>
        public void ActionRfidSetQValue(int theQValue, bool dynamic)
        {
            lock (presentationSync)
            {
                if (!disposed)
                {
                    session.RfidSetQValue(theQValue, dynamic);
                }
            }
        }

        /// <summary>
        /// Executes sending of EPC session parameter value to the handheld.
        /// </summary>
        /// <param name="theSession"></param>
        public void ActionRfidSetSession(RfidEpcSession theSession)
        {
            lock (presentationSync)
            {
                if (!disposed)
                {
                    currentSession = (int)theSession;
                    session.RfidSetSession(theSession);
                }
            }
        }

        /// <summary>
        /// Change session in the reader to support new reading of high-volume of tags.
        /// In high volume reading tags respond each 15 minutes, which is reset only by moving to session 3.
        /// This reset happens only if switch is needed.
        /// </summary>
        public void ActionRfidSwitchSession()
        {
            lock (presentationSync)
            {
                // switch session
                if (changeSession)
                {
                    if (currentSession == 2)
                        ActionRfidSetSession(RfidEpcSession.Session3);
                    else if (currentSession == 3)
                        ActionRfidSetSession(RfidEpcSession.Session2);

                    changeSession = false;
                }
            }
        }

        /// <summary>
        /// Executes sending of EPC target parameter value to the handheld.
        /// </summary>
        /// <param name="theTarget"></param>
        /// <param name="toggle"></param>
        public void ActionRfidSetTarget(RfidEpcTarget theTarget, bool toggle)
        {
            lock (presentationSync)
            {
                if (!disposed)
                {
                    session.RfidSetTarget(theTarget, toggle);
                }
            }
        }

        /// <summary>
        /// Sets RFID on time parameter.
        /// </summary>
        /// <param name="timeout"></param>
        public void ActionRfidSetTimeout(int timeout)
        {
            lock (presentationSync)
            {
                if (!disposed)
                {
                    rfidOnTime = timeout;
                }
            }
        }

        /// <summary>
        /// Executes read of EPC values.
        /// </summary>
        public void ActionRfidReadEpc()
        {
            lock (presentationSync)
            {
                if (!disposed)
                {
                    singleTagsFiltered.Clear();
                    singleTags.Clear();
                    session.RfidReadEpcs(rfidOnTime, readPower, rfidCurrentSelect);
                }
            }
        }

        /// <summary>
        /// Executes write of EPC values.
        /// </summary>
        public void ActionRfidWriteEpc()
        {
            lock (presentationSync)
            {
                if (!disposed)
                {
                    session.RfidWriteEpc(rfidEPCToWrite);
                }
            }
        }

        /// <summary>
        /// Executes read of RFID data.
        /// </summary>
        /// <param name="power"></param>
        public void ActionRfidReadDetailedTagInformation(int power)
        {
            lock (presentationSync)
            {
                if (!disposed)
                {
                    session.RfidReadData(rfidReadDataMemoryBank, rfidReadDataAddressInWords, rfidReadDataLengthInWords, rfidOnTime, power, rfidCurrentSelect, rfidCurrentPassword);
                }
            }
        }

        /// <summary>
        /// Sets power for Rfid reading
        /// </summary>
        /// <param name="power"></param>
        public void ActionRfidSetReadPower(int power)
        {
            this.readPower = power;
        }

        /// <summary>
        /// Sets last read rfid value as current rfidCurrentSelect value. 
        /// </summary>
        public void ActionRFIDSetSelectLastRead()
        {
            EpcCode epcCode = singleTags.Count > 0 ? singleTags[0].EpcCode : null;
            rfidCurrentSelect = new Select(epcCode);
        }
        /// <summary>
        /// Sets last written rfid value as current rfidCurrentSelect value. 
        /// </summary>
        public void ActionRFIDSetSelectLastWritten()
        {
            rfidCurrentSelect = new Select(rfidEPCToWrite);
        }
        /// <summary>
        /// Resets rfidCurrentSelect value (Sets rfidCurrentSelect value to null). 
        /// </summary>
        public void ActionRFIDResetSelect()
        {
            rfidCurrentSelect = null;
        }

         /// <summary>
        /// Executes read of barcode value.
        /// </summary>
        public void ActionBarcodeRead()
        {
            lock (presentationSync)
            {
                if (!disposed)
                {
                    session.BarcodeTryDecode(barcodeCurrentDuration);
                }
            }
        }

        /// <summary>
        /// Executes start or stop of UI led blinking on handheld.
        /// Should UI led blinking start or stop depends on given blinking parameter.
        /// </summary>
        /// <param name="blinking"></param>
        public void ActionUILedSetBlinking(bool blinking)
        {
            lock (presentationSync)
            {
                if (!disposed)
                {
                    session.UILedSetBlinking(blinking);
                }
            }
        }

        /// <summary>
        /// Plays short tune followed woth a vibration
        /// </summary>
        private void ActionUITuneShort()
        {
            lock (presentationSync)
            {
                if (!disposed)
                {
                    if (BeeperEnabled)
                    {
                        session.UIPlay(UITune.Short);
                    }
                    if (VibrationEnabled)
                    {
                        session.UIVibrateTune(UITune.Short);
                    }
                }
            }
        }

        /// <summary>
        /// Executes play of given tune on handheld.
        /// </summary>
        /// <param name="beepTune"></param>
        public void ActionUITune(UITune beepTune)
        {
            lock (presentationSync)
            {
                if (!disposed)
                {
                    if (BeeperEnabled)
                    {
                        session.UIPlay(beepTune);
                    }
                    if (VibrationEnabled)
                    {
                        session.UIVibrateTune(beepTune);
                    }
                }
            }
        }

        /// <summary>
        /// This method is executed when device connects to the handheld.
        /// Triggers OnConnect event and start battery percentage timer.
        /// </summary>
        public void ActionHandheldConnected()
        {
            lock (presentationSync)
            {
                if (!disposed)
                {
                    DoConnect();
                    batteryPercentageTimer = new SyncTimer(OnBatteryPercentageTimer, null, 0, 60000);
                }
            }
        }

        /// <summary>
        /// This method is executed when device disconnects from the handheld.
        /// Triggers OnDisconnect event and stops battery percentage timer.
        /// </summary>
        public void ActionHandheldDisconnected()
        {
            lock (presentationSync)
            {
                if (!disposed)
                {
                    if (batteryPercentageTimer != null)
                    {
                        batteryPercentageTimer.Dispose();
                    }
                    DoDisconnect();
                }
            }
        }

        /// <summary>
        /// Executes configure of the handheld. Sets handheld region to Europe.
        /// </summary>
        public void ActionHandheldConfigure()
        {
            lock (presentationSync)
            {
                if (!disposed)
                {
                    session.RfidSetRegion(RfidRegion.Europe);
                }
            }
        }

        /// <summary>
        /// Acquires battery state (voltage) from the handheld.
        /// </summary>
        public void ActionHandheldGetBatteryPercentage()
        {
            lock (presentationSync)
            {
                if (!disposed)
                {
                    session.HandheldGetBatteryPercentage();
                }
            }
        }

        /// <summary>
        /// Logs given error using debug logging.
        /// </summary>
        /// <param name="errorName"></param>
        public void ActionLogError(String errorName)
        {
            lock (presentationSync)
            {
                if (!disposed)
                {
#if DEBUG
                    log.Error(errorName); 
#endif
                }
            }
        }

        /// <summary>
        /// Logs given event using debug logging.
        /// </summary>
        /// <param name="eventName"></param>
        public void ActionLogEvent(String eventName)
        {
            lock (presentationSync)
            {
                if (!disposed)
                {
#if DEBUG
                    log.Info("LogEvent: " + eventName); 
#endif
                }
            }
        }

        /// <summary>
        /// Logs given timed event using debug logging.
        /// </summary>
        /// <param name="eventName"></param>
        public void ActionLogTimedEvent(String eventName)
        {
            lock (presentationSync)
            {
                if (!disposed)
                {
#if DEBUG
                    log.Info("LogTimedEvent: " + eventName); 
#endif
                    currentTimedEvent = eventName;
                }
            }
        }

        /// <summary>
        /// Logs end of the given timed event using debug logging.
        /// </summary>
        /// <param name="eventName"></param>
        public void ActionLogEndTimedEvent(String eventName)
        {
            lock (presentationSync)
            {
                if (!disposed)
                {
#if DEBUG
                    log.Info("LogEndTimedEvent: " + eventName); 
#endif
                }
            }
        }

        /// <summary>
        /// Logs end of the given current timed event using debug logging.
        /// </summary>
        public void ActionLogEndCurrentTimedEvent()
        {
            lock (presentationSync)
            {
                if (!disposed)
                {
                    if (currentTimedEvent != null)
                    {
                        ActionLogEndTimedEvent(currentTimedEvent);
                        currentTimedEvent = null;
                    }
                }
            }
        }

        /// <summary>
        /// Starts timer for given presentation timer method. Created timer is added to the timers list.
        /// </summary>
        /// <param name="timerMethod"></param>
        /// <param name="duration"></param>
        public void StartTimer(PresentationTimerMethod timerMethod, int duration)
        {
            lock (presentationSync)
            {
                if (!disposed)
                {
                    timers.Add(timerMethod, new SyncTimer(GetTimerCallback(timerMethod), null, duration, duration));
                }
            }
        }

        /// <summary>
        /// Stops timer for given presentation timer method. Stoped timer is removed from the timers list.
        /// </summary>
        /// <param name="timerMethod"></param>
        public void StopTimer(PresentationTimerMethod timerMethod)
        {
            lock (presentationSync)
            {
                if (!disposed)
                {
                    SyncTimer t = timers[timerMethod];
                    timers.Remove(timerMethod);
                    t.Dispose();
                }
            }
        }

        /// <summary>
        /// Changes time when event for given presentation timer method will be triggered. 
        /// </summary>
        /// <param name="timerMethod"></param>
        /// <param name="duration"></param>
        public void ResetTimer(PresentationTimerMethod timerMethod, int duration)
        {
            lock (presentationSync)
            {
                if (!disposed)
                {
                    SyncTimer t = timers[timerMethod];
                    t.Change(0, duration);
                }
            }
        }

        /// <summary>
        /// This method simulates button push on the handheld reader.
        /// </summary>
        public void SimulateButtonPush()
        {
            SessionUIPushButton(this, EventArgs.Empty);
        }

        private TimerCallback GetTimerCallback(PresentationTimerMethod timerMethod)
        {
            switch (timerMethod)
            {
                case PresentationTimerMethod.NfcReadErrorTimer:
                    return new TimerCallback(OnNfcReadErrorTimer);
                case PresentationTimerMethod.NfcReadRetry:
                    return new TimerCallback(OnNfcReadRetry);
                case PresentationTimerMethod.RfidSingleReadReadErrorTimer:
                    return new TimerCallback(OnRfidSingleReadReadErrorTimer);
                case PresentationTimerMethod.RfidInventoryReadErrorTimer:
                    return new TimerCallback(OnRfidInventoryReadErrorTimer);
                case PresentationTimerMethod.RfidInventoryIdleTimer:
                    return new TimerCallback(OnRfidInventoryIdleTimer);
                case PresentationTimerMethod.RfidMonitorReadErrorTimer:
                    return new TimerCallback(OnRfidMonitorReadErrorTimer);
                case PresentationTimerMethod.RfidMonitorIdleTimer:
                    return new TimerCallback(OnRfidMonitorIdleTimer);
                default:
                    throw new NotSupportedException("Not supported timer method:" + timerMethod);
            }
        }

        private void OnNfcReadErrorTimer(Object stateInfo)
        {
            lock (presentationSync)
            {
                if (!disposed)
                {
                    presentationContext.nfcReadErrorTimer();
                }
            }
        }

        private void OnNfcReadRetry(Object stateInfo)
        {
            lock (presentationSync)
            {
                if (!disposed)
                {
                    presentationContext.nfcReadRetry();
                }
            }
        }

        private void OnRfidSingleReadReadErrorTimer(Object stateInfo)
        {
            lock (presentationSync)
            {
                if (!disposed)
                {
                    presentationContext.rfidSingleReadReadErrorTimer();
                }
            }
        }

        private void OnRfidInventoryReadErrorTimer(Object stateInfo)
        {
            lock (presentationSync)
            {
                if (!disposed)
                {
                    presentationContext.rfidInventoryReadErrorTimer();
                }
            }
        }

        private void OnRfidInventoryIdleTimer(Object stateInfo)
        {
            lock (presentationSync)
            {
                if (!disposed)
                {
                    presentationContext.rfidInventoryIdleTimer();
                }
            }
        }

        private void OnRfidMonitorReadErrorTimer(Object stateInfo)
        {
            lock (presentationSync)
            {
                if (!disposed)
                {
                    presentationContext.rfidMonitorReadErrorTimer();
                }
            }
        }

        private void OnRfidMonitorIdleTimer(Object stateInfo)
        {
            lock (presentationSync)
            {
                if (!disposed)
                {
                    presentationContext.rfidMonitorIdleTimer();
                }
            }
        }

        #endregion

        #region Session event handlers

        private void SessionHandheldConnect(object sender, EventArgs e)
        {
            lock (presentationSync)
            {
                if (!disposed)
                {
                    presentationContext.connected();
                }
            }
        }

        private void SessionHandheldDisconnect(object sender, EventArgs e)
        {
            lock (presentationSync)
            {
                if (!disposed)
                {
                    presentationContext.disconnected();
                }
            }
        }

        private void SessionHandheldReceiveBatteryPercentage(object sender, HandheldReceiveBatteryPercentageEventArgs e)
        {
            lock (presentationSync)
            {
                if (!disposed)
                {
                    int percentage = e.Percentage;
                    if (OnReceiveBatteryPercentage != null)
                    {
                        OnReceiveBatteryPercentage(this, new ReceiveBatteryPercentageEventArgs(percentage));
                    }
                }
            }
        }

        private void SessionHandheldReceiveLowBatteryWarning(object sender, EventArgs e)
        {
            lock (presentationSync)
            {
                if (!disposed)
                {
                    if (OnLowBatteryWarning != null)
                    {
                        OnLowBatteryWarning(this, EventArgs.Empty);
                    }
                }
            }
        }

        private void SessionHandheldReceiveEmptyBatteryStatus(object sender, EventArgs e)
        {
            lock (presentationSync)
            {
                if (!disposed)
                {
                    if (OnEmptyBatteryStatus != null)
                    {
                        OnEmptyBatteryStatus(this, EventArgs.Empty);
                    }
                }
            }
        }

        private void SessionHandheldReceiveChargingStatus(object sender, EventArgs e)
        {
            lock (presentationSync)
            {
                if (!disposed)
                {
                    if (OnChargingStatus != null)
                    {
                        OnChargingStatus(this, EventArgs.Empty);
                    }
                }
            }
        }

        private void SessionHandheldReceiveError(object sender, HandheldReceiveErrorEventArgs e)
        {
            lock (presentationSync)
            {
                if (!disposed)
                {
                    ActionLogError(e.ErrorMessage);
                }
            }
        }

        private void SessionRfidReadTag(object sender, RfidReadTagEventArgs e)
        {
            lock (presentationSync)
            {
                if (!disposed)
                {
                    if (!singleTags.Contains(e.EpcObservation))
                    {
                        singleTags.Add(e.EpcObservation);
                        if (!inventoryTags.Contains(e.EpcObservation.EpcCode))
                        {
                            singleTagsFiltered.Add(e.EpcObservation);

                            inventoryTags.Add(e.EpcObservation.EpcCode);

                            EpcObservation epc = new EpcObservation(e.EpcObservation.EpcCode, e.EpcObservation.TimeStamp, e.EpcObservation.RSSI,e.EpcObservation.ReadCount);
                            if (OnReadRFIDTag != null)
                            {
                                OnReadRFIDTag(this, new ReadRfidTagEventArgs(epc));
                            }
                        }
                    }
                }
            }
        }

        private void SessionRfidFinishDownloadTags(object sender, EventArgs e)
        {
            lock (presentationSync)
            {
                if (!disposed)
                {
                    presentationContext.rfidReadFinished();

                    if (singleTagsFiltered.Count > 0)
                    {
                        if (presentationContext.State.Name.StartsWith(INVENTORY_PREFIX, StringComparison.Ordinal))
                        {
                            ActionUITuneShort();
                        }

                        Collection<EpcObservation> filteredTagsCollectiton = new Collection<EpcObservation>();
                        foreach (EpcObservation epcObservation in singleTagsFiltered)
                        {
                            filteredTagsCollectiton.Add(new EpcObservation(epcObservation.EpcCode, epcObservation.TimeStamp, epcObservation.RSSI, epcObservation.ReadCount));
                        }
                        if (OnReadRFIDTagsFiltered != null)
                        {
                            OnReadRFIDTagsFiltered(this, new ReadRfidTagsFilteredEventArgs(filteredTagsCollectiton));
                        }
                    }

                    if (singleTags.Count == 0)
                    {
                        presentationContext.rfidReadFinishedZero();
                    }
                    if (singleTags.Count > 0)
                    {
                        presentationContext.rfidReadFinishedMoreThanZero();
                    }
                    Collection<EpcObservation> tags = new Collection<EpcObservation>();
                    foreach (EpcObservation tag in singleTags)
                    {
                        tags.Add(tag);
                    }
                    if (OnReadRFIDTags != null)
                    {
                        OnReadRFIDTags(this, new ReadRfidTagsEventArgs(tags));
                    }
                }
            }
        }

        private void SessionRfidReadData(object sender, RfidReadDataEventArgs e)
        {
            lock (presentationSync)
            {
                if (!disposed)
                {
                    if (presentationContext.State.Name.StartsWith(READ_DETAILED_TAG_INFORMATION_PREFIX, StringComparison.Ordinal))
                    {
                        int dataInteger = e.Data[3];
                        dataInteger |= e.Data[2] << 8;
                        dataInteger |= e.Data[1] << 16;
                        dataInteger |= e.Data[0] << 24;

                        RfidEpcMdid tagMDID = (RfidEpcMdid)((dataInteger & 0x7FF000) >> 12);
                        int tagModelNumber = dataInteger & 0xFFF;
                        bool tagXTID = (dataInteger & 0x00800000) > 0;

                        if (OnReadDetailedTagInformation != null)
                        {
                            OnReadDetailedTagInformation(this, new ReadDetailedTagInformationEventArgs(tagMDID, tagModelNumber, tagXTID));
                        }
                    }

                    if (e.MoreTagsAvailable)
                    {
                        presentationContext.rfidActionSuccessMoreTags();
                    }
                    else
                    {
                        presentationContext.rfidActionSuccess();
                    }
                }
            }
        }

        private void SessionRfidFailButTagsAvailable(object sender, RfidFailButTagsAvailableEventArgs e)
        {
            lock (presentationSync)
            {
                if (!disposed)
                {
                    if (e.Available)
                    {
                        presentationContext.rfidActionFailMoreTags();
                    }
                    else
                    {
                        presentationContext.rfidActionFail();
                    }
                }
            }
        }

        private void SessionRfidWriteDataMoreTagsAvailable(object sender, RfidWriteDataMoreTagsAvailableEventArgs e)
        {
            lock (presentationSync)
            {
                if (!disposed)
                {
                    if (e.Available)
                    {
                        presentationContext.rfidActionSuccessMoreTags();
                    }
                    else
                    {
                        presentationContext.rfidActionSuccess();
                    }
                }
            }
        }


        void SessionRfidWriteNoTagsInField(object sender, EventArgs e)
        {
            lock (presentationSync)
            {
                if (!disposed)
                {
                    presentationContext.rfidWriteEPCNoTagsInField();
                }
            }
        }

        void SessionRfidWriteTooMuchTagsInField(object sender, EventArgs e)
        {
            lock (presentationSync)
            {
                if (!disposed)
                {
                    presentationContext.rfidWriteEPCTooMuchTagsInField();
                }
            }
        }

        void SessionRfidWriteToTagFailed(object sender, EventArgs e)
        {
            lock (presentationSync)
            {
                if (!disposed)
                {
                    presentationContext.rfidWriteEPCWriteToTagFailed();
                }
            }
        }

        void SessionRfidWriteSuccessful(object sender, EventArgs e)
        {
            lock (presentationSync)
            {
                if (!disposed)
                {
                    presentationContext.rfidWriteEPCWriteSuccessful();
                }
            }
        }

        private void SessionRfidReceiveSupportedRegions(object sender, RfidReceiveSupportedRegionsEventArgs e)
        {
            lock (presentationSync)
            {
                if (!disposed)
                {
                    DoReceiveSupportedRegions(e.RegionFlags);
                }
            }
        }

        private void SessionBarcodeReadBarcode(object sender, RfidReadABarcodeEventArgs e)
        {
            lock (presentationSync)
            {
                if (!disposed)
                {
                    presentationContext.barcodeReadSuccess();
                    string barcode = string.Copy(e.Barcode);
                    if (OnReadBarcode != null)
                    {
                        OnReadBarcode(this, new ReadBarcodeEventArgs(barcode, e.BarcodeType));
                    }
                }
            }
        }

        private void SessionBarcodeFailReadingBarcode(object sender, EventArgs e)
        {
            lock (presentationSync)
            {
                if (!disposed)
                {
                    presentationContext.barcodeReadFail();
                    if (OnNotReadBarcode != null)
                    {
                        OnNotReadBarcode(this, EventArgs.Empty);
                    }
                }
            }
        }

        private void SessionUIPushButton(object sender, EventArgs e)
        {
            lock (presentationSync)
            {
                if (!disposed)
                {
                    presentationContext.uiButtonPushed();
                    if (OnPushButton != null)
                    {
                        OnPushButton(this, EventArgs.Empty);
                    }
                }
            }
        }

        private void SessionUIShake(object sender, EventArgs e)
        {
            lock (presentationSync)
            {
                if (!disposed)
                {
                    if (OnShake != null)
                    {
                        OnShake(this, EventArgs.Empty);
                    }
                }
            }
        }

        private void SessionNfcNotFoundACard(object sender, EventArgs e)
        {
            lock (presentationSync)
            {
                if (!disposed)
                {
                    presentationContext.nfcNoCardFound();
                }
            }
        }

        private void SessionNfcReadCard(object sender, RfidReadCardEventArgs e)
        {
            lock (presentationSync)
            {
                if (!disposed)
                {
                    presentationContext.nfcCardFound();
                    int cardNumber = e.CardNumber;
                    if (OnReadNFCCard != null)
                    {
                        OnReadNFCCard(this, new ReadNfcCardEventArgs(cardNumber));
                    }
                }
            }
        }

        #endregion

        #region PresentationFacade Members

        /// <summary>
        /// This method can be used to try to read NFC cards. It’ll enable the ﬁeld, and continuously try to read cards in the ﬁeld.
        /// </summary>
        public void NfcReadStart()
        {
            lock (presentationSync)
            {
                if (!disposed)
                {
                    presentationContext.nfcReadStart();
                }
            }
        }


        /// <summary>
        /// The inventory methods provide an easy way to do inventory. It automatically filters ‘multiple’ reads in a buffer and generates a ‘beep’ on new tags. 
        /// EPC air interface parameters are set automatically to session 1, target A (no switching) and automatic Q.
        /// </summary>
        /// <param name="bufferName"></param>
        /// <param name="select"></param>
        /// <param name="highVolume"></param>
        public void RfidInventoryStart(String bufferName, Select select, bool highVolume)
        {
            lock (presentationSync)
            {
                if (!disposed)
                {
                    localRfidSetBuffer(bufferName);
                    changeSession = false;
                    rfidCurrentSelect = select;
                    presentationContext.rfidInventoryStart(highVolume);
                }
            }
        }

        /// <summary>
        /// Resets the RFID inventory 'seen' list of the "default" buffer. 
        /// Please make sure that if a list is also kept in the application layer, clean it after the relevant notification is received.
        /// </summary>
        public void RfidInventoryReset()
        {
            lock (presentationSync)
            {
                if (!disposed)
                {
                    changeSession = true;
                    inventoryTags.Clear();
                    singleTagsFiltered.Clear();
                    DoClearInventoryEvent();
                }
            }
        }        

        /// <summary>
        /// Resets the RFID inventory 'seen' list of the given buffer. 
        /// Please make sure that if a list is also kept in the application layer, clean it after the relevant notification is received.
        /// </summary>
        /// <param name="bufferName"></param>
        public void RfidInventoryReset(String bufferName)
        {
            lock (presentationSync)
            {
                if (!disposed)
                {
                    changeSession = true;
                    inventoryTagsDictionary[bufferName].Clear();
                    DoClearInventoryEvent();
                }
            }
        }

        /// <summary>
        /// Monitors the tags in the field, with a select statement.
        /// </summary>
        /// <param name="select">Select statement</param>
        public void RfidMonitorStart(Select select)
        {
            lock (presentationSync)
            {
                if (!disposed)
                {
                    rfidCurrentSelect = select;
                    presentationContext.rfidMonitorStart();
                }
            }
        }

        /// <summary>
        /// Does a single read cycle of tags, with a select statement.
        /// </summary>
        /// <param name="select">Select statement</param>
        public void RfidReadSingle(Select select)
        {
            lock (presentationSync)
            {
                if (!disposed)
                {
                    rfidCurrentSelect = select;
                    presentationContext.rfidSingleReadStart();
                }
            }
        }

        /// <summary>
        /// To obtain the tag manufacturer ID (MDID), the tag model number, or the tag’s extended identifier field (XTID), use this method.
        /// </summary>
        /// <param name="code">Code</param>
        public void RfidReadTag(EpcCode code)
        {
            lock (presentationSync)
            {
                if (!disposed)
                {
                    if (code != null)
                    {
                        rfidCurrentSelect = new Select(code);
                    }
                    else
                    {
                        rfidCurrentSelect = null;
                    }
                    rfidCurrentPassword = null;
                    rfidReadDataMemoryBank = RfidEpcMemoryBank.Tid;
                    rfidReadDataAddressInWords = 0;
                    rfidReadDataLengthInWords = 2;

                    presentationContext.rfidReadDetailedTagInformationStart();
                }
            }
        }

        /// <summary>
        /// Removed from iOS code.
        /// </summary>
        /// <param name="newEpcCode"></param>
        public void RfidWriteEpc(EpcCode newEpcCode)
        {
            lock (presentationSync)
            {
                if (!disposed)
                {
#if DEBUG
                    log.Info("w1"); 
#endif
                    rfidEPCToWrite = newEpcCode;
                    rfidCurrentSelect = null;
                    rfidCurrentPassword = null;

#if DEBUG
                    log.Info("w2"); 
#endif
                    presentationContext.rfidWriteEPCStart();
#if DEBUG
                    log.Info("epc: " + rfidEPCToWrite.HexString); 
#endif
                }
            }
        }

        /// <summary>
        /// Check to see if the handheld reader has a barcode reader on-board.
        /// </summary>
        /// <returns>If true, there is a barcode reader</returns>
        public bool BarcodeHasReader()
        {
            return session.BarcodeHasReader();
        }

        /// <summary>
        /// Read a barcode with a specified array of types, and with a specified timeout.
        /// </summary>
        /// <param name="barcodeTypes">Array of barcode types</param>
        /// <param name="duration">Timeout in seconds</param>
        public void BarcodeRead(Collection<BarcodeType> barcodeTypes, int duration)
        {
            lock (presentationSync)
            {
                if (!disposed)
                {
                    barcodeCurrentTypes = barcodeTypes;
                    barcodeCurrentDuration = duration;

                    //scegar This is a bugfix for continous barcode reading, since it is preventing continous barcode type setting
                    //messages that can be overlapped with button press, and cause an error. Once the types are set, the handheld does
                    //not need these messages for each scan
                    if (!barcodeTypesInitialized)
                    {
                        session.BarcodeSetStatusForType(BarcodeType.All, false);
                        foreach (BarcodeType barcodeType in barcodeCurrentTypes)
                        {
                            session.BarcodeSetStatusForType(barcodeType, true);
                        }
                        barcodeTypesInitialized = true;
                    }
                    
                    presentationContext.barcodeReadStart();
                }
            }
        }

        /// <summary>
        /// Let the beeper on the handheld play a tune.
        /// </summary>
        /// <param name="tune">Type of tune</param>
        public void UIPlay(UITune tune)
        {
            lock (presentationSync)
            {
                if (!disposed)
                {
                    session.UIPlay(tune);
                }
            }
        }

        /// <summary>
        /// Let the vibration motor on the handheld vibrate.
        /// </summary>
        public void UIVibrate()
        {
            lock (presentationSync)
            {
                if (!disposed)
                {
                    session.UIVibrate();
                }
            }
        }

        /// <summary>
        /// Let the vibration motor on the handheld vibrate with beeper playing a tune.
        /// </summary>
        public void UIVibrateTune(UITune tune)
        {
            lock (presentationSync)
            {
                if (!disposed)
                {
                    session.UIVibrateTune(tune);
                }
            }
        }

        /// <summary>
        /// Enables or disables the logging functionality with Flurry.com. The information logged to Flurry is anonymous. 
        /// We’ll log the usages and performance of the !d hand (e.g. how long does an inventory take, how often will the user press the button, how many tags are read in an hour, 
        /// how is the battery quality over time, etc.). 
        /// This will help Nedap to improve the code, and gather information on the performance and usage of the !d hand in the field. 
        /// We kindly ask you to leave this enabled (which is the default). If you’d prefer to turn this off, feel free to do so.
        /// </summary>
        /// <param name="enabled">Enabled</param>
        public void HandheldLogEnabled(bool enabled)
        {
            lock (presentationSync)
            {
                if (!disposed)
                {
                    throw new NotImplementedException();
                    //[FlurryAPI setEventLoggingEnabled:enabled];
                }
            }
        }

        /// <summary>
        /// Stop the current operation, e.g. inventory or monitoring. Can be used when switching from one view to another.
        /// </summary>
        public void HandheldStop()
        {
            lock (presentationSync)
            {
                if (!disposed)
                {
                    presentationContext.handheldStop();
                }
            }
        }

        /// <summary>
        /// Checks if the handheld is connected or not.
        /// </summary>
        /// <returns></returns>
        public bool HandheldIsConnected()
        {
            return session.HandheldIsConnected();
        }

        /// <summary>
        /// Will let the handheld kill the bluetooth connection. Can be used to pair the handheld with another device.
        /// </summary>
        public void HandheldDisconnect()
        {
            session.HandheldDisconnect();
        }

        /// <summary>
        /// Get information on the handheld. 
        /// </summary>
        /// <returns>DeviceInfo class object with three parameters: serial number, firmware and hardware version.</returns>
        public DeviceInfo HandheldGetInformation()
        {
            return session.HandheldGetInformation();
        }

        /// <summary>
        /// Returns the current battery status in percentage.
        /// </summary>
        public void HandheldGetBatteryPercentage()
        {
            session.HandheldGetBatteryPercentage();
        }

        #endregion

        #region Properties

        /// <summary>
        /// RFIDOffTime property
        /// </summary>
        public int RFIDOffTime
        {
            get
            {
                return rfidOffTime;
            }
        }

        /// <summary>
        /// Returns if beeper is enabled.
        /// </summary>
        public bool BeeperEnabled {get; private set;}
        
        /// <summary>
        /// Returns if vibration is enabled
        /// </summary>
        public bool VibrationEnabled { get; private set; }

        #endregion

        private void OnBatteryPercentageTimer(Object stateInfo)
        {
            lock (presentationSync)
            {
                if (!disposed)
                {
                    ActionHandheldGetBatteryPercentage();
                }
            }
        }

        private void localRfidSetBuffer(String bufferName)
        {
            if (inventoryTagsDictionary.ContainsKey(bufferName))
            {
                inventoryTags = inventoryTagsDictionary[bufferName];
            }
            else
            {
                inventoryTags = new Collection<EpcCode>();
                singleTagsFiltered = new Collection<EpcObservation>();
                inventoryTagsDictionary.Add(bufferName, inventoryTags);
            }
        }

        #region IDisposable Members

        public void Dispose()
        {
            lock (presentationSync)
            {
                if (!disposed)
                {
                    disposed = true;
                    UnsubscribefromSessionEvents();
                    if (batteryPercentageTimer != null)
                    {
                        batteryPercentageTimer.Dispose();
                    }

                    foreach (PresentationTimerMethod timerMethod in timers.Keys)
                    {
                        SyncTimer timer = timers[timerMethod];
                        timer.Dispose();
                    }
                    timers.Clear();
                }
            }
        }

        #endregion
    }
}
