﻿using System;

using System.Collections.Generic;
using System.Text;
using Nedap.EPC.TDT;
using System.Collections.ObjectModel;

namespace Nedap.HandheldApi
{
    /// <summary>
    /// Presentation layer: high-level communication layer to communicate with Handheld Reader. 
    /// </summary>
    public class PresentationFacade : IDisposable
    {
        #region Constants
        private const String DEFAULT_BUFFER_NAME = "default";
        #endregion

        #region Fields

        private Presentation presentation;
        private bool disposed;

        #endregion

        #region Events

        /// <summary>
        /// This event is triggered when inventory list is cleared (All items from the inventory list are removed).
        /// </summary>
        public event EventHandler<EventArgs> OnClearInventory;
        /// <summary>
        /// This event is triggered when battery percentage value is received from the handheld.
        /// </summary>
        public event EventHandler<ReceiveBatteryPercentageEventArgs> OnReceiveBatteryPercentage;
        /// <summary>
        /// This event is triggered when low battery warning is received from the handheld.
        /// </summary>
        public event EventHandler<EventArgs> OnLowBatteryWarning;
        /// <summary>
        /// This event is triggered when empty battery status is received from the handheld.
        /// </summary>
        public event EventHandler<EventArgs> OnEmptyBatteryStatus;
        /// <summary>
        /// This event is triggered when charging status is received from the handheld.
        /// </summary>
        public event EventHandler<EventArgs> OnChargingStatus;
        /// <summary>
        /// This event is triggered when one RFID tag is red using the handheld.
        /// </summary>
        public event EventHandler<ReadRfidTagEventArgs> OnReadRfidTag;
        /// <summary>
        /// This event is triggered when more than one RFID tags are red using the handheld.
        /// </summary>
        public event EventHandler<ReadRfidTagsEventArgs> OnReadRfidTags;
        /// <summary>
        /// This event is triggered when more than one new RFID tags are red using the handheld.
        /// </summary>
        public event EventHandler<ReadRfidTagsFilteredEventArgs> OnReadRfidTagsFiltered;
        /// <summary>
        /// This event is triggered when detailed tag information is red using the handheld.
        /// </summary>
        public event EventHandler<ReadDetailedTagInformationEventArgs> OnReadDetailedTagInformation;
        /// <summary>
        /// This event is triggered when barcode is red using the handheld.
        /// </summary>
        public event EventHandler<ReadBarcodeEventArgs> OnReadBarcode;
        /// <summary>
        /// This event is triggered when the handheld was not able to read barcode.
        /// </summary>
        public event EventHandler<EventArgs> OnNotReadBarcode;
        /// <summary>
        /// This event is triggered when button on the handheld is pushed.
        /// </summary>
        public event EventHandler<EventArgs> OnPushButton;
        /// <summary>
        /// This event is triggered when handheld is shaken.
        /// </summary>
        public event EventHandler<EventArgs> OnShake;
        /// <summary>
        /// This event is triggered when NFC card is red using the handheld.
        /// </summary>
        public event EventHandler<ReadNfcCardEventArgs> OnReadNfcCard;
        /// <summary>
        /// This event is triggered when the handheld is disconnected and action is executed on the device.
        /// </summary>
        public event EventHandler<EventArgs> OnIsDisconnected;
        /// <summary>
        /// This event is triggered when single read of RFID is finished.
        /// </summary>
        public event EventHandler<EventArgs> OnFinishSingleRead;
        /// <summary>
        /// This event is triggered when when inventory process or monitor process are continued.
        /// </summary>
        public event EventHandler<EventArgs> OnContinue;
        /// <summary>
        /// This event is triggered when when inventory process or monitor process are paused.
        /// </summary>
        public event EventHandler<EventArgs> OnPause;
        /// <summary>
        /// This event is triggered when writing of EPC has failed with multiple read.
        /// </summary>
        public event EventHandler<EventArgs> OnFailWriteEpcMultipleRead;
        /// <summary>
        /// This event is triggered when writing of EPC has failed with zero read.
        /// </summary>
        public event EventHandler<EventArgs> OnFailWriteEpcZeroRead;
        /// <summary>
        /// This event is triggered when writing of EPC has failed. 
        /// Failure happened during the write of data.
        /// </summary>
        public event EventHandler<EventArgs> OnFailWriteEpcWriteFailed;
        /// <summary>
        /// This event is triggered when writing of EPC has failed. 
        /// Failure happened during the verification of written of data.
        /// </summary>
        public event EventHandler<EventArgs> OnFailWriteEpcVerificationFailed;
        /// <summary>
        /// This event is triggered when writing of EPC has finished successfully.
        /// </summary>
        public event EventHandler<EventArgs> OnSucceedWriteEpc;
        /// <summary>
        /// This event is triggered when device is connected to the handheld.
        /// </summary>
        public event EventHandler<EventArgs> OnConnect;
        /// <summary>
        /// This event is triggered when device is disconnected from the handheld.
        /// </summary>
        public event EventHandler<EventArgs> OnDisconnect;
        /// <summary>
        /// This event is triggered when supported regions are received from the handheld.
        /// </summary>
        public event EventHandler<ReceiveSupportedRegionsEventArgs> OnReceiveSupportedRegions;

        #endregion

        #region Constructor/destructor

        /// <summary>
        /// Constructor
        /// </summary>
        /// <param name="session"></param>
        public PresentationFacade(ISession session)
        {
            presentation = new Presentation(session);
            
            presentation.OnClearInventory += new EventHandler<EventArgs>(presentationFacade_OnClearInventory);
            presentation.OnReceiveBatteryPercentage += new EventHandler<ReceiveBatteryPercentageEventArgs>(presentationFacade_OnReceiveBatteryPercentage);
            presentation.OnLowBatteryWarning += new EventHandler<EventArgs>(presentationFacade_OnLowBatteryWarning);
            presentation.OnEmptyBatteryStatus += new EventHandler<EventArgs>(presentationFacade_OnEmptyBatteryStatus);
            presentation.OnChargingStatus += new EventHandler<EventArgs>(presentationFacade_OnChargingStatus);
            presentation.OnReadRFIDTag += new EventHandler<ReadRfidTagEventArgs>(presentationFacade_OnReadRFIDTag);
            presentation.OnReadRFIDTags += new EventHandler<ReadRfidTagsEventArgs>(presentationFacade_OnReadRFIDTags);
            presentation.OnReadRFIDTagsFiltered += new EventHandler<ReadRfidTagsFilteredEventArgs>(presentationFacade_OnReadRFIDTagsFiltered);
            presentation.OnReadDetailedTagInformation += new EventHandler<ReadDetailedTagInformationEventArgs>(presentationFacade_OnReadDetailedTagInformation);
            presentation.OnReadBarcode += new EventHandler<ReadBarcodeEventArgs>(presentationFacade_OnReadBarcode);
            presentation.OnNotReadBarcode += new EventHandler<EventArgs>(presentationFacade_OnNotReadBarcode);
            presentation.OnPushButton += new EventHandler<EventArgs>(presentationFacade_OnPushButton);
            presentation.OnShake += new EventHandler<EventArgs>(presentationFacade_OnShake);
            presentation.OnReadNFCCard += new EventHandler<ReadNfcCardEventArgs>(presentationFacade_OnReadNFCCard);
            presentation.OnIsDisconnected += new EventHandler<EventArgs>(presentationFacade_OnIsDisconnected);
            presentation.OnFinishSingleRead += new EventHandler<EventArgs>(presentationFacade_OnFinishSingleRead);
            presentation.OnContinue += new EventHandler<EventArgs>(presentationFacade_OnContinue);
            presentation.OnPause += new EventHandler<EventArgs>(presentationFacade_OnPause);
            presentation.OnFailWriteEpcMultipleRead += new EventHandler<EventArgs>(presentationFacade_OnFailWriteEpcMultipleRead);
            presentation.OnFailWriteEpcZeroRead += new EventHandler<EventArgs>(presentationFacade_OnFailWriteEpcZeroRead);
            presentation.OnFailWriteEpcWriteFailed += new EventHandler<EventArgs>(presentationFacade_OnFailWriteEpcWriteFailed);
            presentation.OnFailWriteEpcVerificationFailed += new EventHandler<EventArgs>(presentationFacade_OnFailWriteEpcVerificationFailed);
            presentation.OnSucceedWriteEPC += new EventHandler<EventArgs>(presentationFacade_OnSucceedWriteEPC);
            presentation.OnConnect += new EventHandler<EventArgs>(presentationFacade_OnConnect);
            presentation.OnDisconnect += new EventHandler<EventArgs>(presentationFacade_OnDisconnect);
            presentation.OnReceiveSupportedRegions += new EventHandler<ReceiveSupportedRegionsEventArgs>(presentationFacade_OnReceiveSupportedRegions);
        }
        /// <summary>
        /// Destructor
        /// </summary>
        ~PresentationFacade()
        {
            Dispose();
        }

        #endregion

        #region Event triggers

        private void presentationFacade_OnClearInventory(object sender, EventArgs e)
        {
            if (OnClearInventory != null)
            {
                OnClearInventory(this, e);
            }
        }

        private void presentationFacade_OnReceiveBatteryPercentage(object sender, ReceiveBatteryPercentageEventArgs e)
        {
            if (OnReceiveBatteryPercentage != null)
            {
                OnReceiveBatteryPercentage(this, e);
            }
        }

        private void presentationFacade_OnLowBatteryWarning(object sender, EventArgs e)
        {
            if (OnLowBatteryWarning != null)
            {
                OnLowBatteryWarning(this, e);
            }
        }

        private void presentationFacade_OnEmptyBatteryStatus(object sender, EventArgs e)
        {
            if (OnEmptyBatteryStatus != null)
            {
                OnEmptyBatteryStatus(this, e);
            }
        }

        private void presentationFacade_OnChargingStatus(object sender, EventArgs e)
        {
            if (OnChargingStatus != null)
            {
                OnChargingStatus(this, e);
            }
        }

        private void presentationFacade_OnReadRFIDTag(object sender, ReadRfidTagEventArgs e)
        {
            if (OnReadRfidTag != null)
            {
                OnReadRfidTag(this, e);
            }
        }

        private void presentationFacade_OnReadRFIDTags(object sender, ReadRfidTagsEventArgs e)
        {
            if (OnReadRfidTags != null)
            {
                OnReadRfidTags(this, e);
            }
        }

        void presentationFacade_OnReadRFIDTagsFiltered(object sender, ReadRfidTagsFilteredEventArgs e)
        {
            if (OnReadRfidTagsFiltered != null)
            {
                OnReadRfidTagsFiltered(this, e);
            }
        }

        private void presentationFacade_OnReadDetailedTagInformation(object sender, ReadDetailedTagInformationEventArgs e)
        {
            if (OnReadDetailedTagInformation != null)
            {
                OnReadDetailedTagInformation(this, e);
            }
        }

        private void presentationFacade_OnReadBarcode(object sender, ReadBarcodeEventArgs e)
        {
            if (OnReadBarcode != null)
            {
                OnReadBarcode(this, e);
            }
        }

        private void presentationFacade_OnNotReadBarcode(object sender, EventArgs e)
        {
            if (OnNotReadBarcode != null)
            {
                OnNotReadBarcode(this, e);
            }
        }

        private void presentationFacade_OnPushButton(object sender, EventArgs e)
        {
            if (OnPushButton != null)
            {
                OnPushButton(this, e);
            }
        }

        private void presentationFacade_OnShake(object sender, EventArgs e)
        {
            if (OnShake != null)
            {
                OnShake(this, e);
            }
        }

        private void presentationFacade_OnReadNFCCard(object sender, ReadNfcCardEventArgs e)
        {
            if (OnReadNfcCard != null)
            {
                OnReadNfcCard(this, e);
            }
        }

        private void presentationFacade_OnIsDisconnected(object sender, EventArgs e)
        {
            if (OnIsDisconnected != null)
            {
                OnIsDisconnected(this, e);
            }
        }

        private void presentationFacade_OnFinishSingleRead(object sender, EventArgs e)
        {
            if (OnFinishSingleRead != null)
            {
                OnFinishSingleRead(this, e);
            }
        }

        private void presentationFacade_OnContinue(object sender, EventArgs e)
        {
            if (OnContinue != null)
            {
                OnContinue(this, e);
            }
        }

        private void presentationFacade_OnPause(object sender, EventArgs e)
        {
            if (OnPause != null)
            {
                OnPause(this, e);
            }
        }

        private void presentationFacade_OnFailWriteEpcMultipleRead(object sender, EventArgs e)
        {
            if (OnFailWriteEpcMultipleRead != null)
            {
                OnFailWriteEpcMultipleRead(this, e);
            }
        }

        private void presentationFacade_OnFailWriteEpcZeroRead(object sender, EventArgs e)
        {
            if (OnFailWriteEpcZeroRead != null)
            {
                OnFailWriteEpcZeroRead(this, e);
            }
        }

        private void presentationFacade_OnFailWriteEpcWriteFailed(object sender, EventArgs e)
        {
            if (OnFailWriteEpcWriteFailed != null)
            {
                OnFailWriteEpcWriteFailed(this, e);
            }
        }

        private void presentationFacade_OnFailWriteEpcVerificationFailed(object sender, EventArgs e)
        {
            if (OnFailWriteEpcVerificationFailed != null)
            {
                OnFailWriteEpcVerificationFailed(this, e);
            }
        }

        private void presentationFacade_OnSucceedWriteEPC(object sender, EventArgs e)
        {
            if (OnSucceedWriteEpc != null)
            {
                OnSucceedWriteEpc(this, e);
            }
        }

        private void presentationFacade_OnConnect(object sender, EventArgs e)
        {
            if (OnConnect != null)
            {
                OnConnect(this, e);
            }
        }

        private void presentationFacade_OnDisconnect(object sender, EventArgs e)
        {
            if (OnDisconnect != null)
            {
                OnDisconnect(this, e);
            }
        }

        private void presentationFacade_OnReceiveSupportedRegions(object sender, ReceiveSupportedRegionsEventArgs e)
        {
            if (OnReceiveSupportedRegions != null)
            {
                OnReceiveSupportedRegions(this, e);
            }
        }

        #endregion

        #region Facade methods

        /// <summary>
        /// This method can be used to try to read NFC cards. It’ll enable the ﬁeld, and continuously try to read cards in the ﬁeld.
        /// </summary>
        public void NfcReadStart()
        {
            presentation.NfcReadStart();
        }

        /// <summary>
        /// The inventory methods provide an easy way to do inventory. It automatically filters ‘multiple’ reads in a buffer and generates a ‘beep’ on new tags. 
        /// EPC air interface parameters are set automatically to session 1, target A (no switching) and automatic Q.
        /// </summary>
        public void RfidInventoryStart()
        {
            presentation.RfidInventoryStart(DEFAULT_BUFFER_NAME, null, false);
        }

        /// <summary>
        /// The inventory methods provide an easy way to do inventory. It automatically filters ‘multiple’ reads in a buffer and generates a ‘beep’ on new tags. 
        /// EPC air interface parameters are set automatically to session 1, target A (no switching) and automatic Q.
        /// </summary>
        /// <param name="select">Select statement</param>
        public void RfidInventoryStart(Select select)
        {
            presentation.RfidInventoryStart(DEFAULT_BUFFER_NAME, select, false);
        }

        /// <summary>
        /// The inventory methods provide an easy way to do inventory. It automatically filters ‘multiple’ reads in a buffer and generates a ‘beep’ on new tags. 
        /// EPC air interface parameters are set automatically to session 1, target A (no switching) and automatic Q.
        /// </summary>
        /// <param name="bufferName"></param>
        public void RfidInventoryStart(String bufferName)
        {
            presentation.RfidInventoryStart(bufferName, null, false);
        }

        /// <summary>
        /// The inventory methods provide an easy way to do inventory. It automatically filters ‘multiple’ reads in a buffer and generates a ‘beep’ on new tags. 
        /// EPC air interface parameters are set automatically to session 1, target A (no switching) and automatic Q.
        /// </summary>
        /// <param name="bufferName"></param>
        /// <param name="select"></param>
        public void RfidInventoryStart(String bufferName, Select select)
        {
            presentation.RfidInventoryStart(bufferName, select, false);
        }

        /// <summary>
        /// The inventory methods provide an easy way to do inventory. It automatically filters ‘multiple’ reads in a buffer and generates a ‘beep’ on new tags. 
        /// EPC air interface parameters are set automatically to session 1 or 2 depending on highVolume setting, target A (no switching) and automatic Q.
        /// </summary>
        /// <param name="highVolume">If true session 2 is used, otherwise session 1 is used</param>
        public void RfidInventoryStart(bool highVolume)
        {
            presentation.RfidInventoryStart(DEFAULT_BUFFER_NAME, null, highVolume);
        }

        /// <summary>
        /// The inventory methods provide an easy way to do inventory. It automatically filters ‘multiple’ reads in a buffer and generates a ‘beep’ on new tags. 
        /// EPC air interface parameters are set automatically to session 1 or 2 depending on highVolume setting, target A (no switching) and automatic Q.
        /// </summary>
        /// <param name="select">Select statement</param>
        /// <param name="highVolume">If true session 2 is used, otherwise session 1 is used</param>
        public void RfidInventoryStart(Select select, bool highVolume)
        {
            presentation.RfidInventoryStart(DEFAULT_BUFFER_NAME, select, highVolume);
        }

        /// <summary>
        /// The inventory methods provide an easy way to do inventory. It automatically filters ‘multiple’ reads in a buffer and generates a ‘beep’ on new tags. 
        /// EPC air interface parameters are set automatically to session 1 or 2 depending on highVolume setting, target A (no switching) and automatic Q.
        /// </summary>
        /// <param name="bufferName"></param>
        /// <param name="highVolume">If true session 2 is used, otherwise session 1 is used</param>
        public void RfidInventoryStart(String bufferName, bool highVolume)
        {
            presentation.RfidInventoryStart(bufferName, null, highVolume);
        }

        /// <summary>
        /// The inventory methods provide an easy way to do inventory. It automatically filters ‘multiple’ reads in a buffer and generates a ‘beep’ on new tags. 
        /// EPC air interface parameters are set automatically to session 1 or 2 depending on highVolume setting, target A (no switching) and automatic Q.
        /// </summary>
        /// <param name="bufferName"></param>
        /// <param name="select"></param>
        /// <param name="highVolume">If true session 2 is used, otherwise session 1 is used</param>
        public void RfidInventoryStart(String bufferName, Select select, bool highVolume)
        {
            presentation.RfidInventoryStart(bufferName, select, highVolume);
        }

        /// <summary>
        /// Resets the RFID inventory 'seen' list of the "default" buffer. 
        /// Please make sure that if a list is also kept in the application layer, clean it after the relevant notification is received.
        /// </summary>
        public void RfidInventoryReset()
        {
            presentation.RfidInventoryReset();
        }

        /// <summary>
        /// Resets the RFID inventory 'seen' list of the given buffer. 
        /// Please make sure that if a list is also kept in the application layer, clean it after the relevant notification is received.
        /// </summary>
        /// <param name="bufferName"></param>
        public void RfidInventoryReset(String bufferName)
        {
            presentation.RfidInventoryReset(bufferName);
        }

        /// <summary>
        /// Continously return tags in the field, together with their RSSI values. Can be used to implement 'search' functionality.
        /// </summary>
        public void RfidMonitorStart()
        {
            RfidMonitorStart(null);
        }

        /// <summary>
        /// Monitors the tags in the field, with a select statement.
        /// </summary>
        /// <param name="select">Select statement</param>
        public void RfidMonitorStart(Select select)
        {
            presentation.RfidMonitorStart(select);
        }

        /// <summary>
        /// Does a single read cycle of tags.
        /// </summary>
        public void RfidReadSingle()
        {
            RfidReadSingle(null);
        }

        /// <summary>
        /// Does a single read cycle of tags, with a select statement.
        /// </summary>
        /// <param name="select">Select statement</param>
        public void RfidReadSingle(Select select)
        {
            presentation.RfidReadSingle(select);
        }

        /// <summary>
        /// To obtain the tag manufacturer ID (MDID), the tag model number, or the tag’s extended identifier field (XTID), use this method.
        /// </summary>
        public void RfidReadTag()
        {
            RfidReadTag(null);
        }

        /// <summary>
        /// To obtain the tag manufacturer ID (MDID), the tag model number, or the tag’s extended identifier field (XTID), use this method.
        /// </summary>
        /// <param name="code">Code</param>
        public void RfidReadTag(EpcCode code)
        {
            presentation.RfidReadTag(code);
        }

        /// <summary>
        /// Write a new EPC value to a tag.
        /// </summary>
        /// <param name="newEpcCode">The new EPC value</param>
        public void RfidWriteEpc(EpcCode newEpcCode)
        {
            presentation.RfidWriteEpc(newEpcCode);
        }

        /// <summary>
        /// This method simulates button push on the handheld reader.
        /// </summary>
        public void SimulateButtonPush()
        {
            presentation.SimulateButtonPush();
        }

        /// <summary>
        /// Check to see if the handheld reader has a barcode reader on-board.
        /// </summary>
        /// <returns>If true, there is a barcode reader</returns>
        public bool BarcodeHasReader()
        {
            return presentation.BarcodeHasReader();
        }

        /// <summary>
        /// Read a barcode with a specified type, and with a specified timeout.
        /// </summary>
        /// <param name="barcodeType">Type of barcode</param>
        /// <param name="duration">Timeout in seconds</param>
        public void BarcodeRead(BarcodeType barcodeType, int duration)
        {
            Collection<BarcodeType> barcodeTypes = new Collection<BarcodeType>();
            barcodeTypes.Add(barcodeType);
            BarcodeRead(barcodeTypes, duration);
        }

        /// <summary>
        /// Read a barcode with a specified array of types, and with a specified timeout.
        /// </summary>
        /// <param name="barcodeTypes">Array of barcode types</param>
        /// <param name="duration">Timeout in seconds</param>
        public void BarcodeRead(Collection<BarcodeType> barcodeTypes, int duration)
        {
            presentation.BarcodeRead(barcodeTypes, duration);
        }

        /// <summary>
        /// Let the beeper on the handheld play a tune.
        /// </summary>
        /// <param name="tune">Type of tune</param>
        public void UIPlay(UITune tune)
        {
            presentation.UIPlay(tune);
        }

        /// <summary>
        /// Let the vibration motor on the handheld vibrate with beeper playing a tune.
        /// </summary>
        public void UIVibrateTune(UITune tune)
        {
            presentation.UIVibrateTune(tune);
        }

        /// <summary>
        /// Let the vibration motor on the handheld vibrate.
        /// </summary>
        public void UIVibrate()
        {
            presentation.UIVibrate();
        }

        /// <summary>
        /// Enables or disables the logging functionality with Flurry.com. The information logged to Flurry is anonymous. 
        /// We’ll log the usages and performance of the !d hand (e.g. how long does an inventory take, how often will the user press the button, how many tags are read in an hour, 
        /// how is the battery quality over time, etc.). 
        /// This will help Nedap to improve the code, and gather information on the performance and usage of the !d hand in the field. 
        /// We kindly ask you to leave this enabled (which is the default). If you’d prefer to turn this off, feel free to do so.
        /// </summary>
        /// <param name="enabled">Enabled</param>
        public void HandheldLogEnabled(bool enabled)
        {
            presentation.HandheldLogEnabled(enabled);
        }

        /// <summary>
        /// Stop the current operation, e.g. inventory or monitoring. Can be used when switching from one view to another.
        /// </summary>
        public void HandheldStop()
        {
            presentation.HandheldStop();
        }

        /// <summary>
        /// Checks if the handheld is connected or not.
        /// </summary>
        /// <returns>Returns indication if handheld is connected</returns>
        public bool HandheldIsConnected()
        {
            return presentation.HandheldIsConnected();
        }

        /// <summary>
        /// Will let the handheld kill the bluetooth connection. Can be used to pair the handheld with another device.
        /// </summary>
        public void HandheldDisconnect()
        {
            presentation.HandheldDisconnect();
        }

        /// <summary>
        /// Get information on the handheld. 
        /// </summary>
        /// <returns>DeviceInfo class object with three parameters: serial number, firmware and hardware version.</returns>
        public DeviceInfo HandheldGetInformation()
        {
            return presentation.HandheldGetInformation();
        }

        /// <summary>
        /// Returns the current battery status in percentage.
        /// </summary>
        public void HandheldGetBatteryPercentage()
        {
            presentation.HandheldGetBatteryPercentage();
        }

        #endregion

        #region IDisposable Members

        /// <summary>
        /// Dispose.
        /// </summary>
        public void Dispose()
        {
            if (!disposed)
            {
                disposed = true;
                presentation.OnClearInventory -= new EventHandler<EventArgs>(presentationFacade_OnClearInventory);
                presentation.OnReceiveBatteryPercentage -= new EventHandler<ReceiveBatteryPercentageEventArgs>(presentationFacade_OnReceiveBatteryPercentage);
                presentation.OnLowBatteryWarning -= new EventHandler<EventArgs>(presentationFacade_OnLowBatteryWarning);
                presentation.OnEmptyBatteryStatus -= new EventHandler<EventArgs>(presentationFacade_OnEmptyBatteryStatus);
                presentation.OnChargingStatus -= new EventHandler<EventArgs>(presentationFacade_OnChargingStatus);
                presentation.OnReadRFIDTag -= new EventHandler<ReadRfidTagEventArgs>(presentationFacade_OnReadRFIDTag);
                presentation.OnReadRFIDTags -= new EventHandler<ReadRfidTagsEventArgs>(presentationFacade_OnReadRFIDTags);
                presentation.OnReadRFIDTagsFiltered += new EventHandler<ReadRfidTagsFilteredEventArgs>(presentationFacade_OnReadRFIDTagsFiltered);
                presentation.OnReadDetailedTagInformation -= new EventHandler<ReadDetailedTagInformationEventArgs>(presentationFacade_OnReadDetailedTagInformation);
                presentation.OnReadBarcode -= new EventHandler<ReadBarcodeEventArgs>(presentationFacade_OnReadBarcode);
                presentation.OnNotReadBarcode -= new EventHandler<EventArgs>(presentationFacade_OnNotReadBarcode);
                presentation.OnPushButton -= new EventHandler<EventArgs>(presentationFacade_OnPushButton);
                presentation.OnShake -= new EventHandler<EventArgs>(presentationFacade_OnShake);
                presentation.OnReadNFCCard -= new EventHandler<ReadNfcCardEventArgs>(presentationFacade_OnReadNFCCard);
                presentation.OnIsDisconnected -= new EventHandler<EventArgs>(presentationFacade_OnIsDisconnected);
                presentation.OnFinishSingleRead -= new EventHandler<EventArgs>(presentationFacade_OnFinishSingleRead);
                presentation.OnContinue -= new EventHandler<EventArgs>(presentationFacade_OnContinue);
                presentation.OnPause -= new EventHandler<EventArgs>(presentationFacade_OnPause);
                presentation.OnSucceedWriteEPC -= new EventHandler<EventArgs>(presentationFacade_OnSucceedWriteEPC);
                presentation.OnConnect -= new EventHandler<EventArgs>(presentationFacade_OnConnect);
                presentation.OnDisconnect -= new EventHandler<EventArgs>(presentationFacade_OnDisconnect);
                presentation.OnReceiveSupportedRegions -= new EventHandler<ReceiveSupportedRegionsEventArgs>(presentationFacade_OnReceiveSupportedRegions);

                presentation.Dispose();
            }
        }

        #endregion
    }
}
