﻿using System;
namespace Nedap.HandheldApi
{
    /// <summary>
    /// Session layer interface: low-level communication layer interface that can be used to communicate with Handheld Reader. 
    /// </summary>
    public interface ISession
    {
        /// <summary>
        /// 	Check to see if the handheld reader has a barcode reader on-board.
        /// </summary>
        /// <returns>true if there is a barcode reader</returns>
        bool BarcodeHasReader();
        /// <summary>
        /// Enables or disable a certain type of barcode.
        /// </summary>
        /// <param name="barcodeType">NHBarcodeType</param>
        /// <param name="enabled">enabled or disabled</param>
        void BarcodeSetStatusForType(Nedap.HandheldApi.BarcodeType barcodeType, bool enabled);
        /// <summary>
        /// Try to read a barcode for a defined number of seconds.
        /// </summary>
        /// <param name="duration">duration in seconds</param>
        void BarcodeTryDecode(int duration);
        /// <summary>
        /// Disposes private objects.
        /// </summary>
        void Dispose();
        /// <summary>
        ///  Will let the handheld kill the bluetooth connection. Can be used to pair the handheld with another device.
        /// </summary>
        void HandheldDisconnect();
        /// <summary>
        /// Returns the current battery status in percentage.
        /// </summary>
        void HandheldGetBatteryPercentage();
        /// <summary>
        /// Get information on the handheld. 
        /// </summary>
        /// <returns>DeviceInfo class object with three parameters: serial number, firmware and hardware version.</returns>
        DeviceInfo HandheldGetInformation();
        /// <summary>
        /// Checks if the handheld is connected or not.
        /// </summary>
        /// <returns>true if handheld is connected</returns>
        bool HandheldIsConnected();
        /// <summary>
        /// The handheld will try to read a NFC card.
        /// </summary>
        void NfcTryToReadCard();
        /// <summary>
        /// This event is triggered after handheld fails to read a barcode
        /// </summary>
        event EventHandler<EventArgs> OnSessionBarcodeNotRead;
        /// <summary>
        /// This event is triggered after handheld reads a barcode.
        /// </summary>
        event EventHandler<RfidReadABarcodeEventArgs> OnSessionBarcodeRead;
        /// <summary>
        /// This event is triggered after handheld device connects. 
        /// </summary>
        event EventHandler<EventArgs> OnSessionHandheldConnect;
        /// <summary>
        /// This event is triggered after handheld device disconnects. 
        /// </summary>
        event EventHandler<EventArgs> OnSessionHandheldDisconnect;
        /// <summary>
        /// This event is triggered after battery charge percentage is received.
        /// </summary>
        event EventHandler<HandheldReceiveBatteryPercentageEventArgs> OnSessionHandheldReceiveBatteryPercentage;
        /// <summary>
        /// This event is triggered after charging status responce is received.
        /// </summary>
        event EventHandler<EventArgs> OnSessionHandheldReceiveChargingStatus;
        /// <summary>
        /// This event is triggered after empty battery status responce is received.
        /// </summary>
        event EventHandler<EventArgs> OnSessionHandheldReceiveEmptyBatteryStatus;
        /// <summary>
        /// This event is triggered after handheld device battery reaches minimal level. 
        /// </summary>
        event EventHandler<EventArgs> OnSessionHandheldReceiveLowBatteryWarning;
        /// <summary>
        /// This event is triggered after error responce is received.
        /// </summary>
        event EventHandler<HandheldReceiveErrorEventArgs> OnSessionHandheldReceiveError;
        /// <summary>
        /// This event is triggered after handheld device fails to found a card.
        /// </summary>
        event EventHandler<EventArgs> OnSessionNfcCardNotFound;
        /// <summary>
        /// This event is triggered after handheld device reads a card.
        /// </summary>
        event EventHandler<RfidReadCardEventArgs> OnSessionNfcCardRead;
        /// <summary>
        /// This event is triggered after handheld device read fails but tags are available.
        /// </summary>
        event EventHandler<RfidFailButTagsAvailableEventArgs> OnSessionRfidFailButTagsAvailable;
        /// <summary>
        /// This event is triggered after handheld device finished read cycle.
        /// </summary>
        event EventHandler<EventArgs> OnSessionRfidFinishReadCycle;
        /// <summary>
        /// This event is triggered after handheld device reads data.
        /// </summary>
        event EventHandler<RfidReadDataEventArgs> OnSessionRfidReadData;
        /// <summary>
        /// This event is triggered after handheld device reads tag.
        /// </summary>
        event EventHandler<RfidReadTagEventArgs> OnSessionRfidReadTag;
        /// <summary>
        /// This event is triggered after supported regions responce is received..
        /// </summary>
        event EventHandler<RfidReceiveSupportedRegionsEventArgs> OnSessionRfidReceiveSupportedRegions;
        /// <summary>
        /// This event is triggered after handheld device writes data and more tags are available.
        /// </summary>
        event EventHandler<RfidWriteDataMoreTagsAvailableEventArgs> OnSessionRfidWriteDataMoreTagsAvailable;
        /// <summary>
        /// This event is triggered after handheld device tries to write EPC and there are no tags in the field
        /// </summary>
        event EventHandler<EventArgs> OnSessionRfidWriteNoTagsInField;
        /// <summary>
        /// This event is triggered after handheld device tries to write EPC and there are too much tags in the field
        /// </summary>
        event EventHandler<EventArgs> OnSessionRfidWriteTooMuchTagsInField;
        /// <summary>
        /// This event is triggered after handheld device tries to write EPC and fails.
        /// </summary>
        event EventHandler<EventArgs> OnSessionRfidWriteToTagFailed;
        /// <summary>
        /// This event is triggered after handheld device tries to write EPC and succeeds.
        /// </summary>
        event EventHandler<EventArgs> OnSessionRfidWriteSuccess;
        /// <summary>
        /// This event is triggered after button is pushed on the handheld device.
        /// </summary>
        event EventHandler<EventArgs> OnSessionUIPushButton;
        /// <summary>
        /// This event is triggered after the handheld device is shaken.
        /// </summary>
        event EventHandler<EventArgs> OnSessionUIShake;
        /// <summary>
        /// Gets the supported regions by the handheld reader.
        /// </summary>
        void RfidGetSupportedRegions();
        /// <summary>
        /// Will read specific data from a RFID tag.
        /// </summary>
        /// <param name="memoryBank">the memory bank to read data from</param>
        /// <param name="addressInWords">the address to read data from (in words)</param>
        /// <param name="lengthInWords">the length of the data to read (in words)</param>
        /// <param name="timeout">timeout in milliseconds</param>
        /// <param name="power">the power, possible values range from 10 to 23 dBm</param>
        /// <param name="select">select statement, can also be set to null</param>
        /// <param name="password">password, can also be set to null</param>
        void RfidReadData(Nedap.HandheldApi.RfidEpcMemoryBank memoryBank, int addressInWords, int lengthInWords, int timeout, int power, Nedap.HandheldApi.Select select, Nedap.HandheldApi.Password password);
        /// <summary>
        /// Will read RFID tags in the field, for a specified duration, with a specific power and with a select to select a subset of tags.
        /// </summary>
        /// <param name="timeout">timeout in milliseconds</param>
        /// <param name="power">the power, possible values range from 10 to 23 dBm</param>
        /// <param name="select">select statement, can also be set to null</param>
        void RfidReadEpcs(int timeout, int power, Nedap.HandheldApi.Select select);
        /// <summary>
        /// Enables Continuous Wave (CW) signal, used for testing and measurement purposes.
        /// </summary>
        /// <param name="enabled">CW enabled</param>
        /// <param name="antenna">select antenna one or two</param>
        void RfidSetCWSignal(bool enabled, int antenna);
        /// <summary>
        /// Set frequency hop table.
        /// </summary>
        /// <param name="frequencies">frequencies array with frequencies (NSNumber containing NHRFIDFrequency)</param>
        void RfidSetFrequencyHopTable(System.Collections.Generic.List<Nedap.HandheldApi.RfidFrequency> frequencies);
        /// <summary>
        /// Sets the miller value for the EPC Gen2 protocol.
        /// </summary>
        /// <param name="mValue">mValue miller value</param>
        void RfidSetMValue(Nedap.HandheldApi.RfidEpcMValue mValue);
        /// <summary>
        /// Sets the Q value for the EPC Gen2 protocol.
        /// </summary>
        /// <param name="qValue">qValue the Q value</param>
        /// <param name="dynamic">dynamic whether the Q value is dynamically adjusted or not</param>
        void RfidSetQValue(int qValue, bool dynamic);
        /// <summary>
        /// Sets the region that the handheld reader operates in. The supported regions can be obtained by using +(void)rfidGetSupportedRegions.
        /// </summary>
        /// <param name="region">RfidRegion to set</param>
        void RfidSetRegion(Nedap.HandheldApi.RfidRegion region);
        /// <summary>
        /// Sets the session for the EPC Gen2 protocol.
        /// </summary>
        /// <param name="session">session session (S0, S1, S2 or S3)</param>
        void RfidSetSession(Nedap.HandheldApi.RfidEpcSession session);
        /// <summary>
        /// Sets the target for the EPC Gen2 protocol.
        /// </summary>
        /// <param name="target">target target (A or B)</param>
        /// <param name="toggle">toggle toggle between A and B</param>
        void RfidSetTarget(Nedap.HandheldApi.RfidEpcTarget target, bool toggle);
        /// <summary>
        /// Write data to a specified memory bank and address in a tag.
        /// </summary>
        /// <param name="dataToWrite">the data to write.</param>
        /// <param name="memoryBank">the memory bank where to write the data to</param>
        /// <param name="addressInWords">the address in words where the data needs to be written</param>
        /// <param name="timeout">the address in words where the data needs to be written</param>
        /// <param name="power">the power, possible values range from 10 to 23 dBm</param>
        /// <param name="select">select statement, can also be set to null</param>
        /// <param name="password">password, can also be set to null</param>
        void RfidWriteData(byte[] dataToWrite, Nedap.HandheldApi.RfidEpcMemoryBank memoryBank, int addressInWords, int timeout, int power, Nedap.HandheldApi.Select select, Nedap.HandheldApi.Password password);
        /// <summary>
        /// Writes a new EPC code to a tag.
        /// </summary>
        /// <param name="epc">new EPC code</param>
        void RfidWriteEpc(Nedap.EPC.TDT.EpcCode epc);
        /// <summary>
        ///  Let the white status led on the handheld blink or not.
        /// </summary>
        /// <param name="blinking">blinking enabled</param>
        void UILedSetBlinking(bool blinking);
        /// <summary>
        /// The beeper on the handheld will play the defined tune.
        /// </summary>
        /// <param name="tune"></param>
        void UIPlay(Nedap.HandheldApi.UITune tune);
        /// <summary>
        /// The vibration motor on the handheld will vibrate.
        /// </summary>
        void UIVibrate();
        /// <summary>
        ///  The vibration motor on the handheld will vibrate for a specified duration.
        /// </summary>
        /// <param name="timeout">vibrate duration in tens of a second</param>
        void UIVibrate(int timeout);
        /// <summary>
        /// The vibration motor on the handheld will vibrate and the beeper on the handheld will play the defined tune.
        /// </summary>
        /// <param name="tune"></param>
        void UIVibrateTune(UITune tune);
    }
}
