using System;
using System.Collections.Generic;
using System.Text;
using Nedap.EPC.TDT;
using Nedap.HandheldApi.CommunicationQueue;
using System.Diagnostics;

namespace Nedap.HandheldApi
{
    /// <summary>
    /// Session layer: low-level communication layer to communicate with Handheld Reader. 
    /// </summary>
    public class Session : IDisposable, ISession
    {
        #region Enumerations

        /// <summary>
        /// Enumeration for available commands used in communication with handheld.
        /// </summary>
        public enum Command
        {
            GeneralGetBatteryPercentage = 0x01,
            GeneralGetDeviceDetails = 0x02,
            GeneralDisconnect = 0x03,

            RFIDSetParameter = 0x21,
            RFIDGetParameter = 0x22,
            RFIDReadEPC = 0x23,
            RFIDReadData = 0x24,
            RFIDWrite = 0x25,
            RFIDLock = 0x26,
            RFIDKill = 0x27,
            RFIDSetRegion = 0x28,
            RFIDSetEPCParameters = 0x29,
            RFIDSetCWSignal = 0x2A,
            RFIDSetFrequencyHopTable = 0x2B,
            RFIDGetSupportedRegions = 0x2C,
            RFIDWriteEPC = 0x2D,

            BarcodeSetType = 0x31,
            BarcodeTryDecode = 0x32,

            UILedSetStatusLed = 0x41,
            UIBeeperTune = 0x42,
            UIVibrationVibrate = 0x43,
            UIVibrationTune = 0x44,

            NFCTryToReadCard = 0x51
        }

        /// <summary>
        /// Enumeration for available responces.
        /// </summary>
        public enum Response
        {
            Error = 0xFF,

            GeneralBatteryPercentage = 0x81,
            GeneralDeviceDetails = 0x82,
            GeneralLowBattery = 0x83,
            GeneralEmptyBattery = 0x84,
            GeneralDeviceCharging = 0x85,

            RfidTagCount = 0xA1,
            RfidTags = 0xA2,
            RfidReadData = 0xA3,
            RfidWrite = 0xA4,
            RfidSupportedRegions = 0xA5,
            RFIDWriteEPC = 0xA6,

            BarcodeBarcodeRead = 0xB1,
            BarcodeNothingRead = 0xB2,

            UIButtonPush = 0xC1,
            UIAccelerometerShake = 0xC2,

            NfcNoCardFound = 0xD1,
            NfcCardFound = 0xD2
        }

        private enum Error
        {
            Crc = 0x00,
            WrongNumberOfBytes = 0x01,

            RfidAnalogFrontendNotEnabled = 0x21,
            RfidTemperatureExceedLimits = 0x22,
            RfidHighReturnLoss = 0x23,
            RfidBusy = 0x24,

            BarcodeNotAvailable = 0x31
        }

        private enum RfidEpcParameters
        {
            Session = 0x00,
            Target = 0x01,
            MValue = 0x02,
            QValue = 0x03
        }

        private enum RfidEpcPasswordType
        {
            Kill = 0,
            Access = 1
        }

        private enum RfidFlag
        {
            Select = 1 << 0,
            SelectInvert = 1 << 1,
            Password = 1 << 4
        }

        private enum RFIDLock
        {
            KillPasswordReadWrite = 1 << 9,
            KillPasswordPermanent = 1 << 8,
            AccessPasswordReadWrite = 1 << 7,
            AccessPasswordPermanent = 1 << 6,
            EpcMemoryReadWrite = 1 << 5,
            EpcMemoryPermanent = 1 << 4,
            TidMemoryReadWrite = 1 << 3,
            TidMemoryPermanent = 1 << 2,
            UserMemoryReadWrite = 1 << 1,
            UserMemoryPermanent = 1 << 0
        }

        private enum RFIDWriteEPCResult
        {
            RFIDWriteEPCResultNoTagsInField = 0,
            RFIDWriteEPCResultTooMuchTagsInField = 1,
            RFIDWriteEPCResultWriteToTagFailed = 2,
            RFIDWriteEPCResultWriteSuccess = 3
        }
        #endregion

        #region Fields

        private ICommunicationPort port;
        private ICommunicationQueue communicationQueue;
        private IConnectionState connectionState;
        private RfidRegion currentRegion;
        private bool connectionStateEnabled;
        private bool disposed;
#if DEBUG
        // Create a logger for use in this class
        private static readonly log4net.ILog log = log4net.LogManager.GetLogger(typeof(Session));
#endif

        /// <summary>
        /// Property that holds basic device information.
        /// </summary>
        private DeviceInfo DeviceInfo { get; set; }

        #endregion

        #region Constructor/destructor

        /// <summary>
        /// Constructor for session layer.
        /// </summary>
        /// <param name="port">Port implementation of ICommunicationPort interface</param>
        /// <param name="connectionStateEnable">Set to true if session layer should provide connect/disconnect detection.</param>
        public Session(ICommunicationPort port, bool connectionStateEnable)
        {
            DeviceInfo = new DeviceInfo("0", "0", "0.0.0", "0.0.0");

            this.port = port;
            communicationQueue = new CommunicationQueue.CommunicationQueue(port);
            communicationQueue.OnResponseReceived += new EventHandler<ResponseReceivedEventArgs>(communicationQueue_OnResponseReceived);
            if (connectionStateEnable)
            {
                connectionStateEnabled = true;
                connectionState = new ConnectionState(communicationQueue);
                connectionState.DeviceConnected += new EventHandler<EventArgs>(connectionState_DeviceConnected);
                connectionState.DeviceDisconnected += new EventHandler<EventArgs>(connectionState_DeviceDisconnected);
            }

        }

        /// <summary>
        /// Destructor.
        /// </summary>
        ~Session()
        {
            Dispose();
        }

        #endregion

        #region IDisposable Members

        /// <summary>
        /// Disposes private objects.
        /// </summary>
        public void Dispose()
        {
            if (!disposed)
            {
                if (connectionStateEnabled)
                {
                    connectionState.DeviceConnected -= new EventHandler<EventArgs>(connectionState_DeviceConnected);
                    connectionState.DeviceDisconnected -= new EventHandler<EventArgs>(connectionState_DeviceDisconnected);
                    connectionState.Dispose();
                }
                communicationQueue.OnResponseReceived -= new EventHandler<ResponseReceivedEventArgs>(communicationQueue_OnResponseReceived);
                communicationQueue.Dispose();
                disposed = true;
            }
        }

        #endregion

        #region Events

        /// <summary>
        /// This event is triggered after handheld device connects. 
        /// </summary>
        public event EventHandler<EventArgs> OnSessionHandheldConnect;

        /// <summary>
        /// This event is triggered after handheld device disconnects. 
        /// </summary>
        public event EventHandler<EventArgs> OnSessionHandheldDisconnect;

        /// <summary>
        /// This event is triggered after battery charge percentage is received.
        /// </summary>
        public event EventHandler<HandheldReceiveBatteryPercentageEventArgs> OnSessionHandheldReceiveBatteryPercentage;

        /// <summary>
        /// This event is triggered after handheld device battery reaches minimal level. 
        /// </summary>
        public event EventHandler<EventArgs> OnSessionHandheldReceiveLowBatteryWarning;

        /// <summary>
        /// This event is triggered after empty battery status responce is received.
        /// </summary>
        public event EventHandler<EventArgs> OnSessionHandheldReceiveEmptyBatteryStatus;

        /// <summary>
        /// This event is triggered after charging status responce is received.
        /// </summary>
        public event EventHandler<EventArgs> OnSessionHandheldReceiveChargingStatus;

        /// <summary>
        /// This event is triggered after error responce is received.
        /// </summary>
        public event EventHandler<HandheldReceiveErrorEventArgs> OnSessionHandheldReceiveError;

        /// <summary>
        /// This event is triggered after handheld device reads tag.
        /// </summary>
        public event EventHandler<RfidReadTagEventArgs> OnSessionRfidReadTag;

        /// <summary>
        /// This event is triggered after handheld device reads data.
        /// </summary>
        public event EventHandler<RfidReadDataEventArgs> OnSessionRfidReadData;

        /// <summary>
        /// This event is triggered after handheld device finished read cycle.
        /// </summary>
        public event EventHandler<EventArgs> OnSessionRfidFinishReadCycle;

        /// <summary>
        /// This event is triggered after handheld device read fails but tags are available.
        /// </summary>
        public event EventHandler<RfidFailButTagsAvailableEventArgs> OnSessionRfidFailButTagsAvailable;

        /// <summary>
        /// This event is triggered after handheld device writes data and more tags are available.
        /// </summary>
        public event EventHandler<RfidWriteDataMoreTagsAvailableEventArgs> OnSessionRfidWriteDataMoreTagsAvailable;

        /// <summary>
        /// This event is triggered after handheld device tries to write EPC and there are no tags in the field
        /// </summary>
        public event EventHandler<EventArgs> OnSessionRfidWriteNoTagsInField;

        /// <summary>
        /// This event is triggered after handheld device tries to write EPC and there are too much tags in the field
        /// </summary>
        public event EventHandler<EventArgs> OnSessionRfidWriteTooMuchTagsInField;

        /// <summary>
        /// This event is triggered after handheld device tries to write EPC and fails.
        /// </summary>
        public event EventHandler<EventArgs> OnSessionRfidWriteToTagFailed;

        /// <summary>
        /// This event is triggered after handheld device tries to write EPC and succeeds.
        /// </summary>
        public event EventHandler<EventArgs> OnSessionRfidWriteSuccess;

        /// <summary>
        /// This event is triggered after supported regions responce is received..
        /// </summary>
        public event EventHandler<RfidReceiveSupportedRegionsEventArgs> OnSessionRfidReceiveSupportedRegions;

        /// <summary>
        /// This event is triggered after handheld reads a barcode.
        /// </summary>
        public event EventHandler<RfidReadABarcodeEventArgs> OnSessionBarcodeRead;

        /// <summary>
        /// This event is triggered after handheld fails to read a barcode
        /// </summary>
        public event EventHandler<EventArgs> OnSessionBarcodeNotRead;

        /// <summary>
        /// This event is triggered after button is pushed on the handheld device.
        /// </summary>
        public event EventHandler<EventArgs> OnSessionUIPushButton;

        /// <summary>
        /// This event is triggered after the handheld device is shaken.
        /// </summary>
        public event EventHandler<EventArgs> OnSessionUIShake;

        /// <summary>
        /// This event is triggered after handheld device fails to found a card.
        /// </summary>
        public event EventHandler<EventArgs> OnSessionNfcCardNotFound;

        /// <summary>
        /// This event is triggered after handheld device reads a card.
        /// </summary>
        public event EventHandler<RfidReadCardEventArgs> OnSessionNfcCardRead;

        #endregion

        #region Management

        private void connectionState_DeviceConnected(object sender, EventArgs e)
        {
            HandheldGetDeviceDetails();
            if (OnSessionHandheldConnect != null)
            {
                OnSessionHandheldConnect(this, EventArgs.Empty);
            }
        }

        private void connectionState_DeviceDisconnected(object sender, EventArgs e)
        {
            port.Disconnect();
            if (OnSessionHandheldDisconnect != null)
            {
                OnSessionHandheldDisconnect(this, EventArgs.Empty);
            }
        }

        private void communicationQueue_OnResponseReceived(object sender, ResponseReceivedEventArgs e)
        {
            ParseResponse(e.Response, e.Data);
        }

        private byte[] RfidPrepareStandardParameters(int timeout, int power, Select select, Password password)
        {
            // check timeout
            if (timeout > 1000)
            {
                timeout = 1000;
#if DEBUG
                log.Warn("RFID: Timeout cannot be larger than 1000 ms.");
#endif
            }

            // check power
            if (power > 23)
            {
                power = 23;
#if DEBUG
                log.Warn("RFID: Output power cannot be larger than 23 dBm.");
#endif
            }
            if (power < 10)
            {
#if DEBUG
                log.Warn("RFID: Output power cannot be lower than 10 dBm.");
#endif
                power = 10;
            }

            if (password != null)
            {
                if (password.HexString.Equals("00000000"))
                {
#if DEBUG
                    log.Warn("RFID: When using no password, set it to null instead of '00000000'.");
#endif
                    password = null;
                }
            }

            List<byte> dataByteList = new List<byte>();
            dataByteList.Add(0);
            RfidFlag flags = 0;

            // add select statement if necessary
            if (select != null)
            {
                flags |= RfidFlag.Select;

                if (select.Invert)
                {
                    flags |= RfidFlag.SelectInvert;
                }
                dataByteList.Add(0);
                dataByteList.Add((byte)((int)select.MemoryBank));
                dataByteList.Add((byte)((select.AddressInBits >> 8) & 0xFF));
                dataByteList.Add((byte)(select.AddressInBits & 0xFF));
                dataByteList.Add((byte)((select.DataLengthInBits >> 8) & 0xFF));
                dataByteList.Add((byte)(select.DataLengthInBits & 0xFF));

                byte[] selectData = select.Data;
                for (int i = 0; i < selectData.Length; i++)
                {
                    dataByteList.Add(selectData[i]);
                }

                dataByteList[1] = (byte)(selectData.Length + 5);
            }

            // add password if necessary
            if (password != null)
            {
                flags |= RfidFlag.Password;

                byte[] passwordData = EpcConvert.ConvertToDataFromHexString(password.HexString);
                for (int i = 0; i < passwordData.Length; i++)
                {
                    dataByteList.Add(passwordData[i]);
                }
            }

            dataByteList[0] = (byte)((int)flags);

            dataByteList.Add((byte)((timeout >> 8) & 0xFF));
            dataByteList.Add((byte)(timeout & 0xFF));
            dataByteList.Add((byte)(power));

            return dataByteList.ToArray();
        }
        #endregion //Management

        #region API calls

        /// <summary>
        /// Checks if the handheld is connected or not.
        /// </summary>
        /// <returns>true if handheld is connected</returns>
        public bool HandheldIsConnected()
        {
            return port.Connected;
        }

        /// <summary>
        /// Returns the current battery status in percentage.
        /// </summary>
        public void HandheldGetBatteryPercentage()
        {
            communicationQueue.SendCommand(Command.GeneralGetBatteryPercentage, new byte[0]);
        }

        /// <summary>
        /// Get information on the handheld. 
        /// </summary>
        /// <returns>DeviceInfo class object with three parameters: serial number, firmware and hardware version.</returns>
        public DeviceInfo HandheldGetInformation()
        {
            return DeviceInfo;
        }

        private void HandheldGetDeviceDetails()
        {
            communicationQueue.SendCommand(Command.GeneralGetDeviceDetails, new byte[0]);
        }

        /// <summary>
        ///  Will let the handheld kill the bluetooth connection. Can be used to pair the handheld with another device.
        /// </summary>
        public void HandheldDisconnect()
        {
            communicationQueue.SendCommand(Command.GeneralDisconnect, new byte[0]);
        }

        #region User interaction

        /// <summary>
        /// The beeper on the handheld will play the defined tune.
        /// </summary>
        /// <param name="tune"></param>
        public void UIPlay(UITune tune)
        {
            byte[] data = new byte[1];
            data[0] = (byte)tune;

            communicationQueue.SendCommand(Command.UIBeeperTune, data);
        }

        /// <summary>
        /// The vibration motor on the handheld will vibrate.
        /// </summary>
        public void UIVibrate()
        {
            communicationQueue.SendCommand(Command.UIVibrationVibrate, new byte[0]);
        }

        /// <summary>
        ///  The vibration motor on the handheld will vibrate for a specified duration.
        /// </summary>
        /// <param name="duration">vibrate duration in tens of a second</param>
        public void UIVibrate(int duration)
        {
            byte[] data = new byte[1];
            data[0] = (byte)(duration & 0xFF);

            communicationQueue.SendCommand(Command.UIVibrationVibrate, data);
        }

        /// <summary>
        /// The vibration motor on the handheld will vibrate and the beeper on the handheld will play the defined tune.
        /// </summary>
        /// <param name="tune"></param>
        public void UIVibrateTune(UITune tune)
        {
            byte[] data = new byte[1];
            data[0] = (byte)tune;

            communicationQueue.SendCommand(Command.UIVibrationTune, data);
        }

        /// <summary>
        ///  Let the white status led on the handheld blink or not.
        /// </summary>
        /// <param name="blinking">blinking enabled</param>
        public void UILedSetBlinking(bool blinking)
        {
            byte[] data = new byte[1];
            data[0] = (byte)((blinking) ? 2 : 1);
            communicationQueue.SendCommand(Command.UILedSetStatusLed, data);
        }

        #endregion

        /// <summary>
        /// Gets the supported regions by the handheld reader.
        /// </summary>
        public void RfidGetSupportedRegions()
        {
            communicationQueue.SendCommand(Command.RFIDGetSupportedRegions, new byte[0]);
        }

        /// <summary>
        /// Sets the region that the handheld reader operates in. The supported regions can be obtained by using +(void)rfidGetSupportedRegions.
        /// </summary>
        /// <param name="region">RfidRegion to set</param>
        public void RfidSetRegion(RfidRegion region)
        {
            byte[] data = new byte[1];
            data[0] = (byte)((int)region);

            communicationQueue.SendCommand(Command.RFIDSetRegion, data);
            currentRegion = region;
        }

        /// <summary>
        /// Set frequency hop table.
        /// </summary>
        /// <param name="frequencies">frequencies array with frequencies (NSNumber containing NHRFIDFrequency)</param>
        public void RfidSetFrequencyHopTable(List<RfidFrequency> frequencies)
        {
            if (currentRegion == RfidRegion.Undefined)
            {
#if DEBUG
                log.Warn("Region is undefined. Therefore it's not possible to set hop frequencies.");
#endif
            }

            List<byte> data = new List<byte>();
            // TODO: Add frequency hop table stuff for non-European frequencies.
            foreach (RfidFrequency frequencyNumber in frequencies)
            {
                int frequency = (int)frequencyNumber;

                if (currentRegion == RfidRegion.Europe)
                {
                    if ((frequency < 865700) || (frequency > 867500))
                    {
#if DEBUG
                        log.Warn("Frequency " + frequency.ToString() + " outside allowable frequency range for current region.");
#endif
                    }
                    else
                    {
                        byte[] frequencyBytes = new byte[4];
                        frequencyBytes[0] = (byte)((frequency >> 24) & 0xFF);
                        frequencyBytes[1] = (byte)((frequency >> 16) & 0xFF);
                        frequencyBytes[2] = (byte)((frequency >> 8) & 0xFF);
                        frequencyBytes[3] = (byte)(frequency & 0xFF);
                        data.AddRange(frequencyBytes);
                    }
                }
            }

            if (data.Count > 0)
            {
                communicationQueue.SendCommand(Command.RFIDSetFrequencyHopTable, data.ToArray());
            }
            else
            {
#if DEBUG
                log.Warn("Frequency hop table is empty.");
#endif
            }
        }

        /// <summary>
        /// Sets the session for the EPC Gen2 protocol.
        /// </summary>
        /// <param name="session">session session (S0, S1, S2 or S3)</param>
        public void RfidSetSession(RfidEpcSession session)
        {
            byte[] data = new byte[2];
            data[0] = (byte)((int)RfidEpcParameters.Session);
            data[1] = (byte)((int)session);

            communicationQueue.SendCommand(Command.RFIDSetEPCParameters, data);
        }

        /// <summary>
        /// Sets the target for the EPC Gen2 protocol.
        /// </summary>
        /// <param name="target">target target (A or B)</param>
        /// <param name="toggle">toggle toggle between A and B</param>
        public void RfidSetTarget(RfidEpcTarget target, bool toggle)
        {
            byte[] data = new byte[3];

            data[0] = (byte)((int)RfidEpcParameters.Target);
            data[1] = (byte)((toggle) ? 0x01 : 0x00);
            data[2] = (byte)((int)target);

            communicationQueue.SendCommand(Command.RFIDSetEPCParameters, data);
        }

        /// <summary>
        /// Sets the miller value for the EPC Gen2 protocol.
        /// </summary>
        /// <param name="mValue">mValue miller value</param>
        public void RfidSetMValue(RfidEpcMValue mValue)
        {
            byte[] data = new byte[2];
            data[0] = (byte)((int)RfidEpcParameters.MValue);
            data[1] = (byte)((int)mValue);

            communicationQueue.SendCommand(Command.RFIDSetEPCParameters, data);
        }

        /// <summary>
        /// Sets the Q value for the EPC Gen2 protocol.
        /// </summary>
        /// <param name="qValue">qValue the Q value</param>
        /// <param name="dynamic">dynamic whether the Q value is dynamically adjusted or not</param>
        public void RfidSetQValue(int qValue, bool dynamic)
        {
            byte[] data = new byte[3];
            data[0] = (byte)((int)RfidEpcParameters.QValue);
            data[1] = (byte)((dynamic) ? 0x01 : 0x00);
            data[2] = (byte)(qValue);

            communicationQueue.SendCommand(Command.RFIDSetEPCParameters, data);
        }

        /// <summary>
        /// Will read RFID tags in the field, for a specified duration, with a specific power and with a select to select a subset of tags.
        /// </summary>
        /// <param name="timeout">timeout in milliseconds</param>
        /// <param name="power">the power, possible values range from 10 to 23 dBm</param>
        /// <param name="select">select statement, can also be set to null</param>
        public void RfidReadEpcs(int timeout, int power, Select select)
        {
            byte[] data = RfidPrepareStandardParameters(timeout, power, select, null);
            communicationQueue.SendCommand(Command.RFIDReadEPC, data);
        }

        /// <summary>
        /// Will read specific data from a RFID tag.
        /// </summary>
        /// <param name="memoryBank">the memory bank to read data from</param>
        /// <param name="addressInWords">the address to read data from (in words)</param>
        /// <param name="lengthInWords">the length of the data to read (in words)</param>
        /// <param name="timeout">timeout in milliseconds</param>
        /// <param name="power">the power, possible values range from 10 to 23 dBm</param>
        /// <param name="select">select statement, can also be set to null</param>
        /// <param name="password">password, can also be set to null</param>
        public void RfidReadData(RfidEpcMemoryBank memoryBank, int addressInWords, int lengthInWords, int timeout, int power, Select select, Password password)
        {
            List<byte> dataArray = new List<byte>();

            dataArray.AddRange(RfidPrepareStandardParameters(timeout, power, select, password));
            dataArray.Add((byte)((int)memoryBank)); // bank
            dataArray.Add((byte)((addressInWords >> 8) & 0xFF)); // address in words, high byte
            dataArray.Add((byte)(addressInWords & 0xFF)); // address in words, low byte
            dataArray.Add((byte)(lengthInWords & 0xFF)); // length in words

            communicationQueue.SendCommand(Command.RFIDReadData, dataArray.ToArray());
        }

        /// <summary>
        /// Writes a new EPC code to a tag.
        /// </summary>
        /// <param name="epc">new EPC code</param>
        public void RfidWriteEpc(EpcCode epc)
        {
            if (epc.HexString.Length % 4 > 0)
            {
#if DEBUG
                log.Error("RFID: EPC length not a multiple of four characters.");
#endif
                return;
            }
            List<byte> dataArray = new List<byte>();
            //add FDU
            dataArray.Add(0x00);
            dataArray.Add(0x00);

            byte[] epcData = epc.BinaryData;
            for (int i = 0; i < epcData.Length; i++)
            {
                dataArray.Add(epcData[i]);
            }

            communicationQueue.SendCommand(Command.RFIDWriteEPC, dataArray.ToArray());
        }

        /// <summary>
        /// Write data to a specified memory bank and address in a tag.
        /// </summary>
        /// <param name="dataToWrite">the data to write.</param>
        /// <param name="memoryBank">the memory bank where to write the data to</param>
        /// <param name="addressInWords">the address in words where the data needs to be written</param>
        /// <param name="timeout">the address in words where the data needs to be written</param>
        /// <param name="power">the power, possible values range from 10 to 23 dBm</param>
        /// <param name="select">select statement, can also be set to null</param>
        /// <param name="password">password, can also be set to null</param>
        public void RfidWriteData(byte[] dataToWrite, RfidEpcMemoryBank memoryBank, int addressInWords, int timeout, int power, Select select, Password password)
        {
            if (dataToWrite.Length % 4 > 0)
            {
#if DEBUG
                log.Error("Data length is not a multiple of four characters.");
#endif
                return;
            }

            List<byte> dataArray = new List<byte>();

            dataArray.AddRange(RfidPrepareStandardParameters(timeout, power, select, password));
            int dataToWriteLengthInWords = dataToWrite.Length / 4;
            dataArray.Add((byte)((int)memoryBank)); // bank
            dataArray.Add((byte)((addressInWords >> 8) & 0xFF)); // address in words, high byte
            dataArray.Add((byte)(addressInWords & 0xFF)); // address in words, low byte - includes PC word
            dataArray.Add((byte)(dataToWriteLengthInWords * 2)); // length of the data in bytes

            for (int i = 0; i < dataToWrite.Length; i++)
            {
                dataArray.Add(dataToWrite[i]);
            }

            communicationQueue.SendCommand(Command.RFIDWrite, dataArray.ToArray());
        }

        private void RfidLock(RFIDLock maskBits, RFIDLock lockBits, int timeout, int power, Select select, Password password)
        {
            if (password == null)
            {
#if DEBUG
                log.Error("RFID: Password cannot be zero when locking a tag.");
#endif
                return;
            }

            List<byte> dataArray = new List<byte>();

            dataArray.AddRange(RfidPrepareStandardParameters(timeout, power, select, password));
            dataArray.Add((byte)(((int)maskBits >> 8) & 0xFF)); // bank
            dataArray.Add((byte)((int)maskBits & 0xFF)); // address in words, high byte
            dataArray.Add((byte)(((int)lockBits >> 8) & 0xFF)); // address in words, low byte - includes PC word
            dataArray.Add((byte)((int)lockBits & 0xFF)); // length of the new EPC in bytes + 2 for the PC word

            communicationQueue.SendCommand(Command.RFIDLock, dataArray.ToArray());
        }

        private void RfidKill(int timeout, int power, Select select, Password password)
        {
            if (password == null)
            {
#if DEBUG
                log.Error("RFID: Password cannot be zero when trying to kill a tag.");
#endif
                return;
            }
            byte[] data = RfidPrepareStandardParameters(timeout, power, select, password);

            communicationQueue.SendCommand(Command.RFIDKill, data);
        }

        /// <summary>
        /// Enables Continuous Wave (CW) signal, used for testing and measurement purposes.
        /// </summary>
        /// <param name="enabled">CW enabled</param>
        /// <param name="antenna">select antenna one or two</param>
        public void RfidSetCWSignal(bool enabled, int antenna)
        {
            byte[] data = new byte[2];
            data[0] = (byte)((enabled) ? 1 : 0);
            data[1] = (byte)((antenna == 0) ? 0 : 1);

            communicationQueue.SendCommand(Command.RFIDSetCWSignal, data);
        }

        #region Barcode

        /// <summary>
        /// 	Check to see if the handheld reader has a barcode reader on-board.
        /// </summary>
        /// <returns>true if there is a barcode reader</returns>
        public bool BarcodeHasReader()
        {
            return true;// (DeviceInfo.SerialNumber == 9943978);
        }

        /// <summary>
        /// Enables or disable a certain type of barcode.
        /// </summary>
        /// <param name="barcodeType">NHBarcodeType</param>
        /// <param name="enabled">enabled or disabled</param>
        public void BarcodeSetStatusForType(BarcodeType barcodeType, bool enabled)
        {
            byte[] data = new byte[2];
            data[0] = (byte)((int)barcodeType);
            data[1] = (byte)((enabled) ? 1 : 0);

            communicationQueue.SendCommand(Command.BarcodeSetType, data);
        }

        /// <summary>
        /// Try to read a barcode for a defined number of seconds.
        /// </summary>
        /// <param name="duration">duration in seconds</param>
        public void BarcodeTryDecode(int duration)
        {
            byte[] data = new byte[1] { (byte)(duration * 10) };
            communicationQueue.SendCommand(Command.BarcodeTryDecode, data);
        }

        #endregion //Barcode


        # region NCF

        /// <summary>
        /// The handheld will try to read a NFC card.
        /// </summary>
        public void NfcTryToReadCard()
        {
            communicationQueue.SendCommand(Command.NFCTryToReadCard, new byte[0]);
        }

        #endregion

        #endregion //API calls


        #region Packet sending/receiving

        private void ParseResponse(Response responseCode, byte[] dataInPacket)
        {
            switch (responseCode)
            {
                case Response.Error:
                    ParsePacketError(dataInPacket);
                    break;
                case Response.GeneralBatteryPercentage:
                    ParsePacketGeneralBatteryPercentage(dataInPacket);
                    break;
                case Response.GeneralDeviceDetails:
                    ParsePacketGeneralDeviceDetails(dataInPacket);
                    break;
                case Response.GeneralLowBattery:
                    ParsePacketGeneralLowBattery();
                    break;
                case Response.GeneralEmptyBattery:
                    ParsePacketGeneralEmptyBattery();
                    break;
                case Response.GeneralDeviceCharging:
                    ParsePacketGeneralDeviceCharging();
                    break;
                case Response.RfidTagCount:
                    ParsePacketRfidTagCount(dataInPacket);
                    break;
                case Response.RfidTags:
                    ParsePacketRfidTags(dataInPacket);
                    break;
                case Response.RfidReadData:
                    ParsePacketRfidReadData(dataInPacket);
                    break;
                case Response.RfidWrite:
                    ParsePacketRfidWrite(dataInPacket);
                    break;
                case Response.RfidSupportedRegions:
                    ParsePacketRfidSupportedRegions(dataInPacket);
                    break;
                case Response.RFIDWriteEPC:
                    ParsePacketRfidWriteEPC(dataInPacket);
                    break;
                case Response.BarcodeBarcodeRead:
                    ParsePacketBarcodeRead(dataInPacket);
                    break;
                case Response.BarcodeNothingRead:
                    ParsePacketBarcodeFailed();
                    break;
                case Response.UIButtonPush:
                    ParsePacketUiButtonPush();
                    break;
                case Response.UIAccelerometerShake:
                    ParsePacketUiAccelerometerShake();
                    break;
                case Response.NfcNoCardFound:
                    ParsePacketNfcNoCardFound();
                    break;
                case Response.NfcCardFound:
                    ParsePacketNfcCardFound(dataInPacket);
                    break;
                default:
                    throw new NotSupportedException();
            }
        }


        private void ParsePacketError(byte[] data)
        {
            Error errorCode = (Error)Convert.ToInt32(data[0]);
            string error = string.Empty;

            switch (errorCode)
            {
                case Error.Crc:
                    error = "Handheld: Error in CRC of transmitted packet";
                    break;
                case Error.WrongNumberOfBytes:
                    error = "Handheld: Wrong number of bytes in tranmitted packet.";
                    break;
                case Error.RfidAnalogFrontendNotEnabled:
                    error = "RFID: Analog frontend not enabled";
                    break;
                case Error.RfidTemperatureExceedLimits:
                    error = "RFID: Temperature exceed limits.";
                    break;
                case Error.RfidHighReturnLoss:
                    error = "RFID: High return loss on transmitter.";
                    break;
                case Error.BarcodeNotAvailable:
                    error = "Barcode: Not available.";
                    break;
            }

#if DEBUG
            log.Error("Error " + error);
#endif

            if (error != string.Empty)
            {
                if (OnSessionHandheldReceiveError != null)
                {
                    OnSessionHandheldReceiveError(this, new HandheldReceiveErrorEventArgs(error));
                }
#if DEBUG
                log.Error(error);
#endif
            }
        }

        private void ParsePacketGeneralBatteryPercentage(byte[] data)
        {
            int percentage = Convert.ToInt32(data[0]);
            if (OnSessionHandheldReceiveBatteryPercentage != null)
            {
                OnSessionHandheldReceiveBatteryPercentage(this, new HandheldReceiveBatteryPercentageEventArgs(percentage));
            }
        }

        private void ParsePacketGeneralDeviceDetails(byte[] data)
        {
            if (data.Length == 18)
            {
                DeviceInfo.NedapProductionCode =
                    //year
                    (Convert.ToChar(data[0])).ToString()
                    //month
                    + (Convert.ToChar(data[1])).ToString()
                    //day
                    + (Convert.ToChar(data[2])).ToString() + (Convert.ToChar(data[3])).ToString() + " "
                    //modification char
                    + (Convert.ToChar(data[4])).ToString() + " "
                    //serial number
                    + (Convert.ToChar(data[5])).ToString() + (Convert.ToChar(data[6])).ToString() + (Convert.ToChar(data[7])).ToString();
                DeviceInfo.SerialNumber = (Convert.ToChar(data[5])).ToString() + (Convert.ToChar(data[6])).ToString() + (Convert.ToChar(data[7])).ToString();
                DeviceInfo.ModelNumber = ((uint)((data[8] << 24) | (data[9] << 16) | (data[10] << 8) | data[11])).ToString();
                DeviceInfo.FirmwareVersion = (Convert.ToInt16(data[12])).ToString() + "." + (Convert.ToInt16(data[13])).ToString() + "." + (Convert.ToInt16(data[14])).ToString();
                DeviceInfo.HardwareVersion = (Convert.ToInt16(data[15])).ToString() + "." + (Convert.ToInt16(data[16])).ToString() + "." + (Convert.ToInt16(data[17])).ToString();
            }
        }

        private void ParsePacketGeneralLowBattery()
        {
            if (OnSessionHandheldReceiveLowBatteryWarning != null)
            {
                OnSessionHandheldReceiveLowBatteryWarning(this, EventArgs.Empty);
            }
        }

        private void ParsePacketGeneralEmptyBattery()
        {
            if (OnSessionHandheldReceiveEmptyBatteryStatus != null)
            {
                OnSessionHandheldReceiveEmptyBatteryStatus(this, EventArgs.Empty);
            }
        }

        private void ParsePacketGeneralDeviceCharging()
        {
            if (OnSessionHandheldReceiveChargingStatus != null)
            {
                OnSessionHandheldReceiveChargingStatus(this, EventArgs.Empty);
            }
        }

        private void ParsePacketRfidTagCount(byte[] data)
        {
            if (data[0] == 0)
            {
                if (OnSessionRfidFinishReadCycle != null)
                {
                    OnSessionRfidFinishReadCycle(this, EventArgs.Empty);
                }
            }
        }

        private void ParsePacketRfidTags(byte[] data)
        {
            Debug.Assert(data.Length >= 2, "Unexpected RfidTags packet length");

            bool lastPacket = Convert.ToBoolean(data[0]);
            uint numberOfTags = Convert.ToUInt32(data[1]);

            if (numberOfTags == 0)
            {
                if (OnSessionRfidFinishReadCycle != null)
                {
                    OnSessionRfidFinishReadCycle(this, EventArgs.Empty);
                }
                return;
            }

            int index = 2;
            for (int i = 0; i < numberOfTags; i++)
            {
                Debug.Assert((data.Length >= index + 2), "Unexpected format of RfidTags packet.");
                int readCount = data[index++];
                int rssi = data[index++];

                int epcLength = ((data[index] << 8) | data[index + 1]) / 8;
                index += 4;

                if (index + epcLength - 4 > data.Length)
                {
#if DEBUG
                    log.Error("Error in data packet while parsing EPC tags.");
#endif
                    return;
                }

                Debug.Assert((data.Length >= epcLength - 4), "Bad length of tag in RfidTags packet.");

                byte[] epc = new byte[epcLength - 4];
                Buffer.BlockCopy(data, index, epc, 0, epcLength - 4);

                EpcCode tag = new EpcCode(epc);
                EpcObservation tagObservation = new EpcObservation(tag, DateTime.Now, rssi, readCount);

                index += epcLength - 4;

                // skip CRC of tag
                index += 2;

                if (OnSessionRfidReadTag != null)
                {
                    OnSessionRfidReadTag(this, new RfidReadTagEventArgs(tagObservation));
                }
            }

            if (lastPacket)
            {
                if (OnSessionRfidFinishReadCycle != null)
                {
                    OnSessionRfidFinishReadCycle(this, EventArgs.Empty);
                }
            }
        }

        private void ParsePacketRfidReadData(byte[] data)
        {
            int count = Convert.ToInt32(data[0]);

            bool moreTagsAvailable = false;
            bool success = Convert.ToBoolean(data[1]);

            if (success)
            {
                byte[] readData = new byte[data.Length - 2];
                Buffer.BlockCopy(data, 2, readData, 0, data.Length - 2);

                if (count > 1)
                {
                    moreTagsAvailable = true;
                }
                if (OnSessionRfidReadData != null)
                {
                    OnSessionRfidReadData(this, new RfidReadDataEventArgs(readData, moreTagsAvailable));
                }
            }
            else
            {
                if (count > 0)
                {
                    moreTagsAvailable = true;
                }
                if (OnSessionRfidFailButTagsAvailable != null)
                {
                    OnSessionRfidFailButTagsAvailable(this, new RfidFailButTagsAvailableEventArgs(moreTagsAvailable));
                }
            }
        }

        private void ParsePacketRfidWrite(byte[] data)
        {
            // get write status

            int count = data[0];
            bool moreTagsAvailable = false;

            bool success = Convert.ToBoolean(data[1]);

            if (success)
            {
                if (OnSessionRfidWriteDataMoreTagsAvailable != null)
                {
                    OnSessionRfidWriteDataMoreTagsAvailable(this, new RfidWriteDataMoreTagsAvailableEventArgs(moreTagsAvailable));
                }
            }
            else
            {
                if (count > 0)
                {
                    moreTagsAvailable = true;
                }
                if (OnSessionRfidFailButTagsAvailable != null)
                {
                    OnSessionRfidFailButTagsAvailable(this, new RfidFailButTagsAvailableEventArgs(moreTagsAvailable));
                }
            }
        }

        private void ParsePacketRfidSupportedRegions(byte[] data)
        {
            // get write status
            int regionFlags = Convert.ToInt32(data[0]);
            if (OnSessionRfidReceiveSupportedRegions != null)
            {
                OnSessionRfidReceiveSupportedRegions(this, new RfidReceiveSupportedRegionsEventArgs(regionFlags));
            }
        }


        private void ParsePacketRfidWriteEPC(byte[] data)
        {
            RFIDWriteEPCResult result = (RFIDWriteEPCResult)data[0];

            switch (result)
            {
                case RFIDWriteEPCResult.RFIDWriteEPCResultNoTagsInField:
                    if (OnSessionRfidWriteNoTagsInField != null)
                    {
                        OnSessionRfidWriteNoTagsInField(this, EventArgs.Empty);
                    }
                    break;
                case RFIDWriteEPCResult.RFIDWriteEPCResultTooMuchTagsInField:
                    if (OnSessionRfidWriteTooMuchTagsInField != null)
                    {
                        OnSessionRfidWriteTooMuchTagsInField(this, EventArgs.Empty);
                    }
                    break;
                case RFIDWriteEPCResult.RFIDWriteEPCResultWriteToTagFailed:
                    if (OnSessionRfidWriteToTagFailed != null)
                    {
                        OnSessionRfidWriteToTagFailed(this, EventArgs.Empty);
                    }
                    break;
                case RFIDWriteEPCResult.RFIDWriteEPCResultWriteSuccess:
                    if (OnSessionRfidWriteSuccess != null)
                    {
                        OnSessionRfidWriteSuccess(this, EventArgs.Empty);
                    }
                    break;
                default:
                    break;
            }

        }

        private void ParsePacketBarcodeRead(byte[] data)
        {
            // get barcode type
            BarcodeType barcodeType = (BarcodeType)Convert.ToInt32(data[0]);
            String barcode = Encoding.ASCII.GetString(data, 1, data.Length - 1);
            if (OnSessionBarcodeRead != null)
            {
                OnSessionBarcodeRead(this, new RfidReadABarcodeEventArgs(barcode, barcodeType));
            }
        }

        private void ParsePacketBarcodeFailed()
        {
            if (OnSessionBarcodeNotRead != null)
            {
                OnSessionBarcodeNotRead(this, EventArgs.Empty);
            }
        }

        private void ParsePacketUiButtonPush()
        {
            // generate notification
            if (OnSessionUIPushButton != null)
            {
                OnSessionUIPushButton(this, EventArgs.Empty);
            }
        }

        private void ParsePacketUiAccelerometerShake()
        {
            // generate notification
            if (OnSessionUIShake != null)
            {
                OnSessionUIShake(this, EventArgs.Empty);
            }
        }

        private void ParsePacketNfcNoCardFound()
        {
            if (OnSessionNfcCardNotFound != null)
            {
                OnSessionNfcCardNotFound(this, EventArgs.Empty);
            }
        }

        private void ParsePacketNfcCardFound(byte[] data)
        {
            // get barcode type
            int cardNumber = Convert.ToInt32((data[0] << 24) | (data[1] << 16) | (data[2] << 8) | data[3]);
#if DEBUG
            log.Info(String.Format("cardNumber: {0:x4}", cardNumber));
#endif

            if (OnSessionNfcCardRead != null)
            {
                OnSessionNfcCardRead(this, new RfidReadCardEventArgs(cardNumber));
            }
        }

        #endregion //Packet sending/receiving
    }
}
