//
//  EPCCode.m
//  EPCGlobalTDS
//
//  Created by Jasper Kremer on 12-07-10.
//  Copyright 2010 N.V. Nederlandsche Apparatenfabriek "Nedap". All rights reserved.
//

#import "EPCCode.h"
#import "EPCConvert.h"
#import "EPCTDTEngine.h"

@implementation EPCCode

+ (id)codeWithBinaryData:(NSData *)someData
{
	return [[[self alloc] initWithBinaryData:someData] autorelease];
}

+ (id)codeWithString:(NSString *)aString
{
	return [[[self alloc] initWithString:aString] autorelease];
}

+ (id)codeWithGTIN14:(NSString *)GTIN14 serialNumber:(NSNumber *)serial companyPrefixLength:(NSUInteger)companyPrefixLength
{
	return [[[self alloc] initWithGTIN14:GTIN14 serialNumber:serial companyPrefixLength:companyPrefixLength] autorelease];
}

- (id)initWithBinaryData:(NSData *)someData
{
	self = [super init];
	if (nil != self) {
		binaryData = [someData retain];
        hash = 0;
	}
    return self;
}

- (id)initWithString:(NSString *)aString
{
	self = [super init];
	if (nil != self) {
		// try to parse with TDT
		NSString *binary = [EPCTDTEngine translate:aString withParameterList:nil outputFormat:@"BINARY"];
        hash = 0;
		
		if (binary != nil)
		{
			hexString = [[EPCConvert convertToHexStringFromBinaryString:binary] retain];
			binaryData = [[EPCConvert convertToDataFromHexString:hexString] retain];
		}
		else 
		{
			// TODO: check if it's an HEX string
			
			binaryData = [EPCConvert convertToDataFromHexString:aString];
			[binaryData retain];
		}
	}
	return self;
}

- (id)initWithGTIN14:(NSString *)GTIN14 serialNumber:(NSNumber *)serial companyPrefixLength:(NSUInteger)companyPrefixLength;
{
	self = [super init];
	if (nil != self) {
		// create legacy encoding
		legacy = [[NSString alloc] initWithFormat:@"gtin=%@;serial=%llu", GTIN14, [serial unsignedLongLongValue]];

		// create parameter list
		NSString *parameterList = [NSString stringWithFormat:@"filter=1;taglength=96;gs1companyprefixlength=%u", companyPrefixLength];
		
		NSString *binary = [EPCTDTEngine translate:legacy withParameterList:parameterList outputFormat:@"BINARY"];
		hexString = [[EPCConvert convertToHexStringFromBinaryString:binary] retain];
		binaryData = [[EPCConvert convertToDataFromHexString:hexString] retain];
        hash = 0;
	}
	return self;
}

- (NSUInteger)lengthInBits
{
	return [binaryData length] * 8;
}

- (BOOL)isEqual:(id)anObject;
{
    if (anObject == self)
        return YES;
    if (!anObject || ![anObject isKindOfClass:[self class]])
        return NO;
    return [self isEqualToEPCCode:anObject];
}

- (BOOL)isEqualToEPCCode:(EPCCode *)anEPCCode;
{
	return [binaryData isEqual:anEPCCode.binaryData];
}

// See http://www.sersc.org/journals/IJUNESST/vol5_no2/5.pdf for more detail on EPC hash methods
- (NSUInteger)hash
{
    if (hash == 0) {
        const unsigned char* buffer = [binaryData bytes];
        NSUInteger size =  [binaryData length];
        
        hash = buffer[size - 2] * 4194304 + buffer[size - 8] * 2097152 + buffer[size - 4] * 1048576 + buffer[size - 6] * 32768 + buffer[size - 5] * 16384 + buffer[size - 7] * 8192 + buffer[size - 3] * 4096 + buffer[size - 9] * 128 + buffer[size - 1];
        
        // max NSUInteger value = 2147483647 (== NSNotFound)
        //       max hash value = 1894809856
        
//        if (size >= 12) {
//        }
//        // !Hash based on serial number (last 2 bytes) alone is not safe enough! For example:
//        // hash =    for                                      [  ,  ,   ,  ,   ,XXX,XXX,   , , ,X,XX]
//        // hash = 63 for urn:epc:id:sgtin:2004452.692835.63 = [48,52,122,87,146,164,152,192,0,0,0,63]
//        // hash = 63 for urn:epc:id:sgtin:2004451.003988.63 = [48,52,122,87,140,  3,229,  0,0,0,0,63]
//        else {
//            hash = buffer[size - 2] * 256 + buffer[size - 1];
//        }
    }
    return hash;
}

- (NSData *)binaryData
{
	return binaryData;
}

- (NSString *)binaryString
{
	if (binaryString == nil)
	{
		binaryString = [EPCConvert convertToBinaryStringFromData:binaryData];
		[binaryString retain];
	}

	return binaryString;
}

- (NSString *)hexString
{	
	if (hexString == nil)
	{
		hexString = [EPCConvert convertToHexStringFromBinaryString:self.binaryString];
		[hexString retain];
	}
	
	return hexString;
}

- (NSString *)tagEncoding
{
	if (tagEncoding == nil)
		tagEncoding = [EPCTDTEngine translateData:binaryData withParameterList:@"" outputFormat:@"TAG_ENCODING"];
	
	if (tagEncoding == nil)
	{
		tagEncoding = [[self hexString] copy];
	}
	else 
	{
		[tagEncoding retain];
	}
	
	return tagEncoding;
}

- (NSString *)pureIdentityURI
{
	if (pureIdentityURI == nil)
		pureIdentityURI = [EPCTDTEngine translateData:binaryData withParameterList:@"" outputFormat:@"PURE_IDENTITY"];
	
	if (pureIdentityURI == nil)
	{
		pureIdentityURI = [[self hexString] copy];
	}
	else 
	{
		[pureIdentityURI retain];
	}
	
	return pureIdentityURI;
}

- (NSString *)legacy
{
	if (legacy == nil)
		legacy = [EPCTDTEngine translateData:binaryData withParameterList:@"" outputFormat:@"LEGACY"];
	
	if (legacy == nil)
	{
		legacy = [[self hexString] copy];
	}
	else 
	{
		[legacy retain];
	}
	
	return legacy;
}

- (NSString *)onsHostname
{
	if (onsHostname == nil)
		onsHostname = [EPCTDTEngine translateData:binaryData withParameterList:@"" outputFormat:@"ONS_HOSTNAME"];
	
	if (onsHostname == nil)
	{
		onsHostname = [[self hexString] copy];
	}
	else 
	{
		[onsHostname retain];
	}
	
	return onsHostname;
}

- (NSString *)description {
	return self.pureIdentityURI;
}

- (void) dealloc {
    [binaryData release];
    [binaryString release];
	[hexString release];
	[tagEncoding release];
    [pureIdentityURI release];
    [legacy release];
	[onsHostname release];
    
	binaryData = nil;
	binaryString = nil;
	hexString = nil;
	tagEncoding = nil;
	pureIdentityURI = nil;
	legacy = nil;
	onsHostname = nil;
    
	[super dealloc];
}

@end
