//
//  EPCTDTEngine.m
//  EPCGlobalTDS
//
//  Created by Danny Haak on 19-04-10.
//  Copyright 2010 N.V. Nederlandsche Apparatenfabriek "Nedap". All rights reserved.
//

#import "EPCConvert.h"
#import "EPCTDTEngine.h"
#import "EPCTagDataTranslationStandard.h"
#import "EPCScheme.h"
#import "EPCLevel.h"
#import "EPCOption.h"
#import "EPCField.h"
#import "EPCRule.h"
#import "EPCPrefixMatch.h"

static EPCTagDataTranslationStandard *epcTagDataTranslation;
static NSMutableArray *prefixMatchArray;
static EPCScheme *currentScheme;
static EPCLevel *currentLevel;
static EPCOption *currentOption;

@interface EPCTDTEngine()

+ (bool)loadEPCTagDataTranslation:(NSString *)uri;
+ (void)buildPrefixTree;
+ (EPCPrefixMatch *)getPrefixMatch:(NSString *)input parameters:(NSMutableDictionary *)parameters;
+ (NSMutableDictionary *)parseParameterList:(NSString *)parameterList;
+ (EPCOption *)getOption:(NSString *)input parameters:(NSMutableDictionary *)parameters level:(EPCLevel *)level;
+ (void)processRule:(EPCRule *)rule parameters:(NSMutableDictionary *)parameters;
+ (NSInteger)getIntValue:(NSString *)fieldName parameters:(NSMutableDictionary *)parameters;
+ (EPCLevel *)findOutputLevelFromScheme:(EPCScheme *)scheme withString:(NSString *)outputFormat;
+ (EPCOption *)findOutputOptionFromLevel:(EPCLevel *)level withOption:(EPCOption *)option;
+ (NSString *)decimalToBinary:(NSString *)decimal;
+ (NSString *)binaryToDecimal:(NSString *)binary;
+ (NSString *)buildGrammer:(NSString *)grammerField parameters:(NSMutableDictionary *)parameters;
+ (void)binaryInputPadding:(NSMutableDictionary *)parameters option:(EPCOption *)option optionNonBinary:(EPCOption *)optionNonBinary;
+ (void)binaryOutputPadding:(NSMutableDictionary *)parameters option:(EPCOption *)option optionNonBinary:(EPCOption *)optionNonBinary;

@end

@implementation EPCTDTEngine

+ (EPCTDTEngine *)sharedSingleton 
{
	static EPCTDTEngine *sharedSingleton;
	
	@synchronized(self)	
	{
		if (!sharedSingleton)
			sharedSingleton = [[EPCTDTEngine alloc] init];
		
		return sharedSingleton;
	}
	
	return nil;
}

+ (void)initialize 
{
	// TODO: Replace integers with NSDecimalNumber. This is only an issue for >96 bit RFID represenations (e.g. SGTIN-196).
	// TODO: Verify maximum/minimum, etc.
	// TODO: Implement unit tests.
	
	NSLog(@"Initialize TDTEngine");
	
	epcTagDataTranslation = [[EPCTagDataTranslationStandard alloc] init];
	prefixMatchArray = [[NSMutableArray alloc] init];
	
	//[self loadEPCTagDataTranslation:@"GDTI-96"];
	//[self loadEPCTagDataTranslation:@"GDTI-113"];
	//[self loadEPCTagDataTranslation:@"GIAI-64"];
	//[self loadEPCTagDataTranslation:@"GIAI-96"];
	//[self loadEPCTagDataTranslation:@"GIAI-202"];
	//[self loadEPCTagDataTranslation:@"GID-96"];
	//[self loadEPCTagDataTranslation:@"GRAI-64"];
	//[self loadEPCTagDataTranslation:@"GRAI-96"];
	//[self loadEPCTagDataTranslation:@"GRAI-170"];
	//[self loadEPCTagDataTranslation:@"GSRN-96"];
	//[self loadEPCTagDataTranslation:@"SGLN-64"];
	[self loadEPCTagDataTranslation:@"SGLN-96"];
	//[self loadEPCTagDataTranslation:@"SGLN-195"];
	//[self loadEPCTagDataTranslation:@"SGTIN-64"];
	[self loadEPCTagDataTranslation:@"SGTIN-96"];
	//[self loadEPCTagDataTranslation:@"SGTIN-198"];
	//[self loadEPCTagDataTranslation:@"SSCC-64"];
	//[self loadEPCTagDataTranslation:@"SSCC-96"];
	//[self loadEPCTagDataTranslation:@"USDOD-64"];
	//[self loadEPCTagDataTranslation:@"USDOD-96"];
	
	[self buildPrefixTree];
	
}

+ (bool)loadEPCTagDataTranslation:(NSString *)uri 
{
	NSString *path = [[NSBundle bundleForClass:[self class]] pathForResource:uri ofType:@"xml"];
	
	if (path != nil)
	{
		NSURL *xmlURL = [NSURL fileURLWithPath:path];
		NSXMLParser *xmlParser = [[[NSXMLParser alloc] initWithContentsOfURL:xmlURL] autorelease];
		
		[xmlParser setDelegate:(id <NSXMLParserDelegate>)self];
		[xmlParser setShouldResolveExternalEntities:YES];
		
		BOOL succeeded = [xmlParser parse];
		
		return succeeded;
	}
	else 
	{
		NSLog(@"File not found: %@", path);
	}
		
	return NO;
}

+ (void)buildPrefixTree 
{
	NSArray *schemes = epcTagDataTranslation.schemes;
	
	for (EPCScheme *scheme in schemes)
	{
		for (EPCLevel *level in scheme.levels)
		{
			EPCPrefixMatch *prefixMatch = [[EPCPrefixMatch alloc] init];
			prefixMatch.prefix = level.prefixMatch;
			prefixMatch.scheme = scheme;
			prefixMatch.level = level;
			[prefixMatchArray addObject:prefixMatch];
			[prefixMatch release];
		}
	}
}

+ (BOOL)isValidEPC:(NSString *)binaryString
{
	// detect inbound pattern
	EPCPrefixMatch *prefixMatch = [self getPrefixMatch:binaryString parameters:nil];
	if (prefixMatch == nil)
    {
		return NO;
    }
    return YES;
}

/**
 * Translates given input following the EPC TDT standard
 * examples: 
 * - input = "urn:epc:id:sgtin:3327252.089333.141702337387", parameterList = "taglength=96;filter=0;gs1companyprefixlength=7" and outputFormat = "LEGACY"
 * >> returns gtin14 value "33272520893332"
 * 
 * @param input
 * @param parameterList
 * @outputFormat
 * @return translated string
 */
+ (NSString *)translate:(NSString *)input withParameterList:(NSString *)parameterList outputFormat:(NSString *)outputFormat 
{
	[input retain];
    
	NSString *returnString;
	
	// parse parameter list
	NSMutableDictionary *parameters = [self parseParameterList:parameterList];
		
	// detect inbound pattern
	EPCPrefixMatch *prefixMatch = [self getPrefixMatch:input parameters:parameters];
	if (prefixMatch == nil)
    {
        [input release];
		return nil;
    }

	EPCScheme *scheme = prefixMatch.scheme;
	EPCLevel *level = prefixMatch.level;
	
	// determine the option
	EPCOption *option = [self getOption:input parameters:parameters level:level];
	
	if (option == nil)
	{
#ifdef DEBUG
		NSLog(@"NedapEPCGlobalGen2: No matching inbound option.");
#endif
        [input release];
		return nil;
	}
	
	// find outbound level
	EPCLevel *levelOut = [self findOutputLevelFromScheme:scheme withString:outputFormat];
	if (levelOut == nil)
	{
#ifdef DEBUG
		NSLog(@"NedapEPCGlobalGen2: No matching outbound level.");
#endif
        [input release];
		return nil;
	}
		
	EPCOption *optionOut = [self findOutputOptionFromLevel:levelOut withOption:option];
	if (optionOut == nil)
	{
#ifdef DEBUG
		NSLog(@"NedapEPCGlobalGen2: No matching inbound option.");
#endif
        [input release];
		return nil;
	}
	
	// retrieve fields - using NSRegularExpression
    NSCharacterSet *blank = [NSCharacterSet whitespaceCharacterSet];
	NSRegularExpression *regularExpression = [NSRegularExpression regularExpressionWithPattern:option.pattern 
                                                                                       options:(NSRegularExpressionCaseInsensitive | NSRegularExpressionDotMatchesLineSeparators) 
                                                                                         error:nil];
    NSArray *regMatches = [regularExpression matchesInString:input 
                                                  options:NSMatchingReportCompletion 
                                                    range:NSMakeRange(0, input.length)];

    [option.fields enumerateObjectsUsingBlock:^(EPCField *fieldEnumerator, NSUInteger index, BOOL *stop) {
        NSUInteger sequence = fieldEnumerator.seq;
		NSString *name = fieldEnumerator.name;
        
        NSTextCheckingResult *result = [regMatches objectAtIndex:0];
        if (result != nil) {
            NSString *item = [[input substringWithRange:[result rangeAtIndex:sequence]] stringByTrimmingCharactersInSet:blank];
            [parameters setValue:item forKey:name];
        }
    }];
	
	// find non-binary level
	EPCLevel *levelNonBinary = [self findOutputLevelFromScheme:scheme withString:@"TAG_ENCODING"];
	EPCOption *optionNonBinary = [self findOutputOptionFromLevel:levelNonBinary withOption:option];
	
	// binary input padding
	[self binaryInputPadding:parameters option:option optionNonBinary:optionNonBinary];
	
	// extract rules
    [level.rules enumerateObjectsUsingBlock:^(EPCRule *ruleEnumerator, NSUInteger index, BOOL *stop) {
		if ([ruleEnumerator.type isEqualToString:@"EXTRACT"]) {
			[self processRule:ruleEnumerator parameters:parameters];
		}
    }];
	
	// format rules
    [levelOut.rules enumerateObjectsUsingBlock:^(EPCRule *ruleEnumerator, NSUInteger index, BOOL *stop) {
		if ([ruleEnumerator.type isEqualToString:@"FORMAT"]) {
			[self processRule:ruleEnumerator parameters:parameters];
		}
    }];
	
	// binary output padding
	[self binaryOutputPadding:parameters option:optionOut optionNonBinary:optionNonBinary];
	
	// build the output
	returnString = [self buildGrammer:optionOut.grammar parameters:parameters];
	
	[input release];
	
	return returnString;
}

+ (NSString *)translateData:(NSData *)input withParameterList:(NSString *)parameterList outputFormat:(NSString *)outputFormat
{
	// convert to binary string
	NSString *inputString = [EPCConvert convertToBinaryStringFromData:input];
	
	return [self translate:inputString withParameterList:parameterList outputFormat:outputFormat];
}

+ (NSMutableDictionary *)parseParameterList:(NSString *)parameterList
{
	NSMutableDictionary *parameters = [[[NSMutableDictionary alloc] initWithCapacity:20] autorelease];
	
	// get parameters
	if ([parameterList length] > 0) {
		NSArray *chunks = [parameterList componentsSeparatedByString:@";"];	
        [chunks enumerateObjectsUsingBlock:^(NSString *aChunck, NSUInteger index, BOOL *stop) {
            NSArray *keyValue = [aChunck componentsSeparatedByString:@"="];
            NSString *key = [keyValue objectAtIndex:0];
            NSString *value = [keyValue objectAtIndex:1];
            [parameters setValue:value forKey:key];
        }];
	}
	
	return parameters;
}

+ (EPCPrefixMatch *)getPrefixMatch:(NSString *)input parameters:(NSMutableDictionary *)parameters
{
	NSMutableArray *remainingPrefixMatches = [[[NSMutableArray alloc] initWithCapacity:3] autorelease];

    [prefixMatchArray enumerateObjectsUsingBlock:^(EPCPrefixMatch *prefixMatch, NSUInteger index, BOOL *stop){
        if (prefixMatch.prefix != nil
            && [input hasPrefix:prefixMatch.prefix]) {
            [remainingPrefixMatches addObject:prefixMatch];
        }
    }];
    if ([remainingPrefixMatches count] > 1) {
        [remainingPrefixMatches enumerateObjectsUsingBlock:^(EPCPrefixMatch *prefixMatch, NSUInteger index, BOOL *stop){
            if (prefixMatch.scheme.tagLength != [[parameters valueForKey:@"taglength"] unsignedIntValue]) {
                [remainingPrefixMatches removeObject:prefixMatch];
            }
        }];
    }
    
    EPCPrefixMatch *result = nil;
    if ([remainingPrefixMatches count] > 1) {
        [remainingPrefixMatches removeAllObjects];
    }
    else if ([remainingPrefixMatches count] == 1) {
        result = [remainingPrefixMatches objectAtIndex:0];
    }
    return result;
}

+ (EPCOption *)getOption:(NSString *)input parameters:(NSMutableDictionary *)parameters level:(EPCLevel *)level
{
	[input retain];
	
	NSString *optionKey = level.scheme.optionKey;
	NSString *optionKeyValue = [parameters valueForKey:optionKey];
	
	NSMutableArray *suitableOptions = [[[NSMutableArray alloc] initWithCapacity:1] autorelease];
	
    [level.options enumerateObjectsUsingBlock:^(EPCOption *optionEnumerator, NSUInteger index, BOOL *stop){
        @autoreleasepool {
            NSRegularExpression *regularExpression = [NSRegularExpression regularExpressionWithPattern:optionEnumerator.pattern options:(NSRegularExpressionCaseInsensitive | NSRegularExpressionDotMatchesLineSeparators) error:nil];
            NSUInteger numberOfMatches = [regularExpression numberOfMatchesInString:input options:0 range:NSMakeRange(0, [input length])];
            regularExpression = nil;
            
            if (numberOfMatches > 0) {
                if ((optionKeyValue == nil)
                    || ([[optionEnumerator optionKey] isEqualToString:optionKeyValue])) {
                    [suitableOptions addObject:optionEnumerator];
                }
            }
        }
    }];
	[input release];

    if ([suitableOptions count] == 1) {
        return [suitableOptions objectAtIndex:0];
    }
#ifdef DEBUG
    else if ([suitableOptions count] == 0) {
            NSLog(@"NedapEPCGlobalGen2: No suitable options found.");
    }
    else if ([suitableOptions count] > 1) {
        NSLog(@"NedapEPCGlobalGen2: Too much options found.");
    }
#endif
    return nil;
}

+ (void)binaryInputPadding:(NSMutableDictionary *)parameters option:(EPCOption *)option optionNonBinary:(EPCOption *)optionNonBinary
{
	// input is binary
	if ([option.level.type isEqualToString:@"BINARY"])
	{
		for (EPCField *fieldEnumerator in option.fields)
		{
			NSString *name = fieldEnumerator.name;
			NSString *value = [parameters valueForKey:name];
			
			EPCField *fieldNonBinary = [[optionNonBinary fields] objectAtIndex:fieldEnumerator.seq - 1];
			
			// compaction is specified
			if (fieldEnumerator.compaction != nil)
			{
#ifdef DEBUG
				NSLog(@"NedapEPCGlobalGen2: Required compaction is not implemented.");
#endif
				// TODO: Implement compaction.
			}
			// compaction is not specified
			else 
			{
				// strip bits
				if (fieldEnumerator.padDir != nil)
				{
					if ([fieldEnumerator.bitPadDir isEqualToString:@"LEFT"])
					{
						while ([value characterAtIndex:0] == '0')
						{
							value = [value substringFromIndex:1];
						}
					}
					else if ([fieldEnumerator.bitPadDir isEqualToString:@"RIGHT"])
					{
						while ([value characterAtIndex:[value length] - 1] == '0')
						{
							value = [value substringToIndex:[value length] - 1];
						}
					}
				}
				
				// consider sequence of bits as unsigned integer
				value = [self binaryToDecimal:value];
			}
			
			// pad char is defined at binary level
			if (fieldEnumerator.padChar != nil)
			{
				// padchar is defined in non-binary level
				if (fieldNonBinary.padChar != nil)
				{
					NSLog(@"NedapEPCGlobalGen2: Invalid TDT definition file.");
				}
				// padchar is not defined in binary level
				else 
				{
					if ([fieldEnumerator.bitPadDir isEqualToString:@"LEFT"])
					{
						while ([value characterAtIndex:0] == [fieldEnumerator.padChar characterAtIndex:0])
						{
							value = [value substringFromIndex:1];
						}
					}
					else if ([fieldEnumerator.bitPadDir isEqualToString:@"RIGHT"])
					{
						while ([value characterAtIndex:[value length] - 1] == [fieldEnumerator.padChar characterAtIndex:0])
						{
							value = [value substringToIndex:[value length] - 1];
						}
					}
				}
			}
			// pad char is not defined at binary level
			else 
			{
				// padchar is defined in non-binary level
				if (fieldNonBinary.padChar != nil)
				{
					NSMutableString *valueMutable = [[[NSMutableString alloc] initWithCapacity:250] autorelease];
					
					if ([fieldNonBinary.padDir isEqualToString:@"LEFT"])
					{
						while (([valueMutable length] + [value length]) < fieldNonBinary.length)
						{
							[valueMutable appendString:fieldNonBinary.padChar];
						}
						[valueMutable appendString:value];
					}
					else if ([fieldNonBinary.padDir isEqualToString:@"RIGHT"])
					{
						[valueMutable setString:value];
						while ([valueMutable length] < fieldNonBinary.length)
						{
							[valueMutable appendString:fieldNonBinary.padChar];
						}
					}
					
					value = valueMutable;
				}
			}
			
			[parameters setValue:value forKey:name];	
		}
	}
}

+ (void)binaryOutputPadding:(NSMutableDictionary *)parameters option:(EPCOption *)option optionNonBinary:(EPCOption *)optionNonBinary
{
	// output is binary
	if ([option.level.type isEqualToString:@"BINARY"])
	{
		for (EPCField *fieldEnumerator in option.fields)
		{
			NSString *name = fieldEnumerator.name;
			NSString *value = [parameters valueForKey:name];
			
			EPCField *fieldNonBinary = [[optionNonBinary fields] objectAtIndex:fieldEnumerator.seq - 1];
			
			// padchar is defined in non binary level
			if (fieldNonBinary.padChar != nil)
			{
				// padchar is defined in binary level
				if (fieldEnumerator.padChar != nil)
				{
					NSLog(@"NedapEPCGlobalGen2: Invalid TDT definition file.");
				}
				// padchar is not defined in binary level
				else 
				{
					if ([fieldEnumerator.bitPadDir isEqualToString:@"LEFT"])
					{
						while ([value characterAtIndex:0] == [fieldNonBinary.padChar characterAtIndex:0])
						{
							value = [value substringFromIndex:1];
							
							if ([value length] == 0)
								break;
						}
					}
					else if ([fieldEnumerator.bitPadDir isEqualToString:@"RIGHT"])
					{
						while ([value characterAtIndex:[value length] - 1] == [fieldNonBinary.padChar characterAtIndex:0])
						{
							value = [value substringToIndex:[value length] - 1];
							
							if ([value length] == 0)
								break;
						}
					}
				}
			}
			// padchar is not defined in non binary level
			else 
			{
				if (fieldEnumerator.padChar != nil)
				{
					NSMutableString *valueMutable = [[[NSMutableString alloc] initWithCapacity:250] autorelease];
					[valueMutable setString:value];
					
					if ([fieldEnumerator.padDir isEqualToString:@"LEFT"])
					{
						while ([value length] < fieldEnumerator.length)
						{
							[valueMutable insertString:fieldEnumerator.padChar atIndex:0];
						}
					}
					else if ([fieldEnumerator.padDir isEqualToString:@"RIGHT"])
					{
						while ([value length] < fieldEnumerator.length)
						{
							[valueMutable insertString:fieldEnumerator.padChar atIndex:[valueMutable length]];
						}
					}
					
					value = valueMutable;
				}
			}
			
			if (fieldEnumerator.compaction != nil)
			{
#ifdef DEBUG
				NSLog(@"NedapEPCGlobalGen2: Required compaction not implemented.");
#endif
			}
			else 
			{
				value = [self decimalToBinary:value];
			}
			
			if (fieldEnumerator.bitPadDir != nil)
			{
				NSMutableString *valueMutable = [[[NSMutableString alloc] initWithCapacity:250] autorelease];
				
				if ([fieldEnumerator.bitPadDir isEqualToString:@"LEFT"])
				{
					while (([valueMutable length] + [value length]) < fieldEnumerator.bitLength)
					{
						[valueMutable appendString:@"0"];
					}
					
					[valueMutable appendString:value];
				}
				else if ([fieldEnumerator.bitPadDir isEqualToString:@"RIGHT"])
				{
					[valueMutable setString:value];
					
					while ([valueMutable length] < fieldEnumerator.bitLength)
					{
						[valueMutable appendString:@"0"];
					}
				}
				value = valueMutable;
			}
			
			[parameters setValue:value forKey:name];
		}
	}	
}

+ (NSString *)calculateGS1Checksum:(NSString *)input 
{
	NSInteger checksum;
	NSInteger weight;
	NSInteger total = 0;
	NSInteger length = [input length];
	NSInteger d;
	for (NSInteger i = 0; i < length; i++)
	{
		if (i % 2 == 0)
		{
			weight = -3;
		}
		else 
		{
			weight = -1;
		}
		NSRange substringRange = {
			length - 1 - i,
			1
		};
		d = [[input substringWithRange:substringRange] intValue];
		total += weight * d;
	}
	checksum = (10 + total % 10) % 10;
	return [NSString stringWithFormat:@"%i", checksum];
}

+ (void)processRule:(EPCRule *)rule parameters:(NSMutableDictionary *)parameters
{
	NSString *newFieldName = rule.fieldNameNew;
	NSString *ruleFunction = rule.function;
    
	NSCharacterSet *searchSet = [NSCharacterSet characterSetWithCharactersInString:@"("];
	NSRange braceOpening = [ruleFunction rangeOfCharacterFromSet:searchSet];
	NSString *ruleFunctionName = [ruleFunction substringToIndex:braceOpening.location];
	
	NSRange parametersRange = {
		braceOpening.location + 1,
		[ruleFunction length] - braceOpening.location - 2
	};
	NSArray *functionParameters = [[ruleFunction substringWithRange:parametersRange] componentsSeparatedByString:@","];
	
	if ([ruleFunctionName isEqualToString:@"TABLELOOKUP"])
	{
#ifdef DEBUG
		NSLog(@"NedapEPCGlobalGen2: Required function TABLELOOKUP not implemented.");
#endif
		// TODO: Implement rule TABLELOOKUP.
	}
	
	if ([ruleFunctionName isEqualToString:@"LENGTH"])
	{
		NSString *inputValue = [parameters objectForKey:[functionParameters objectAtIndex:0]];
		NSUInteger length = [inputValue length];
		
		[parameters setValue:[NSString stringWithFormat:@"%u", length] forKey:newFieldName];
	}
	
	if ([ruleFunctionName isEqualToString:@"SUBSTR"])
	{	
		NSString *inputValue = [parameters objectForKey:[functionParameters objectAtIndex:0]];
		NSString *outputValue = @"";
		
		if ([functionParameters count] == 2)
		{
			NSInteger start = [self getIntValue:[functionParameters objectAtIndex:1] parameters:parameters];
			
			outputValue = [inputValue substringFromIndex:start];		
		}
		
		if ([functionParameters count] == 3)
		{
			NSInteger start = [self getIntValue:[functionParameters objectAtIndex:1] parameters:parameters];
			NSInteger length = [self getIntValue:[functionParameters objectAtIndex:2] parameters:parameters];
			NSRange substringRange = {
				start,
				length
			};

			outputValue = [inputValue substringWithRange:substringRange];
		}
		
		[parameters setValue:outputValue forKey:newFieldName];
	}
	
	if ([ruleFunctionName isEqualToString:@"GS1CHECKSUM"])
	{
		NSString *inputValue = [parameters objectForKey:[functionParameters objectAtIndex:0]];
		NSString *outputValue = [self calculateGS1Checksum:inputValue];
		[parameters setValue:outputValue forKey:newFieldName];
	}
	
	if ([ruleFunctionName isEqualToString:@"CONCAT"])
	{
		NSMutableString *buffer = [[[NSMutableString alloc] initWithCapacity:250] autorelease];
		for (NSString *parameter in functionParameters)
		{
			NSString *value = [parameters valueForKey:parameter];
			if (value != nil)
			{
				[buffer appendString:value];
			}
			else 
			{
				[buffer appendString:parameter];
			}
		}
		
		[parameters setValue:buffer forKey:newFieldName];
	}
}

+ (NSInteger)getIntValue:(NSString *)fieldName parameters:(NSMutableDictionary *)parameters
{
	NSString *result = [parameters objectForKey:fieldName];
	
	if (result != nil)
	{	
		return [result intValue];
	}
	else 
	{
		return [fieldName intValue];
	}

	// NOTE: removed regular expression due to speed issues (@"^\\d+$")
}

+ (EPCLevel *)findOutputLevelFromScheme:(EPCScheme *)scheme withString:(NSString *)outputFormat
{
	for (EPCLevel *levelEnumerator in scheme.levels)
	{
		if ([levelEnumerator.type isEqualToString:outputFormat])
		{
			return levelEnumerator;
		}
	}
	
	return nil;
}

+ (EPCOption *)findOutputOptionFromLevel:(EPCLevel *)level withOption:(EPCOption *)option
{
	for (EPCOption *optionEnumerator in level.options)
	{
		if ([optionEnumerator.optionKey isEqualToString:option.optionKey])
		{
			return optionEnumerator;
		}
	}
	
	return nil;
}
static NSNumberFormatter *numberFormatter;

+ (NSString *)decimalToBinary:(NSString *)decimal
{
    if (numberFormatter == nil) {
        numberFormatter = [[NSNumberFormatter alloc] init];
        [numberFormatter setNumberStyle:NSNumberFormatterDecimalStyle];
    }
	NSNumber *decimalNumber = [numberFormatter numberFromString:decimal];
	unsigned long long decimalUnsignedLongLong = 0;
	
	NSMutableString *binaryString = [[[NSMutableString alloc] initWithCapacity:250] autorelease];
	for(decimalUnsignedLongLong = [decimalNumber unsignedLongLongValue]; decimalUnsignedLongLong > 0ULL; decimalUnsignedLongLong >>= 1ULL) {
		// Prepend "0" or "1", depending on the bit
		[binaryString insertString:((decimalUnsignedLongLong & 1ULL) ? @"1" : @"0") atIndex:0];
	}
    decimalNumber = nil;

	return binaryString;
}


+ (NSString *)binaryToDecimal:(NSString *)binary
{
	unsigned long long decimal = 0;
	unsigned long long power = 1;
	for (NSUInteger i = [binary length]; i > 0 ; i--)
	{
		NSRange substringRange = {
			i - 1,
			1
		};
		decimal += ((unsigned long long) [[binary substringWithRange:substringRange] intValue]) * power;
		power *= 2ULL;
	}
	
	return [NSString stringWithFormat:@"%llu", decimal];
}

+ (NSString *)buildGrammer:(NSString *)grammerField parameters:(NSMutableDictionary *)parameters
{
	NSMutableString *outboundString = [[[NSMutableString alloc] initWithCapacity:250] autorelease];

    NSArray *matches = [grammerField componentsSeparatedByString:@" "];
    for (NSString *item in matches) {
		if ([item characterAtIndex:0] == '\'') {// if the item starts with ', append the item-string without '
			NSRange rangeWithoutChars = {
				1,
				[item length] - 2
			};
			[outboundString appendString:[item substringWithRange:rangeWithoutChars]];
		}
		else  {// otherwise fetch the value for this key
            [outboundString appendString:[parameters valueForKey:item]];
		}
    }

	return outboundString;
}


+ (void)parser:(NSXMLParser *)parser didStartElement:(NSString *)elementName 
  namespaceURI:(NSString *)namespaceURI qualifiedName:(NSString *)qName
  attributes:(NSDictionary *)attributeDict 
{
	
	if ([elementName isEqualToString:@"tdt:epcTagDataTranslation"])
	{
		epcTagDataTranslation.version = [attributeDict valueForKey:@"version"];
		epcTagDataTranslation.epcTDSVersion = [attributeDict valueForKey:@"epcTDSVersion"];
		epcTagDataTranslation.date = [attributeDict valueForKey:@"date"];
	}
	
	if ([elementName isEqualToString:@"scheme"])
	{
		// add new scheme
		EPCScheme *scheme = [[EPCScheme alloc] init];
		[epcTagDataTranslation.schemes addObject:scheme];
		currentScheme = scheme;
		[scheme release];
		
		scheme.name = [attributeDict valueForKey:@"name"];
		scheme.optionKey = [attributeDict valueForKey:@"optionKey"];
		scheme.tagLength = [[attributeDict valueForKey:@"tagLength"] intValue];
	}
	
	if ([elementName isEqualToString:@"level"])
	{
		// add new level
		EPCLevel *level = [[EPCLevel alloc] init];
		[currentScheme.levels addObject:level];
		currentLevel = level;
		[level release];
		
		level.type = [attributeDict valueForKey:@"type"];
		level.prefixMatch = [attributeDict valueForKey:@"prefixMatch"];
		level.requiredParsingParameters = [attributeDict valueForKey:@"requiredParsingParameters"];
		level.requiredFormattingParameters = [attributeDict valueForKey:@"requiredFormattingParameters"];
		level.scheme = currentScheme;
	}
	
	if ([elementName isEqualToString:@"option"])
	{
		// add new option
		EPCOption *option = [[EPCOption alloc] init];
		[currentLevel.options addObject:option];
		currentOption = option;
		[option release];
		
		option.optionKey = [attributeDict valueForKey:@"optionKey"];
		option.pattern = [attributeDict valueForKey:@"pattern"];
		option.grammar = [attributeDict valueForKey:@"grammar"];
		option.level = currentLevel;
	}
	
	if ([elementName isEqualToString:@"field"])
	{
		// add new field
		EPCField *field = [[EPCField alloc] init];
		[currentOption.fields addObject:field];
		[field release];

		field.seq = [[attributeDict valueForKey:@"seq"] intValue];
		field.name = [attributeDict valueForKey:@"name"];
		field.bitLength = [[attributeDict valueForKey:@"bitLength"] intValue];
		field.characterSet = [attributeDict valueForKey:@"characterSet"];
		field.compaction = [attributeDict valueForKey:@"compaction"];
		field.compression = [attributeDict valueForKey:@"compression"];
		field.padChar = [attributeDict valueForKey:@"padChar"];
		field.padDir = [attributeDict valueForKey:@"padDir"];
		field.bitPadDir = [attributeDict valueForKey:@"bitPadDir"];
		field.decimalMinimum = [attributeDict valueForKey:@"decimalMinimum"];
		field.decimalMaximum = [attributeDict valueForKey:@"decimalMaximum"];
		field.length = [[attributeDict valueForKey:@"length"] intValue];
		field.option = currentOption;
	}
	
	if ([elementName isEqualToString:@"rule"])
	{
		// add new rule
		EPCRule *rule = [[EPCRule alloc] init];
		[currentLevel.rules addObject:rule];
		[rule release];
		
		rule.type = [attributeDict valueForKey:@"type"];
		rule.inputFormat = [attributeDict valueForKey:@"inputFormat"];
		rule.seq = [attributeDict valueForKey:@"seq"];
		rule.fieldNameNew = [attributeDict valueForKey:@"newFieldName"];
		rule.characterSet = [attributeDict valueForKey:@"characterSet"];
		rule.length = [[attributeDict valueForKey:@"length"] intValue];
		rule.padChar = [attributeDict valueForKey:@"padChar"];
		rule.padDir = [attributeDict valueForKey:@"padDir"];
		rule.bitPadDir = [attributeDict valueForKey:@"bitPadDir"];
		rule.decimalMinimum = [[attributeDict valueForKey:@"decimalMinimum"] intValue];
		rule.decimalMaximum = [[attributeDict valueForKey:@"decimalMaximum"] intValue];
		rule.function = [attributeDict valueForKey:@"function"];
		rule.tableURL = [attributeDict valueForKey:@"tableURL"];
		rule.tableParams = [attributeDict valueForKey:@"tableParams"];
		rule.tableXPath = [attributeDict valueForKey:@"tableXPath"];
		rule.tableSQL = [attributeDict valueForKey:@"tableSQL"];
	}		
}

+ (void)parser:(NSXMLParser *)parser didEndElement:(NSString *)elementName
  namespaceURI:(NSString *)namespaceURI qualifiedName:(NSString *)qName {
	if ([elementName isEqualToString:@"scheme"])
		currentScheme = nil;
	
	if ([elementName isEqualToString:@"level"])
		currentLevel = nil;
	
	if ([elementName isEqualToString:@"option"])
		currentOption = nil;
}

+ (void)parser:(NSXMLParser *)parser parseErrorOccured:(NSError *)parseError {
	NSLog(@"NedapEPCGlobalGen2: NXSMLParserError: %@", [[parser parserError] localizedDescription]);
}

@end
