//
//  NHPresentation.h
//  NedapHandheldAPI
//
//  Created by Danny Haak on 04-07-10.
//  Copyright 2010 N.V. Nederlandsche Apparatenfabriek "Nedap". All rights reserved.
//

#import <Foundation/Foundation.h>

#import "NHDefines.h"
#import "NHSessionDelegateProtocol.h"

#import "ihex.h"

@class NHSession;
@class NHSelect;
@class NHPassword;
@class NHPresentationContext;

typedef enum {
	NHPresentationRFIDReadingPowerLevelLow  = 0,
	NHPresentationRFIDReadingPowerLevelHigh = 1
} NHPresentationRFIDReadingPowerLevel;

/**
 Presentation layer: high-level communication layer to communicate with Handheld Reader. 
 */

@interface NHPresentation : NSObject <NHSessionDelegateProtocol> {  
    // rfid inventory properties
    BOOL rfidInventoryHighVolume;
    NHPresentationRFIDReadingPowerLevel rfidInventoryPowerLevel;
    NHSelect *rfidInventorySelect;
    NSString *rfidInventoryBuffer;
    
    // rfid monitor properties
    NHPresentationRFIDReadingPowerLevel rfidMonitorPowerLevel;
    NHSelect *rfidMonitorSelect;
    
    // barcode reading properties
    NSArray *barcodeTypes;
    NSUInteger barcodeReadDuration;
    BOOL (^barcodeValidationBlock)(id);
}

@property (nonatomic) BOOL rfidInventoryHighVolume;
@property (nonatomic) NHPresentationRFIDReadingPowerLevel rfidInventoryPowerLevel;
@property (nonatomic, retain) NHSelect *rfidInventorySelect;
@property (nonatomic, retain) NSString *rfidInventoryBuffer;

@property (nonatomic) NHPresentationRFIDReadingPowerLevel rfidMonitorPowerLevel;
@property (nonatomic, retain) NHSelect *rfidMonitorSelect;

@property (nonatomic, retain) NSArray *barcodeTypes;
@property (nonatomic) NSUInteger barcodeReadDuration;
@property (nonatomic, copy) BOOL (^barcodeValidationBlock)(id);

/**
 NHPresentation singleton.
 @returns singleton
 */
+ (NHPresentation *)sharedPresentation;

#pragma mark -
#pragma mark general

/**
 Notifies listeners
 */
+ (void) notify:(NSString *)notificationName object:(id)theObject userInfo:(id)theUserInfo;

/**
 Gets reading power
 @return read power
 */
+ (NSUInteger)actionRFIDGetReadPower;

/**
 Sets reading power
 @param power the new value for read power
 */
+ (void)actionRFIDSetReadPower:(NSUInteger)power;

/**
 Check if the handheld is connected.
 @returns bool that states whether the handheld is connected or not
 */
+ (BOOL)handheldIsConnected;

/**
 Let the handheld disconnect from the device it is connected to.
 */
+ (void)handheldDisconnect;

/**
 Retrieves the percentage of battery capacity that is remaining in the !d hand.
 */
+ (void)handheldGetBatteryPercentage;

/**
 Stop the current operation, e.g. inventory or monitoring. Can be used when switching from one view to another.
 */
+ (void)handheldStop;

/**
	Get information on the handheld. 
	@returns dictionary with four parameters on the handheld.
 */
+ (NSDictionary *)handheldGetInformation;

/**
 Returns the included firmware version of the !D Hand in the API.
 @returns string with version, formatted like x.y.z.
 */
+ (NSString *)handheldUpdateFirmwareAvailableVersion;

+ (BOOL)handheldUpdateFirmwareAvailable;

/**
 Update the firmware of the !D Hand.
 */
+ (void)handheldUpdateFirmware;

#pragma mark user interaction

+ (void)uiBeeperEnabled:(BOOL)enabled;
+ (void)uiVibrationEnabled:(BOOL)enabled;

/**
 Let the beeper on the handheld play a tune.
 @param tune type of tune
 */
+ (void)uiPlayTune:(NHUITune)tune;

/**
 Let the vibration motor on the handheld vibrate.
 */
+ (void)uiVibrate;

/**
 Let the vibration motor on the handheld play a tune.
 @param tune type of tune
 */
+ (void)uiVibrateTune:(NHUITune)tune;

/** 
 Let the beeper and vibration motor play a tune
 @param tune type of tune
 */
+ (void)actionUITune:(NHUITune)beepTune;

#pragma mark rfid

+ (void)rfidSetRegion:(NHRFIDRegion)region;

/**
    The inventory methods provide an easy way to do inventory. It automatically filters ‘multiple’ reads in a buffer and generates a ‘beep’ on new tags. EPC air interface parameters are set automatically to session 1, target A (no switching) and automatic Q.
 */
+ (void)rfidInventoryStart;
/**
    @see + (void)rfidInventoryStart
    @param session RFID read session to use
 */
+ (void)rfidInventoryStartWithSession:(NHRFIDSession)session;

/**	
    @see + (void)rfidInventoryStart
    @param select select statement
 */
+ (void)rfidInventoryStartWithSelect:(NHSelect *)select;
/**	
    @see + (void)rfidInventoryStart
    @param select select statement
    @param session RFID read session to use
 */
+ (void)rfidInventoryStartWithSelect:(NHSelect *)select withSession:(NHRFIDSession)session;

/**
    @see + (void)rfidInventoryStart
	@param bufferName name of the buffer
 */
+ (void)rfidInventoryStartInBuffer:(NSString *)bufferName;
/**
    @see + (void)rfidInventoryStart
    @param bufferName name of the buffer
    @param session RFID read session to use
 */
+ (void)rfidInventoryStartInBuffer:(NSString *)bufferName withSession:(NHRFIDSession)session;

/**
 @see + (void)rfidInventoryStart
	@param bufferName name of the buffer
	@param select select statement
 */
+ (void)rfidInventoryStartInBuffer:(NSString *)bufferName withSelect:(NHSelect *)select;
/**
    @see + (void)rfidInventoryStart
    @param bufferName name of the buffer
    @param select select statement
    @param session RFID read session to use
 */
+ (void)rfidInventoryStartInBuffer:(NSString*)bufferName withSelect:(NHSelect *)select withSession:(NHRFIDSession)session;


+ (void)rfidInventoryHighVolumeStart;
+ (void)rfidInventoryHighVolumeStartWithSelect:(NHSelect *)select;
+ (void)rfidInventoryHighVolumeStartInBuffer:(NSString *)bufferName;
+ (void)rfidInventoryHighVolumeStartInBuffer:(NSString *)bufferName withSelect:(NHSelect *)select;

+ (void)rfidInventoryLowPowerStart;
+ (void)rfidInventoryLowPowerStartInBuffer:(NSString *)bufferName;

/**
	Resets the RFID inventory 'seen' list of the "default" buffer. 
	@attention Please make sure that if a list is also kept in the application layer, clean it after the relevant notification is received.
 */
+ (void)rfidInventoryReset;

/**
	Resets the RFID inventory 'seen' list of the custom buffer. 
	@param bufferName name of the buffer
 */
+ (void)rfidInventoryResetBuffer:(NSString *)bufferName;

/**
 Fills the 'default' inventory buffer with hashes of previously read EPCs. Use when application was terminated earlier and continuing reading
 @param dataSet the set with binaryData of EPCs to add
 @attention Please make sure that if a list is also kept in the application layer, clean it after the relevant notification is received.
 */
+ (void)rfidInventoryBufferFill:(NSSet*)dataSet;

/**
 Fills the specified inventory buffer with hashes of previously read EPCs. Use when application was terminated earlier and continuing reading
 @param dataSet the set with binaryData of EPCs to add
 @param bufferName name of the buffer
 */
+ (void)rfidInventoryBufferFill:(NSSet*)dataSet forBuffer:(NSString *)bufferName;

/**
 Returns the rfid buffer containing the indices of the read EPCs
 @return set of hashed index values
 @see EPCCode:hash
 */
+ (NSSet *)rfidInventoryBuffer;

/**
 Returns the rfid buffer containing the indices of the read EPCs
 @param bufferName name of the buffer
 @return set of hashed index values
 @see EPCCode:hash
 */
+ (NSSet *)rfidInventoryBufferWithName:(NSString*)bufferName;

/** 
 Adds EPC to list of observed items + adds EPC hash
 @param epc The EPC to add
 */
+ (void)addEPC:(EPCCode*)epc;

/**
	Continously return tags in the field, together with their RSSI values. Can be used to implement 'search' functionality.
 */
+ (void)rfidMonitorStart;

/**
	Monitors the tags in the field, with a select statement.
	@param select select statement
 */
+ (void)rfidMonitorStartWithSelect:(NHSelect *)select;

/**
 Does a single read cycle of tags.
 */
+ (void)rfidReadSingle;

/**
 Does a single read cycle of tags, with a select statement.
 @param select select statement
 */
+ (void)rfidReadSingleWithSelect:(NHSelect *)select;

/**
	To obtain the tag manufacturer ID (MDID), the tag model number, or the tag’s extended identifier field (XTID), use this method.
 */
+ (void)rfidReadDetailedTagInformation;

/**
 To obtain the tag manufacturer ID (MDID), the tag model number, or the tag’s extended identifier field (XTID), use this method.
 */
+ (void)rfidReadDetailedTagInformationFromTag:(EPCCode *)code;

/**
	Write a new EPC value to a tag.
	@param newEPCCode the new EPC value
 */
+ (void)rfidWriteEPC:(EPCCode *)newEPCCode;

#pragma mark barcode
 
/**
	Check to see if the handheld reader has a barcode reader on-board.
	@returns if true, there is a barcode reader
 */
+ (BOOL)barcodeHasReader;

+ (void)barcodeStartReading;

+ (void)barcodeResetParameters;

/**
	Read a barcode with a specified type, and with a specified timeout.
	@param barcodeType type of barcode
	@param duration timeout in seconds
 */
+ (void)barcodeReadType:(NHBarcodeType)barcodeType duringSeconds:(NSUInteger)duration;

/**
	Read a barcode with a specified array of types, and with a specified timeout.
	@param barcodeTypes array of barcode types (NHBarcodeType)
	@param duration timeout in seconds
 */
+ (void)barcodeReadTypes:(NSArray *)barcodeTypes duringSeconds:(NSUInteger)duration;



#pragma mark nfc

+ (void)nfcReadStart;

@end