//
//  NHSession.h
//  NedapHandheldAPI
//
//  Created by Danny Haak on 16-03-10.
//  Copyright 2010 N.V. Nederlandsche Apparatenfabriek "Nedap". All rights reserved.
//

#import <Foundation/Foundation.h>
#import <UIKit/UIKit.h>
#import <ExternalAccessory/ExternalAccessory.h>

#import "NHSessionDelegateProtocol.h"

typedef enum {
	NHRFIDLockKillPasswordReadWrite		= 1 << 9,
	NHRFIDLockKillPasswordPermanent		= 1 << 8,
	NHRFIDLockAccessPasswordReadWrite	= 1 << 7,
	NHRFIDLockAccessPasswordPermanent	= 1 << 6,
	NHRFIDLockEPCMemoryReadWrite		= 1 << 5,
	NHRFIDLockEPCMemoryPermanent		= 1 << 4,
	NHRFIDLockTIDMemoryReadWrite		= 1 << 3,
	NHRFIDLockTIDMemoryPermanent		= 1 << 2,
	NHRFIDLockUserMemoryReadWrite		= 1 << 1,
	NHRFIDLockUserMemoryPermanent		= 1 << 0
} NHRFIDLock;

@class NHSelect;
@class NHPassword;

/**
 Session layer: low-level communication layer to communicate with Handheld Reader. 
 */
@interface NHSession : NSObject <EAAccessoryDelegate, NSStreamDelegate>
{
}

/**
 NHSession singleton.
 @returns singleton
 */
+ (NHSession *)sharedSession;

#pragma mark -
#pragma mark management

/**
	Sets the delegate for NHSession.
	@param delegate The delegate that handles the callback, implementing the NHSessionDelegateProtocol protocol.
 */
+ (void)setDelegate:(id <NHSessionDelegateProtocol>)delegate;

#pragma mark general

/**
 Checks if the accessory is connected or not.
 @returns connected
 */
+ (BOOL)handheldIsConnected;

/**
 Returns the current battery status in percentage.
 */
+ (void)handheldGetBatteryPercentage;

/**
 Will let the handheld kill the bluetooth connection. Can be used to pair the handheld with another device.
 */
+ (void)handheldDisconnect;

/**
 Get information on the handheld. 
 @returns dictionary with four parameters on the handheld.
 */
+ (NSDictionary *)handheldGetInformation;

/**
 Let the handheld go to the bootloader, without losing the Bluetooth connection.
 */
+ (void)handheldBootloader;

+ (void)handheldBootloaderSendPage:(NSData *)page address:(NSUInteger)address;
+ (void)handheldBootloaderFinish:(NSUInteger)crc;

#pragma mark user interaction

/**
 The beeper on the handheld will play the defined tune.
 @param tune specify a tune
 */
+ (void)uiPlayTune:(NHUITune)tune;

/**
 The vibration motor on the handheld will vibrate.
 */
+ (void)uiVibrate;

/**
 The vibration motor on the handheld will vibrate for a specified duration.
 @param timeout duration in tens of a second
 */
+ (void)uiVibrate:(NSUInteger)timeout;

/**
	The vibration motor on the handheld will vibrate a specified tune.
	@param tune specify a tune
 */
+ (void)uiVibrateTune:(NHUITune)tune;

/**
 Let the white status led on the handheld blink or not.
 @param blinking blinking enabled
 */
+ (void)uiSetStatusLedBlinking:(BOOL)blinking;

#pragma mark rfid

/**
	Gets the supported regions by the handheld reader.
 */
+ (void)rfidGetSupportedRegions;

/**
	Sets the region that the handheld reader operates in. The supported regions can be obtained by using +(void)rfidGetSupportedRegions.
	@param region region
 */
+ (void)rfidSetRegion:(NHRFIDRegion)region;

/**
	Set frequency hop table.
	@param frequencies array with frequencies (NSNumber containing NHRFIDFrequency)
 */
+ (void)rfidSetFrequencyHopTable:(NSArray *)frequencies;

/**
	Sets the session for the EPC Gen2 protocol.
	@param session session (S0, S1, S2 or S3)
 */
+ (void)rfidSetSession:(NHRFIDSession)session;

/**
	Sets the target for the EPC Gen2 protocol.
	@param target target (A or B)
	@param toggle toggle between A and B
 */
+ (void)rfidSetTarget:(NHRFIDTarget)target toggle:(BOOL)toggle;

/**
	Sets the miller value for the EPC Gen2 protocol.
	@param mValue miller value
 */
+ (void)rfidSetMValue:(NHRFIDMValue)mValue;

/**
	Sets the Q value for the EPC Gen2 protocol.
	@param qValue the Q value
	@param dynamic whether the Q value is dynamically adjusted or not
 */
+ (void)rfidSetQValue:(NSUInteger)qValue dynamic:(BOOL)dynamic;

/**
	Will read RFID tags in the field, for a specified duration, with a specific power and with a select to select a subset of tags.
	@param timeout timeout in milliseconds
	@param power the power, possible values range from 10 to 23 dBm
	@param select select statement, can also be set to nil
 */
+ (void)rfidReadEPCsWithTimeout:(NSUInteger)timeout power:(NSUInteger)power select:(NHSelect *)select;

/**
	Will read specific data from a RFID tag.
	@param memoryBank the memory bank to read data from
	@param addressInWords the address to read data from (in words)
	@param lengthInWords the length of the data to read (in words)
	@param timeout timeout in milliseconds
	@param power the power, possible values range from 10 to 23 dBm
	@param select select statement, can also be set to nil
	@param password password, can also be set to nil
 */
+ (void)rfidReadData:(NHRFIDMemoryBank)memoryBank addressInWords:(NSUInteger)addressInWords lengthInWords:(NSUInteger)lengthInWords timeout:(NSUInteger)timeout power:(NSUInteger)power select:(NHSelect *)select password:(NHPassword *)password;

/**
	Writes a new EPC code to a tag.
	@param epc new EPC code
 */
+ (void)rfidWriteEPC:(EPCCode *)epc;

/**
	Write data to a specified memory bank and address in a tag.
	@param someData the data
	@param memoryBank the memory bank where to write the data to
	@param addressInWords the address in words where the data needs to be written
	@param timeout timeout in milliseconds	
	@param power the power, possible values range from 10 to 23 dBm
	@param select select statement, can also be set to nil
	@param password password, can also be set to nil
 */
+ (void)rfidWriteData:(NSData *)someData memoryBank:(NHRFIDMemoryBank)memoryBank addressInWords:(NSUInteger)addressInWords timeout:(NSUInteger)timeout power:(NSUInteger)power select:(NHSelect *)select password:(NHPassword *)password;

/**
	Enables Continuous Wave (CW) signal, used for testing and measurement purposes.
	@param enabled enabled
	@param antenna select antenna one or two
 */
+ (void)rfidSetCWSignal:(BOOL)enabled onAntenna:(NSUInteger)antenna;

#pragma mark barcode

/**
	Check to see if the handheld reader has a barcode reader on-board.
	@returns if true, there is a barcode reader
 */
+ (BOOL)barcodeHasReader;

/**
	Enables or disable a certain type of barcode.
	@param barcodeType NHBarcodeType
	@param enabled enabled or disabled
 */
+ (void)barcodeSetStatusForType:(NHBarcodeType)barcodeType enabled:(BOOL)enabled;

/**
	Try to read a barcode for a defined number of seconds.
	@param duration duration in seconds
 */
+ (void)barcodeTryDecode:(NSUInteger)duration;

#pragma mark nfc

/**
	The handheld will try to read a NFC card.
 */
+ (void)nfcTryToReadCard;

@end